/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  GeneralIO.c
 * desc  :  supply basic functions for file handling
 * author:  lampke, keller
 *
 ************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "DynArray.h"			  /* Reuse-Library */
#include "Idents.h"			  /* Reuse-Library */
#include "FileIO.h"
#include "GeneralIO.h"
#include "Positions.h"

#define DEFMODPATH      "DEFMOD_PATH"

#define INITIALSIZE     4L

static BOOLEAN isInitialised = FALSE;

/*
 * local variables
 */
static char **PathTab;
static unsigned long PathTabSize;
static unsigned long ActPathCount;

/*
 * local function prototypes
 */
static void AddEnvPathes( char * pathes );

/*
 * local function implementation
 */
/*************************************************************************
 * name :  AddEnvPathes
 * in   :
 * out  :
 ************************************************************************/
static void AddEnvPathes( char * pathes )
{
  char *work;

  if (pathes != NULL) {

    /* The following three lines simply implement "pathes = strdup(pathes)" */
    work = (char *) malloc(strlen(pathes)+1); /* strdup is not in every */
    if (work != NULL) strcpy(work, pathes);   /* library. So here's the */
    pathes = work;			      /* explicit code for it as */
					      /* shown in K&R-C, 2nd ed. */
    do {
      work = strchr(pathes, PATHLISTDELIM);
      if (work != NULL) {
        *work = '\0';
      } /* endif */
      GeneralIO_AddSearchPath( pathes );
      pathes = work + 1;
    } while (work != NULL && *pathes != '\0');
  }
}

/*
 * global function implementation
 */
/*------- AddSearchPath ----------------------------------------------------*/
void GeneralIO_AddSearchPath( char *path )
{
  char *lastChar = path + strlen(path) - 1;

  if (*lastChar == '\\' || *lastChar == '/') {
    *lastChar = '\0';
  } /* endif */

  if ( ActPathCount == PathTabSize ) {
    ExtendArray((char **) &PathTab, &PathTabSize, sizeof(char *));
  } /* endif */

  PathTab[ActPathCount++] = path;
}

/*------- CheckOpen --------------------------------------------------------*/

void GeneralIO_CheckOpen( FILE *File, char *FileName )
{
  if (File == NULL) {
    fprintf(stderr, " couldn't open file '%s'\n" , FileName );
    FileIO_CloseAll();
    exit(1);
  }
}

/*------- CheckModFile -----------------------------------------------------*/

BOOLEAN GeneralIO_CheckModFile( char *ProgName, char *FileName )
{
  FILE *f;

  if ((f = FileIO_OpenInput(ProgName)) != NULL) {
    strcpy( FileName, ProgName );
    FileIO_Close(f);
    return TRUE;
  } else {
    strcpy(FileName, ProgName);
    strcat(FileName, MODFILE_EXT);
    if ((f = FileIO_OpenInput(FileName)) != NULL) {
      FileIO_Close(f);
      return TRUE;
    }
  }
  return FALSE;
}

/*------- CheckDefFile -----------------------------------------------------*/

BOOLEAN GeneralIO_CheckDefFile( tIdent ModuleName, char *FileName )
{
  char  name[FILENAME_MAX];
  FILE *f;
  SHORT i = 0;

  GetString( ModuleName, name );
  strcat( name, DEFFILE_EXT );
  strcpy( FileName, name );
  if ((f = FileIO_OpenInput(FileName)) == NULL) {
    while ((f == NULL) && (i < ActPathCount)) {
      strcpy(FileName, PathTab[i++]);
      strcat(FileName, PATHDELIM);
      strcat(FileName, name);
      f = FileIO_OpenInput(FileName);
    }
  }
  if (f != NULL) {
    FileIO_Close(f);
    return TRUE;
  } else {
    return FALSE;
  }
}

/*------- RealName ---------------------------------------------------------*/

tIdent GeneralIO_RealName(tIdent Ident, tPosition Position)
{
  char Name[256], NewName[256], OldName[256];
  long ActualPos, DotPos, SpecialPos;

  GetString(Position.File, Name);
  DotPos = -1;
  SpecialPos = -1;
  for (ActualPos = 0; ActualPos < strlen(Name); ActualPos++) {
    if (Name[ActualPos] == '.') {
      DotPos = ActualPos; /* position of the last dot */
    } else if (!(('0' <= Name[ActualPos] && Name[ActualPos] <= '9') ||
                 ('a' <= Name[ActualPos] && Name[ActualPos] <= 'z') ||
                 ('A' <= Name[ActualPos] && Name[ActualPos] <= 'Z') ||
                 ('_' == Name[ActualPos]))) {
      SpecialPos = ActualPos; /* position of the last special character */
    }
  }
  if (DotPos != -1) { /* dot found */
    Name[DotPos] = '\0'; /* cut name */
  }
  if (SpecialPos != -1) { /* special character found */
    strcpy(NewName, &Name[SpecialPos + 1]);
  } else {
    strcpy(NewName, Name);
  }
  GetString(Ident, OldName);
  if (strlen(NewName) > 0 && strcmp(OldName, NewName) != 0) {
    return MakeIdent(NewName, strlen(NewName));
  } else {
    return Ident;
  }
}

/*------- OpenCHeader ------------------------------------------------------*/

FILE *GeneralIO_OpenCHeader(tIdent Ident)
{
  char fileName[FILENAME_MAX];
  FILE *f;

  GetString(Ident, fileName);
  strcat(fileName, CHEADER_EXT);
  f = FileIO_OpenOutput(fileName);
  GeneralIO_CheckOpen(f, fileName);
  return f;
}

/*------- OpenCModule ------------------------------------------------------*/

FILE *GeneralIO_OpenCModule(tIdent Ident)
{
  char fileName[FILENAME_MAX];
  FILE *f;

  GetString(Ident, fileName);
  strcat(fileName, CMODULE_EXT);
  f = FileIO_OpenOutput(fileName);
  GeneralIO_CheckOpen(f, fileName);
  return f;
}

/*------- OpenCrossreferenceFile -------------------------------------------*/

FILE *GeneralIO_OpenCrossreferenceFile(tIdent Ident)
{
  char fileName[FILENAME_MAX];
  FILE *f;

  GetString(Ident, fileName);
  strcat(fileName, CROSSFILE_EXT);
  f = FileIO_OpenOutput(fileName);
  GeneralIO_CheckOpen(f, fileName);
  return f;
}

/*------- OpenMPLHeader ----------------------------------------------------*/

FILE *GeneralIO_OpenMPLHeader(tIdent Ident)
{
  char fileName[FILENAME_MAX];
  FILE *f;

  GetString(Ident, fileName);
  strcat(fileName, MPLHEADER_EXT);
  f = FileIO_OpenOutput(fileName);
  GeneralIO_CheckOpen(f, fileName);
  return f;
}

/*------- OpenMPLModule ----------------------------------------------------*/

FILE *GeneralIO_OpenMPLModule(tIdent Ident)
{
  char fileName[FILENAME_MAX];
  FILE *f;

  GetString(Ident, fileName);
  strcat(fileName, MPLMODULE_EXT);
  f = FileIO_OpenOutput(fileName);
  GeneralIO_CheckOpen(f, fileName);
  return f;
}

void GeneralIO_Init(void)
{
  if (!isInitialised) {
    isInitialised = TRUE;
    PathTabSize = INITIALSIZE;
    MakeArray((char **) &PathTab, &PathTabSize, sizeof(char *));
    ActPathCount = 0;
    AddEnvPathes(getenv(DEFMODPATH));
    FileIO_Init();
  } /* endif */
}

