/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * Name  :  Values.c
 * Author:  lampke, keller
 * Date  :  Wed Jan  8 14:40:19 1997
 *
 * Description:
 * ------------
 * Evaluate expressions, e.g. index boundaries.
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>			  /* Linux finds atof() here */
#include <math.h>
#include <ctype.h>
#include "BigSet.h"
#include "Tree.h"
#include "Symbols.h"
#include "Types.h"
#include "Values.h"

static BOOLEAN isInitialised = FALSE;

/*======================================================================*/

/*
 * local variables
 */
static tEnv env;

/*======================================================================*/

/*
 * local function prototypes
 */
static BOOLEAN Evaluate( tTree tree, tValue *value, tValue *multivalue );

/*======================================================================*/

/*
 * local function implementation
 */
/*****************************************************************************
 *  Synopsis:       BOOLEAN Evaluate( tTree tree, tValue *value, 
 *                                                   tValue *multivalue )
 *
 *  Parameters:     tree:       expression tree
 *                  value:      result value
 *                  multivalue: result value of righthand side expression of
 *                              multiple expressions (dummy elsewhere)
 *                            
 *  Result:	    TRUE:  expression is constant, value is valid
 *                  FALSE: expression is not constant, value undefined
 *
 *  Description:
 *  ------------
 *  Evaluate the constant expression given by tree. The result is returned
 *  in value if computable. Then also TRUE is returned. If the expression
 *  contains values that can not be determined (e.g. variables, etc.), FALSE
 *  is returned. The multivalue is only relevant for multiple expressions. 
 *****************************************************************************/

static BOOLEAN Evaluate( tTree tree, tValue *value, tValue *multivalue )
{
   tValue   value1, value2, multivalue1, multivalue2;
   tObject  object1, object2;
   tType    type;
   char     string1[1024], string2[1024];
   USHORT   i;

   if ( tree == NoTree ) {
      /* stop recursion */
      return TRUE;
   }

   switch ( tree->Kind ) {


/*------ identifier without module qualifier */

   case kQualid0:
      tree->Qualid0.Object = Symbols_Identify( tree->Qualid0.Ident, env );
      switch ( tree->Qualid0.Object->Kind ) {
      case kSymConst:			  /* take value */
         *value = tree->Qualid0.Object->SymConst.Value;
         return TRUE;
      case kSymEnumLiteral:		  /* take pointer to literals */
         value->Kind = eEnum;
         value->value.enumValue = tree->Qualid0.Object;
         return TRUE;
      case kSymTypeDecl:
      case kSymOpaque:			  /* take type itself */
         value->Kind = eType;
         value->value.typeValue = Symbols_GetType( tree->Qualid0.Object );
         return TRUE;
      case kSymStdProc:			  /* take id of procedure */
         value->Kind = eStdProc;
         value->value.stdProcValue = tree->Qualid0.Object->SymStdProc.StdProc;
         return TRUE;
      default:				  /* Error: Expression not constant */
         value->Kind = eNoValue;
         value->value.error.code = xxExprNotConst;
         value->value.error.pos  = tree->Qualid0.Pos;
         value->value.error.info.ident = tree->Qualid0.Ident;
      }
      break;


/*------ identifier with module or record qualifier */

   case kQualid1:
      /* get the value of the next Qualid (=Qualid1 or Qualid0) */
      if ( Evaluate( tree->Qualid1.Qualid, value, multivalue ) ) {
	 /* lookup the id in the environment of the other module */
         tree->Qualid1.Object =
            Symbols_Identify2( tree->Qualid1.Ident,
               Symbols_GetExport2( tree->Qualid1.Qualid->Qualid.Object ) );
	 /* get the value like in Qualid0 */
         switch ( tree->Qualid1.Object->Kind ) {
         case kSymConst:
            *value = tree->Qualid1.Object->SymConst.Value;
            return TRUE;
         case kSymEnumLiteral:
            value->Kind = eEnum;
            value->value.enumValue = tree->Qualid1.Object;
            return TRUE;
         case kSymTypeDecl:
         case kSymOpaque:
            value->Kind = eType;
            value->value.typeValue = Symbols_GetType( tree->Qualid1.Object );
            return TRUE;
         case kSymStdProc:
            value->Kind = eStdProc;
            value->value.stdProcValue = tree->Qualid1.Object->SymStdProc.StdProc;
            return TRUE;
         default:
            value->Kind = eNoValue;
            value->value.error.code = xxExprNotConst;
            value->value.error.pos  = tree->Qualid1.Pos;
            value->value.error.info.ident = tree->Qualid1.Ident;
         }
      }
      break;


/*------ expression with multiple comparisons */

   case kMultiple:
      if ( Evaluate( tree->Multiple.Mop, &value1, &multivalue1  ) &&
           /* Because 'Mop' can't be a multiple expression, parameter
              'multivalue1' is of no interest. 'multivalue2' contains the
              AND-result of the right multiple expressions, value1 the left
              operand and value2 the right operand. */
           Evaluate( tree->Multiple.Multiples, &multivalue2, &value2 ) ) {
         *multivalue = value1;
         /* copies the actual left operand to right operand of former
            expression */
         switch ( tree->Multiple.Operator ) {
         case OpEqual:
            value->Kind = eBoolean;
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->value.booleanValue =
                  ( value1.value.cardinalValue == value2.value.cardinalValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eInteger && value2.Kind == eInteger ) {
               value->value.booleanValue =
                  ( value1.value.integerValue == value2.value.integerValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eCardinal && value2.Kind == eInteger ) {
               if ( value1.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( (LONG)(value1.value.cardinalValue) ==
                        value2.value.integerValue &&
                        multivalue2.value.booleanValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eInteger && value2.Kind == eCardinal ) {
               if ( value2.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( value1.value.integerValue ==
                       (LONG)(value2.value.cardinalValue) &&
                       multivalue2.value.booleanValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Rop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->value.booleanValue =
                  ( value1.value.realValue == value2.value.realValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
               value->value.booleanValue =
                  ( value1.value.booleanValue == value2.value.booleanValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                        (value2.Kind == eChar || value2.Kind == eStringChar) ) {
               value->value.booleanValue =
                  ( value1.value.charValue == value2.value.charValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eString && value2.Kind == eString ) {
               StGetString( value1.value.stringValue, string1 );
               StGetString( value2.value.stringValue, string2 );
               value->value.booleanValue =
                  ( strcmp( string1, string2 ) == 0 &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->value.booleanValue =
                  ( BITSET_IsEqual(value1.value.bitsetValue,
                      value2.value.bitsetValue) &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eEnum && value2.Kind == eEnum ) {
               object1 = value1.value.enumValue;
               object2 = value2.value.enumValue;
               value->value.booleanValue =
                  ( object1->SymEnumLiteral.Index ==
                    object2->SymEnumLiteral.Index &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eNilType && value2.Kind == eNilType ) {
               value->value.booleanValue = multivalue2.value.booleanValue;
               return TRUE;
            } /* endif */
            break;
         case OpLess:
            value->Kind = eBoolean;
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->value.booleanValue =
                  ( value1.value.cardinalValue < value2.value.cardinalValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eInteger && value2.Kind == eInteger ) {
               value->value.booleanValue =
                  ( value1.value.integerValue < value2.value.integerValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eCardinal && value2.Kind == eInteger ) {
               if ( value1.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( (LONG)(value1.value.cardinalValue) <
                       value2.value.integerValue &&
                       multivalue2.value.booleanValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eInteger && value2.Kind == eCardinal ) {
               if ( value2.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( value1.value.integerValue <
                       (LONG)(value2.value.cardinalValue) &&
                       multivalue2.value.booleanValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Rop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->value.booleanValue =
                  ( value1.value.realValue < value2.value.realValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
               value->value.booleanValue =
                  ( value1.value.booleanValue < value2.value.booleanValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                        (value2.Kind == eChar || value2.Kind == eStringChar) ) {
               value->value.booleanValue =
                  ( value1.value.charValue < value2.value.charValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eString && value2.Kind == eString ) {
               StGetString( value1.value.stringValue, string1 );
               StGetString( value2.value.stringValue, string2 );
               value->value.booleanValue =
                 ( strcmp( string1, string2 ) < 0 &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->value.booleanValue =
                  ( BITSET_IsStrictSubset(value1.value.bitsetValue,
                    value2.value.bitsetValue) &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eEnum && value2.Kind == eEnum ) {
               object1 = value1.value.enumValue;
               object2 = value2.value.enumValue;
               value->value.booleanValue =
                  ( object1->SymEnumLiteral.Index <
                    object2->SymEnumLiteral.Index &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eNilType && value2.Kind == eNilType ) {
               value->value.booleanValue =  multivalue2.value.booleanValue;
               return TRUE;
            } /* endif */
            break;
         case OpLessEqual:
            value->Kind = eBoolean;
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->value.booleanValue =
                  ( value1.value.cardinalValue <= value2.value.cardinalValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eInteger && value2.Kind == eInteger ) {
               value->value.booleanValue =
                  ( value1.value.integerValue <= value2.value.integerValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eCardinal && value2.Kind == eInteger ) {
               if ( value1.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( (LONG)(value1.value.cardinalValue) <=
                       value2.value.integerValue &&
                       multivalue2.value.booleanValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eInteger && value2.Kind == eCardinal ) {
               if ( value2.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( value1.value.integerValue <=
                       (LONG)(value2.value.cardinalValue) &&
                       multivalue2.value.booleanValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Rop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->value.booleanValue =
                  ( value1.value.realValue <= value2.value.realValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
               value->value.booleanValue =
                  ( value1.value.booleanValue <= value2.value.booleanValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                        (value2.Kind == eChar || value2.Kind == eStringChar) ) {
               value->value.booleanValue =
                  ( value1.value.charValue <= value2.value.charValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eString && value2.Kind == eString ) {
               StGetString( value1.value.stringValue, string1 );
               StGetString( value2.value.stringValue, string2 );
               value->value.booleanValue =
                  ( strcmp( string1, string2 ) <= 0 &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->value.booleanValue =
                  ( BITSET_IsSubset(value1.value.bitsetValue,
                    value2.value.bitsetValue) &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eEnum && value2.Kind == eEnum ) {
               object1 = value1.value.enumValue;
               object2 = value2.value.enumValue;
               value->value.booleanValue =
                  ( object1->SymEnumLiteral.Index <=
                    object2->SymEnumLiteral.Index &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eNilType && value2.Kind == eNilType ) {
               value->value.booleanValue =  multivalue2.value.booleanValue;
               return TRUE;
            } /* endif */
            break;
         case OpGreater:
            value->Kind = eBoolean;
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->value.booleanValue =
                  ( value1.value.cardinalValue > value2.value.cardinalValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eInteger && value2.Kind == eInteger ) {
               value->value.booleanValue =
                  ( value1.value.integerValue > value2.value.integerValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eCardinal && value2.Kind == eInteger ) {
               if ( value1.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( (LONG)(value1.value.cardinalValue) >
                       value2.value.integerValue &&
                       multivalue2.value.booleanValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eInteger && value2.Kind == eCardinal ) {
               if ( value2.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( value1.value.integerValue >
                       (LONG)(value2.value.cardinalValue) &&
                       multivalue2.value.booleanValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Rop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->value.booleanValue =
                  ( value1.value.realValue > value2.value.realValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
               value->value.booleanValue =
                  ( value1.value.booleanValue > value2.value.booleanValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                        (value2.Kind == eChar || value2.Kind == eStringChar) ) {
               value->value.booleanValue =
                  ( value1.value.charValue > value2.value.charValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eString && value2.Kind == eString ) {
               StGetString( value1.value.stringValue, string1 );
               StGetString( value2.value.stringValue, string2 );
               value->value.booleanValue =
                  ( strcmp( string1, string2 ) > 0 &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->value.booleanValue =
                  ( BITSET_IsStrictSubset(value2.value.bitsetValue,
                    value1.value.bitsetValue) &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eEnum && value2.Kind == eEnum ) {
               object1 = value1.value.enumValue;
               object2 = value2.value.enumValue;
               value->value.booleanValue =
                  ( object1->SymEnumLiteral.Index >
                    object2->SymEnumLiteral.Index &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eNilType && value2.Kind == eNilType ) {
               value->value.booleanValue = multivalue2.value.booleanValue;
               return TRUE;
            } /* endif */
            break;
         case OpGreaterEqual:
            value->Kind = eBoolean;
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->value.booleanValue =
                  ( value1.value.cardinalValue >= value2.value.cardinalValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eInteger && value2.Kind == eInteger ) {
               value->value.booleanValue =
                  ( value1.value.integerValue >= value2.value.integerValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eCardinal && value2.Kind == eInteger ) {
               if ( value1.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( (LONG)(value1.value.cardinalValue) >=
                       value2.value.integerValue &&
                       multivalue2.value.booleanValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eInteger && value2.Kind == eCardinal ) {
               if ( value2.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( value1.value.integerValue >=
                       (LONG)(value2.value.cardinalValue) &&
                       multivalue2.value.booleanValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Rop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->value.booleanValue =
                  ( value1.value.realValue >= value2.value.realValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
               value->value.booleanValue =
                  ( value1.value.booleanValue >= value2.value.booleanValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                        (value2.Kind == eChar || value2.Kind == eStringChar) ) {
               value->value.booleanValue =
                  ( value1.value.charValue >= value2.value.charValue &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eString && value2.Kind == eString ) {
               StGetString( value1.value.stringValue, string1 );
               StGetString( value2.value.stringValue, string2 );
               value->value.booleanValue =
                  ( strcmp( string1, string2 ) >= 0 &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->value.booleanValue =
                  ( BITSET_IsSubset(value2.value.bitsetValue,
                    value1.value.bitsetValue) &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eEnum && value2.Kind == eEnum ) {
               object1 = value1.value.enumValue;
               object2 = value2.value.enumValue;
               value->value.booleanValue =
                  ( object1->SymEnumLiteral.Index >=
                    object2->SymEnumLiteral.Index &&
                    multivalue2.value.booleanValue );
               return TRUE;
            } else if ( value1.Kind == eNilType && value2.Kind == eNilType ) {
               value->value.booleanValue = multivalue2.value.booleanValue;
               return TRUE;
            } /* endif */
            break;
         } /* endswitch */
         value->Kind = eNoValue;
         value->value.error.code = xxIllegalOperator;
         value->value.error.pos  = tree->Multiple.Pos;
         multivalue->Kind = eNoValue;
         multivalue->value.error.code = xxIllegalOperator;
         multivalue->value.error.pos  = tree->Multiple.Pos;
         return FALSE;
      } /* endif */
      /* error: Expression not constant */
      if ( value1.Kind == eNoValue ) {
         *value = value1;
         *multivalue = value1;
      } else {
         *value = value2;
         *multivalue = value2;
      } /* endif */
      return FALSE;
      break;


/*------*/

   case kLastMultiple:
      if ( Evaluate( tree->LastMultiple.Mop, multivalue, value ) ) {
         value->Kind = eBoolean;
         value->value.booleanValue = TRUE;
         return TRUE;
      } else {
         value->Kind = eBoolean;
         value->value.booleanValue = FALSE;
         return FALSE;
      }
      /* 'multivalue' contains the result of the Mop-expression, value
         contains TRUE to initialise the AND-operation for the multiple
         expression. */
      break;


/*------ binary expression (2 operands) */

   case kBinary:
      /* evaluate left and right operand */
      if ( Evaluate( tree->Binary.Lop, &value1, multivalue ) &&
           Evaluate( tree->Binary.Rop, &value2, multivalue ) ) {
	 /* evaluate combination */
         switch ( tree->Binary.Operator ) {

	 /* "=" ----------*/
         case OpEqual:
            value->Kind = eBoolean;	  /* result is BOOLEAN */
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
	       /* compare two CARDINALs */
               value->value.booleanValue =
                  ( value1.value.cardinalValue == value2.value.cardinalValue );
               return TRUE;
            } else if ( value1.Kind == eInteger && value2.Kind == eInteger ) {
	       /* compare two INTEGERs */
               value->value.booleanValue =
                  ( value1.value.integerValue == value2.value.integerValue );
               return TRUE;
            } else if ( value1.Kind == eCardinal && value2.Kind == eInteger ) {
	       /* compare CARDINAL with INTEGER only in cases, when */
	       /* CARDINAL<MAX(INTEGER); then compare signed. */
               if ( value1.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( (LONG)(value1.value.cardinalValue) == value2.value.integerValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eInteger && value2.Kind == eCardinal ) {
	       /* compare INTEGER with CARDINAL only in cases, when */
	       /* CARDINAL<MAX(INTEGER); then compare signed. */
               if ( value2.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( value1.value.integerValue == (LONG)(value2.value.cardinalValue) );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Rop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
	       /* compare two REALs */
               value->value.booleanValue =
                  ( value1.value.realValue == value2.value.realValue );
               return TRUE;
            } else if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
	       /* compare two BOOLEANs */
               value->value.booleanValue =
                  ( value1.value.booleanValue == value2.value.booleanValue );
               return TRUE;
            } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                        (value2.Kind == eChar || value2.Kind == eStringChar) ) {
	       /* compare two CHARs */
               value->value.booleanValue =
                  ( value1.value.charValue == value2.value.charValue );
               return TRUE;
            } else if ( value1.Kind == eString && value2.Kind == eString ) {
	       /* compare two strings */
               StGetString( value1.value.stringValue, string1 );
               StGetString( value2.value.stringValue, string2 );
               if ( strcmp( string1, string2 ) == 0 ) {
                  value->value.booleanValue = TRUE;
               } else {
                  value->value.booleanValue = FALSE;
               } /* endif */
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
	       /* compare two BITSETs */
               value->value.booleanValue =
                  BITSET_IsEqual(value1.value.bitsetValue, value2.value.bitsetValue);
               return TRUE;
            } else if ( value1.Kind == eEnum && value2.Kind == eEnum ) {
	       /* compare two enumeration constants */
               object1 = value1.value.enumValue;
               object2 = value2.value.enumValue;
               value->value.booleanValue =
                  ( object1->SymEnumLiteral.Index == object2->SymEnumLiteral.Index );
               return TRUE;
            } else if ( value1.Kind == eNilType && value2.Kind == eNilType ) {
	       /* NIL=NIL is always true */
               value->value.booleanValue = TRUE;
               return TRUE;
            } /* endif */
            break;

	 /* "<>", "#" ---------- */
         case OpNotEqual:
            value->Kind = eBoolean;
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->value.booleanValue =
                  ( value1.value.cardinalValue != value2.value.cardinalValue );
               return TRUE;
            } else if ( value1.Kind == eInteger && value2.Kind == eInteger ) {
               value->value.booleanValue =
                  ( value1.value.integerValue != value2.value.integerValue );
               return TRUE;
            } else if ( value1.Kind == eCardinal && value2.Kind == eInteger ) {
               if ( value1.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( (LONG)(value1.value.cardinalValue) != value2.value.integerValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eInteger && value2.Kind == eCardinal ) {
               if ( value2.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( value1.value.integerValue != (LONG)(value2.value.cardinalValue) );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Rop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->value.booleanValue =
                  ( value1.value.realValue != value2.value.realValue );
               return TRUE;
            } else if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
               value->value.booleanValue =
                  ( value1.value.booleanValue != value2.value.booleanValue );
               return TRUE;
            } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                        (value2.Kind == eChar || value2.Kind == eStringChar) ) {
               value->value.booleanValue =
                  ( value1.value.charValue != value2.value.charValue );
               return TRUE;
            } else if ( value1.Kind == eString && value2.Kind == eString ) {
               StGetString( value1.value.stringValue, string1 );
               StGetString( value2.value.stringValue, string2 );
               if ( strcmp( string1, string2 ) != 0 ) {
                  value->value.booleanValue = TRUE;
               } else {
                  value->value.booleanValue = FALSE;
               } /* endif */
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->value.booleanValue =
                  !BITSET_IsEqual(value1.value.bitsetValue, value2.value.bitsetValue);
               return TRUE;
            } else if ( value1.Kind == eEnum && value2.Kind == eEnum ) {
               object1 = value1.value.enumValue;
               object2 = value2.value.enumValue;
               value->value.booleanValue =
                  ( object1->SymEnumLiteral.Index != object2->SymEnumLiteral.Index );
               return TRUE;
            } else if ( value1.Kind == eNilType && value2.Kind == eNilType ) {
               value->value.booleanValue = FALSE;
               return TRUE;
            } /* endif */
            break;

	 /* "<" ----------*/
         case OpLess:
            value->Kind = eBoolean;
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->value.booleanValue =
                  ( value1.value.cardinalValue < value2.value.cardinalValue );
               return TRUE;
            } else if ( value1.Kind == eInteger && value2.Kind == eInteger ) {
               value->value.booleanValue =
                  ( value1.value.integerValue < value2.value.integerValue );
               return TRUE;
            } else if ( value1.Kind == eCardinal && value2.Kind == eInteger ) {
               if ( value1.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( (LONG)(value1.value.cardinalValue) < value2.value.integerValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eInteger && value2.Kind == eCardinal ) {
               if ( value2.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( value1.value.integerValue < (LONG)(value2.value.cardinalValue) );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Rop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->value.booleanValue =
                  ( value1.value.realValue < value2.value.realValue );
               return TRUE;
            } else if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
               value->value.booleanValue =
                  ( value1.value.booleanValue < value2.value.booleanValue );
               return TRUE;
            } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                        (value2.Kind == eChar || value2.Kind == eStringChar) ) {
               value->value.booleanValue =
                  ( value1.value.charValue < value2.value.charValue );
               return TRUE;
            } else if ( value1.Kind == eString && value2.Kind == eString ) {
               StGetString( value1.value.stringValue, string1 );
               StGetString( value2.value.stringValue, string2 );
               if ( strcmp( string1, string2 ) < 0 ) {
                  value->value.booleanValue = TRUE;
               } else {
                  value->value.booleanValue = FALSE;
               } /* endif */
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->value.booleanValue =
                  BITSET_IsStrictSubset(value1.value.bitsetValue, value2.value.bitsetValue);
               return TRUE;
            } else if ( value1.Kind == eEnum && value2.Kind == eEnum ) {
               object1 = value1.value.enumValue;
               object2 = value2.value.enumValue;
               value->value.booleanValue =
                  ( object1->SymEnumLiteral.Index < object2->SymEnumLiteral.Index );
               return TRUE;
            } else if ( value1.Kind == eNilType && value2.Kind == eNilType ) {
               value->value.booleanValue = FALSE;
               return TRUE;
            } /* endif */
            break;

	 /* "<=" ----------*/
         case OpLessEqual:
            value->Kind = eBoolean;
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->value.booleanValue =
                  ( value1.value.cardinalValue <= value2.value.cardinalValue );
               return TRUE;
            } else if ( value1.Kind == eInteger && value2.Kind == eInteger ) {
               value->value.booleanValue =
                  ( value1.value.integerValue <= value2.value.integerValue );
               return TRUE;
            } else if ( value1.Kind == eCardinal && value2.Kind == eInteger ) {
               if ( value1.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( (LONG)(value1.value.cardinalValue) <= value2.value.integerValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eInteger && value2.Kind == eCardinal ) {
               if ( value2.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( value1.value.integerValue <= (LONG)(value2.value.cardinalValue) );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Rop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->value.booleanValue =
                  ( value1.value.realValue <= value2.value.realValue );
               return TRUE;
            } else if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
               value->value.booleanValue =
                  ( value1.value.booleanValue <= value2.value.booleanValue );
               return TRUE;
            } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                        (value2.Kind == eChar || value2.Kind == eStringChar) ) {
               value->value.booleanValue =
                  ( value1.value.charValue <= value2.value.charValue );
               return TRUE;
            } else if ( value1.Kind == eString && value2.Kind == eString ) {
               StGetString( value1.value.stringValue, string1 );
               StGetString( value2.value.stringValue, string2 );
               if ( strcmp( string1, string2 ) <= 0 ) {
                  value->value.booleanValue = TRUE;
               } else {
                  value->value.booleanValue = FALSE;
               } /* endif */
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->value.booleanValue =
                  BITSET_IsSubset(value1.value.bitsetValue, value2.value.bitsetValue);
               return TRUE;
            } else if ( value1.Kind == eEnum && value2.Kind == eEnum ) {
               object1 = value1.value.enumValue;
               object2 = value2.value.enumValue;
               value->value.booleanValue =
                  ( object1->SymEnumLiteral.Index <= object2->SymEnumLiteral.Index );
               return TRUE;
            } else if ( value1.Kind == eNilType && value2.Kind == eNilType ) {
               value->value.booleanValue = TRUE;
               return TRUE;
            } /* endif */
            break;

	 /* ">" ----------*/
         case OpGreater:
            value->Kind = eBoolean;
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->value.booleanValue =
                  ( value1.value.cardinalValue > value2.value.cardinalValue );
               return TRUE;
            } else if ( value1.Kind == eInteger && value2.Kind == eInteger ) {
               value->value.booleanValue =
                  ( value1.value.integerValue > value2.value.integerValue );
               return TRUE;
            } else if ( value1.Kind == eCardinal && value2.Kind == eInteger ) {
               if ( value1.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( (LONG)(value1.value.cardinalValue) > value2.value.integerValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eInteger && value2.Kind == eCardinal ) {
               if ( value2.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( value1.value.integerValue > (LONG)(value2.value.cardinalValue) );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Rop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->value.booleanValue =
                  ( value1.value.realValue > value2.value.realValue );
               return TRUE;
            } else if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
               value->value.booleanValue =
                  ( value1.value.booleanValue > value2.value.booleanValue );
               return TRUE;
            } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                        (value2.Kind == eChar || value2.Kind == eStringChar) ) {
               value->value.booleanValue =
                  ( value1.value.charValue > value2.value.charValue );
               return TRUE;
            } else if ( value1.Kind == eString && value2.Kind == eString ) {
               StGetString( value1.value.stringValue, string1 );
               StGetString( value2.value.stringValue, string2 );
               if ( strcmp( string1, string2 ) > 0 ) {
                  value->value.booleanValue = TRUE;
               } else {
                  value->value.booleanValue = FALSE;
               } /* endif */
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->value.booleanValue =
                  BITSET_IsStrictSubset(value2.value.bitsetValue, value1.value.bitsetValue);
               return TRUE;
            } else if ( value1.Kind == eEnum && value2.Kind == eEnum ) {
               object1 = value1.value.enumValue;
               object2 = value2.value.enumValue;
               value->value.booleanValue =
                  ( object1->SymEnumLiteral.Index > object2->SymEnumLiteral.Index );
               return TRUE;
            } else if ( value1.Kind == eNilType && value2.Kind == eNilType ) {
               value->value.booleanValue = FALSE;
               return TRUE;
            } /* endif */
            break;

	 /* ">=" ----------*/
         case OpGreaterEqual:
            value->Kind = eBoolean;
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->value.booleanValue =
                  ( value1.value.cardinalValue >= value2.value.cardinalValue );
               return TRUE;
            } else if ( value1.Kind == eInteger && value2.Kind == eInteger ) {
               value->value.booleanValue =
                  ( value1.value.integerValue >= value2.value.integerValue );
               return TRUE;
            } else if ( value1.Kind == eCardinal && value2.Kind == eInteger ) {
               if ( value1.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( (LONG)(value1.value.cardinalValue) >= value2.value.integerValue );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eInteger && value2.Kind == eCardinal ) {
               if ( value2.value.cardinalValue <= Types_MaxInteger ) {
                  value->value.booleanValue =
                     ( value1.value.integerValue >= (LONG)(value2.value.cardinalValue) );
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Rop);
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->value.booleanValue =
                  ( value1.value.realValue >= value2.value.realValue );
               return TRUE;
            } else if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
               value->value.booleanValue =
                  ( value1.value.booleanValue >= value2.value.booleanValue );
               return TRUE;
            } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                        (value2.Kind == eChar || value2.Kind == eStringChar) ) {
               value->value.booleanValue =
                  ( value1.value.charValue >= value2.value.charValue );
               return TRUE;
            } else if ( value1.Kind == eString && value2.Kind == eString ) {
               StGetString( value1.value.stringValue, string1 );
               StGetString( value2.value.stringValue, string2 );
               if ( strcmp( string1, string2 ) >= 0 ) {
                  value->value.booleanValue = TRUE;
               } else {
                  value->value.booleanValue = FALSE;
               } /* endif */
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->value.booleanValue =
                  BITSET_IsSubset(value2.value.bitsetValue, value1.value.bitsetValue);
               return TRUE;
            } else if ( value1.Kind == eEnum && value2.Kind == eEnum ) {
               object1 = value1.value.enumValue;
               object2 = value2.value.enumValue;
               value->value.booleanValue =
                  ( object1->SymEnumLiteral.Index >= object2->SymEnumLiteral.Index );
               return TRUE;
            } else if ( value1.Kind == eNilType && value2.Kind == eNilType ) {
               value->value.booleanValue = TRUE;
               return TRUE;
            } /* endif */
            break;

	 /* "IN" ----------*/
         case OpIn:
            value->Kind = eBoolean;
            if ( value1.Kind == eCardinal && value2.Kind == eBitset ) {
               if ( value1.value.cardinalValue <= Types_MaxBitset ) {
                  value->value.booleanValue =
                     BITSET_IsElement( value1.value.cardinalValue, value2.value.bitsetValue );
                  return TRUE;
               } else {
                  /* error: BITSET overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxSetOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeBITSET;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eInteger && value2.Kind == eBitset ) {
               if ( value1.value.integerValue >= 0 &&
                    value1.value.integerValue <= Types_MaxBitset ) {
                  value->value.booleanValue =
                     BITSET_IsElement( value1.value.integerValue, value2.value.bitsetValue );
                  return TRUE;
               } else {
                  /* error: BITSET overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxSetOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeBITSET;
                  return FALSE;
               } /* endif */
            } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                        value2.Kind == eBitset ) {
               if ( value1.value.charValue <= Types_MaxBitset ) {
                  value->value.booleanValue =
                     BITSET_IsElement( value1.value.charValue, value2.value.bitsetValue );
                  return TRUE;
               } else {
                  /* error: BITSET overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxSetOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Lop);
                  value->value.error.info.type = TypeBITSET;
                  return FALSE;
               } /* endif */
            } else if ( value1.Kind == eEnum && value2.Kind == eBitset ) {
               object1 = value1.value.enumValue;
               if ( object1->SymEnumLiteral.Index <= Types_MaxBitset ) {
                  value->value.booleanValue =
                     BITSET_IsElement( object1->SymEnumLiteral.Index, value2.value.bitsetValue );
                  return TRUE;
               } else {
                  /* error: BITSET overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxSetOverFlow;
                  value->value.error.pos  = Tree_LeftmostPos(tree->Binary.Rop);
                  value->value.error.info.type = TypeBITSET;
                  return FALSE;
               } /* endif */
            } /* endif */
            break;

	 /* "+" ----------*/
         case OpPlus:
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->Kind = eCardinal;
               value->value.cardinalValue =
                  value1.value.cardinalValue + value2.value.cardinalValue;
               return TRUE;
            } else if ( (value1.Kind == eInteger || value1.Kind == eCardinal) &&
                        (value2.Kind == eInteger || value2.Kind == eCardinal) ) {
               value->Kind = eInteger;
               value->value.integerValue =
                  value1.value.integerValue + value2.value.integerValue;
               return TRUE;
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->Kind = eReal;
               value->value.realValue =
                  value1.value.realValue + value2.value.realValue;
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->Kind = eBitset;
               value->value.bitsetValue =
                  BITSET_Union( value1.value.bitsetValue, value2.value.bitsetValue );
               return TRUE;
            } /* endif */
            break;

	 /* "-" ----------*/
         case OpMinus:
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->Kind = eCardinal;
               if ( value1.value.cardinalValue >= value2.value.cardinalValue ) {
                  value->value.cardinalValue =
                     value1.value.cardinalValue - value2.value.cardinalValue;
                  return TRUE;
               } else {
                  /* error: type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = tree->Binary.Pos;
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( (value1.Kind == eInteger || value1.Kind == eCardinal) &&
                        (value2.Kind == eInteger || value2.Kind == eCardinal) ) {
               value->Kind = eInteger;
               value->value.integerValue =
                  value1.value.integerValue - value2.value.integerValue;
               return TRUE;
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->Kind = eReal;
               value->value.realValue =
                  value1.value.realValue - value2.value.realValue;
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->Kind = eBitset;
               value->value.bitsetValue =
                  BITSET_Diff( value1.value.bitsetValue, value2.value.bitsetValue );
               return TRUE;
            } /* endif */
            break;

	 /* "OR" ----------*/
         case OpOr:
            if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
               value->Kind = eBoolean;
               value->value.booleanValue =
                  value1.value.booleanValue || value2.value.booleanValue;
               return TRUE;
            } /* endif */
            break;

	 /* "*" ----------*/
         case OpTimes:
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->Kind = eCardinal;
               value->value.cardinalValue =
                  value1.value.cardinalValue * value2.value.cardinalValue;
               return TRUE;
            } else if ( (value1.Kind == eInteger || value1.Kind == eCardinal) &&
                        (value2.Kind == eInteger || value2.Kind == eCardinal) ) {
               value->Kind = eInteger;
               value->value.integerValue =
                  value1.value.integerValue * value2.value.integerValue;
               return TRUE;
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->Kind = eReal;
               value->value.realValue =
                  value1.value.realValue * value2.value.realValue;
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->Kind = eBitset;
               value->value.bitsetValue =
                  BITSET_Intersect( value1.value.bitsetValue, value2.value.bitsetValue );
               return TRUE;
            } /* endif */
            break;

	 /* "/" ----------*/
         case OpDivide:
            if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->Kind =  eReal;
               value->value.realValue =
                  value1.value.realValue / value2.value.realValue;
               return TRUE;
            } else if ( value1.Kind == eBitset && value2.Kind == eBitset ) {
               value->Kind = eBitset;
               value->value.bitsetValue =
                  BITSET_SymDiff( value1.value.bitsetValue, value2.value.bitsetValue );
               return TRUE;
            } /* endif */
            break;

	 /* "DIV" ----------*/
         case OpDiv:
            if ( (value1.Kind == eInteger || value1.Kind == eCardinal) &&
                 (value2.Kind == eInteger || value2.Kind == eCardinal) ) {
               if ( value2.value.integerValue != 0L ) {
                  value->Kind = eInteger;
                  value->value.integerValue =
                     value1.value.integerValue / value2.value.integerValue;
                  return TRUE;
               } else {
                  /* error: Division by Zero */
                  value->Kind = eNoValue;
                  value->value.error.code = xxDivisionByZero;
                  value->value.error.pos  = tree->Binary.Pos;
                  return FALSE;
               } /* endif */
            } /* endif */
            break;

	 /* "MOD" ----------*/
         case OpMod:
            if ( (value1.Kind == eInteger || value1.Kind == eCardinal) &&
                 (value2.Kind == eInteger || value2.Kind == eCardinal) ) {
               if ( value2.value.integerValue != 0L ) {
                  value->Kind = eInteger;
                  value->value.integerValue =
                     value1.value.integerValue % value2.value.integerValue;
                  /* because the C- and Parallaxis-MOD are different for
                   * negative values, a correction for negative results
                   * must be done
                   */
                  if (value->value.integerValue < 0) {
                    if (value2.Kind == eCardinal) { /* always positive */
                      value->value.integerValue += value2.value.cardinalValue;
                    } else if (value2.value.integerValue > 0) { /* positive */
                      value->value.integerValue += value2.value.integerValue;
                    } else { /* negative */
                      value->value.integerValue -= value2.value.integerValue;
                    }
                  }
                  return TRUE;
               } else {
                  /* error: Division by Zero */
                  value->Kind = eNoValue;
                  value->value.error.code = xxDivisionByZero;
                  value->value.error.pos  = tree->Binary.Pos;
                  return FALSE;
               } /* endif */
            } /* endif */
            break;

	 /* "AND", "&" ----------*/
         case OpAnd:
            if ( value1.Kind == eBoolean && value2.Kind == eBoolean ) {
               value->Kind = eBoolean;
               value->value.booleanValue =
                  value1.value.booleanValue && value2.value.booleanValue;
               return TRUE;
            } /* endif */
            break;

	 /* "**" ----------*/
         case OpPower:
            if ( value1.Kind == eCardinal && value2.Kind == eCardinal ) {
               value->Kind = eCardinal;
               value->value.cardinalValue = (ULONG)
                  pow( (REAL)value1.value.cardinalValue,
                       (REAL)value2.value.cardinalValue );
               return TRUE;
            } else if ( (value1.Kind == eInteger || value2.Kind == eCardinal) &&
                        (value2.Kind == eInteger || value2.Kind == eCardinal) ) {
               value->Kind = eInteger;
               value->value.integerValue = (LONG)
                  pow( (REAL)value1.value.integerValue,
                       (REAL)value2.value.integerValue );
               return TRUE;
            } else if ( value1.Kind == eReal && value2.Kind == eReal ) {
               value->Kind = eReal;
               value->value.realValue =
                  pow( value1.value.realValue, value2.value.realValue );
               return TRUE;
            } /* endif */
            break;
         } /* endswitch ( tree->Binary.Operator ) */
         value->Kind = eNoValue;
         value->value.error.code = xxIllegalOperator;
         value->value.error.pos  = tree->Binary.Pos;
         return FALSE;
      } /* endif */
      /* error: Expression not constant */
      if ( value1.Kind == eNoValue ) {
         *value = value1;
      } else {
         *value = value2;
      } /* endif */
      return FALSE;


/*------ Unary expressions (only 1 operand) */

   case kUnary:
      /* Evaluate single operand */
      if ( Evaluate( tree->Unary.Mop, value, multivalue ) ) {
         switch ( tree->Unary.Operator ) {

	 /* "+" (positive sign) ---------- */
         case OpPlus:
            return TRUE;

	 /* "-" (negative sign) ---------- */
         case OpMinus:
            if ( value->Kind == eCardinal ) {
               if ( value->value.cardinalValue <= Types_MaxInteger ) {
                  value->Kind = eInteger;
                  value->value.integerValue = - value->value.cardinalValue;
                  return TRUE;
               } else {
                  /* error: Type overflow */
                  value->Kind = eNoValue;
                  value->value.error.code = xxTypeOverFlow;
                  value->value.error.pos  = tree->Unary.Pos;
                  value->value.error.info.type = TypeCARDINAL;
                  return FALSE;
               } /* endif */
            } else if ( value->Kind == eInteger ) {
               value->value.integerValue = - value->value.integerValue;
               return TRUE;
            } else if ( value->Kind == eReal ) {
               value->value.realValue = - value->value.realValue;
               return TRUE;
            } /* endif */
            break;

	 /* NOT ---------- */
         case OpNot:
            if ( value->Kind == eBoolean ) {
               value->value.booleanValue = !value->value.booleanValue;
               return TRUE;
            } /* endif */
            break;
         } /* endswitch */
         value->Kind = eNoValue;
         value->value.error.code = xxIllegalOperator;
         value->value.error.pos  = tree->Binary.Pos;
      } /* endif */
      /* error: Expression not constant */
      return FALSE;


/*------ Integer constant */

   case kIntConst:
      value->Kind = eCardinal;
      value->value.cardinalValue = tree->IntConst.IntVal;
      return TRUE;


/*------ Real constant */

   case kRealConst:
      StGetString( tree->RealConst.RealVal, string1 );
      value->Kind = eReal;
      value->value.realValue = atof( string1 );
      return TRUE;


/*------ String constant */

   case kStringConst:
      StGetString( tree->StringConst.StringVal, string1 );
      if ( strlen( string1 ) == 1 ) {
         value->Kind = eStringChar;
         value->value.charValue = *string1;
      } else {
         value->Kind = eString;
         value->value.stringValue = tree->StringConst.StringVal;
      } /* endif */
      return TRUE;


/*------ Character constant */

   case kCharConst:
      value->Kind = eChar;
      value->value.charValue = tree->CharConst.CharVal;
      return TRUE;


/*------ Function call */

   case kFuncCall:
      if ( Evaluate( tree->FuncCall.Designator, value, multivalue ) &&
           value->Kind == eStdProc ) {
	 /* only standard procedures can be evaluated */
         if ( Evaluate( tree->FuncCall.Actuals, &value1, multivalue ) ) {
            switch ( value->value.stdProcValue ) {

	    /* ABS() ---------- */
            case eSystemAbs:
	       if ( value1.Kind == eCardinal ) {
	          value->value.cardinalValue = value1.value.cardinalValue;
		  return TRUE;		  /* is already positive */
	       } else if ( value1.Kind == eInteger ) {
                  value->Kind = eInteger;
                  value->value.integerValue = labs( value1.value.integerValue );
		  /* value->Kind = eCardinal; 
                  value->value.cardinalValue = 
		     labs(value1.value.integerValue ); */
                  return TRUE;
               } else if ( value1.Kind == eReal ) {
                  value->Kind = eReal;
                  value->value.realValue = fabs( value1.value.realValue );
                  return TRUE;
               } /* endif */
               break;

	    /* CAP() ---------- */
            case eSystemCap:
               if ( value1.Kind == eChar ||
                    value1.Kind == eStringChar ) {
                  value->Kind = eChar;
                  value->value.charValue = toupper( value1.value.charValue );
                  return TRUE;
               }
               break;

	    /* CHR() ---------- */
            case eSystemChr:
               if ( value1.Kind == eInteger ||
                    value1.Kind == eCardinal ) {
                  if ( (value1.value.integerValue >= 0) &&
                       (value1.value.integerValue <= Types_MaxChar) ) {
                     value->Kind = eChar;
                     value->value.charValue = value1.value.integerValue;
                     return TRUE;
                  } else {
                     /* error: type overflow */
                     value->Kind = eNoValue;
                     value->value.error.code = xxTypeOverFlow;
                     value->value.error.pos  = Tree_LeftmostPos(tree->FuncCall.Designator);
                     value->value.error.info.type = TypeCHAR;
                     return FALSE;
                  } /* endif */
               } /* endif */
               break;

	    /* FLOAT() ---------- */
            case eSystemFloat:
               if ( value1.Kind == eCardinal ) {
                  value->Kind = eReal;
                  value->value.realValue = (REAL)value1.value.cardinalValue;
                  return TRUE;
               } else if ( value1.Kind == eInteger ) {
                  value->Kind = eReal;
                  value->value.realValue = (REAL)value1.value.integerValue;
                  return TRUE;
               } /* endif */
               break;

	    /* MIN() ---------- */
            case eSystemMin:
               if ( value1.Kind == eType ) {
                  /* TBC EKA
                   * type constants
                   */
                  switch ( ((tType)value1.value.typeValue)->Kind ) {
                  case kSymInteger:
                     *value = Values_MinInteger;
                     break;
                  case kSymCardinal:
                     *value = Values_Zero;
                     break;
                  case kSymReal:
                     *value = Values_MinReal;
                     break;
                  case kSymBoolean:
                     *value = Values_False;
                     break;
                  case kSymChar:
                     *value = Values_MinChar;
                     break;
                  case kSymEnumeration:
                     value->Kind = eEnum;
                     value->value.enumValue =
                        Symbols_GetLiteral(
                           ((tType)value1.value.typeValue)->SymEnumeration.Objects, 0 );
                     break;
                  case kSymSubrange:
                     *value = ((tType)value1.value.typeValue)->SymSubrange.Lwb;
                     break;
                  default:
                     /* error: not an ordinal type */
                     return FALSE;
                  } /* endswitch */
                  return TRUE;
               } /* endif */
               /* error: not a type specifier */
               return FALSE;

	    /* MAX() ---------- */
            case eSystemMax:
               if ( value1.Kind == eType ) {
                  /* TBC EKA
                   * type constants
                   */
                  switch ( ((tType)value1.value.typeValue)->Kind ) {
                  case kSymInteger:
                     *value = Values_MaxInteger;
                     break;
                  case kSymCardinal:
                     *value = Values_MaxCardinal;
                     break;
                  case kSymReal:
                     *value = Values_MaxReal;
                     break;
                  case kSymBoolean:
                     *value = Values_True;
                     break;
                  case kSymChar:
                     *value = Values_MaxChar;
                     break;
                  case kSymEnumeration:
                     value->Kind = eEnum;
                     value->value.enumValue =
                        Symbols_GetLiteral(
                           ((tType)value1.value.typeValue)->SymEnumeration.Objects,
                           ((tType)value1.value.typeValue)->SymEnumeration.MaxValue );
                     break;
                  case kSymSubrange:
                     *value = ((tType)value1.value.typeValue)->SymSubrange.Upb;
                     break;
                  default:
                     /* error: not a ordinal type */
                     return FALSE;
                  } /* endswitch */
                  return TRUE;
               } /* endif */
               /* error: not a type specifier */
               return FALSE;

	    /* ODD() ---------- */
            case eSystemOdd:
               if ( value1.Kind == eInteger ||
                    value1.Kind == eCardinal ) {
                  value->Kind = eBoolean;
                  value->value.booleanValue =
                     ( (value1.value.cardinalValue & 0x00000001L) ? TRUE : FALSE );
                  return TRUE;
               } /* endif */
               break;

	    /* ORD() ---------- */
            case eSystemOrd:
               if ( value1.Kind == eInteger ||
                    value1.Kind == eCardinal ) {
                  *value = value1;
                  return TRUE;
               } else if ( value1.Kind == eChar ||
                           value1.Kind == eStringChar ) {
                  value->Kind = eCardinal;
                  value->value.cardinalValue = (ULONG)value1.value.charValue;
                  return TRUE;
               } else if ( value1.Kind == eEnum ) {
                  value->Kind = eCardinal;
                  value->value.cardinalValue = (ULONG)
                     ((tObject)value1.value.enumValue)->SymEnumLiteral.Index;
                  return TRUE;
               } /* endif */
               break;

	    /* SIZE() = TSIZE() ---------- */
            case eSystemSize:
            case eSystemTSize:
               if ( value1.Kind == eType ) {
                  value->Kind = eCardinal;
                  value->value.cardinalValue =
                     Types_TypeSize( value1.value.typeValue );
                  return TRUE;
               } /* endif */
               break;

	    /* TRUNC() ---------- */
            case eSystemTrunc:
               if ( value1.Kind == eReal ) {
                  value->Kind = eInteger;
                  value->value.integerValue =
                     (LONG)floor( value1.value.realValue );
                  return TRUE;
               } /* endif */
               break;

	    /* VAL() ---------- */
            case eSystemVal:
               if ( tree->FuncCall.Actuals->Kind != kNoActual ) {
		  /* Evaluate second argument */
                  if ( Evaluate( tree->FuncCall.Actuals->Actual.Next, &value2, multivalue ) ) {
                     if ( value1.Kind == eType &&
                         (value2.Kind == eInteger || value2.Kind == eCardinal ) ) {
                        type = (tType)value1.value.typeValue;
                        if ( type->Kind == kSymSubrange ) {
                           type = type->SymSubrange.Type;
                        } /* endif */
                        switch ( type->Kind ) {
                        case kSymChar:
                           if ( value2.value.cardinalValue <= Types_MaxChar ) {
                              value->Kind = eChar;
                              value->value.charValue = (UCHAR)
                                 value2.value.cardinalValue;
                              return TRUE;
                           } else {
                              /* error: value out of type range */
                              return FALSE;
                           } /* endif */
                        case kSymInteger:
                        case kSymCardinal:
                           *value = value2;
                           return TRUE;
                        case kSymEnumeration:
                           object1 =
                              Symbols_GetLiteral( type->SymEnumeration.Objects,
                                                  value2.value.cardinalValue );
                           if ( object1 != NoObject ) {
                              value->Kind = eEnum;
                              value->value.enumValue = object1;
                              return TRUE;
                           } else {
                              /* error: value out of type range */
                              return FALSE;
                           } /* endif */
                           break;
                        } /* endswitch */
                     } else {
                        /* error: wrong types for procedure */
                        return FALSE;
                     } /* endif */
                  } /* endif */
                  /* error: expression not constant */
                  return FALSE;
               } else {
                  /* error: too view parameters */
                  return FALSE;
               } /* endif */
               break;
            } /* endswitch */
         } /* endif */
      } /* endif */
      return FALSE;
/*------*/
   case kSet:
      /* the compiler treats the values of all set constants as
       * values of Kind Bitset
       */
      return Evaluate( tree->Set.Elems, value, multivalue );
/*------*/
   case kBitSet:
      return Evaluate( tree->BitSet.Elems, value, multivalue );
/*------*/
   case kNoElem:
      value->Kind = eBitset;
      value->value.bitsetValue = EMPTYBITSET;
      return TRUE;
/*------*/
   case kElemExpr:
      if ( Evaluate( tree->ElemExpr.Next, value, multivalue ) &&
           Evaluate( tree->ElemExpr.Elem, &value1, multivalue ) ) {
         switch ( value1.Kind ) {
         case eCardinal:
         case eInteger:
            if ( value1.value.integerValue >= 0 &&
                 value1.value.integerValue <= Types_MaxBitset ) {
               BITSET_Include( value->value.bitsetValue,
                               value1.value.cardinalValue );
               return TRUE;
            } else {
               /* error: value out of type range */
               return FALSE;
            } /* endif */
            break;
         case eChar:
         case eStringChar:
            if ( value1.value.charValue <= Types_MaxBitset ) {
               BITSET_Include( value->value.bitsetValue,
                               value1.value.charValue );
               return TRUE;
            } else {
               /* error: value out of type range */
               return FALSE;
            } /* endif */
            break;
         case eEnum:
            if ( ((tObject)value1.value.enumValue)->SymEnumLiteral.Index <= Types_MaxBitset ) {
               BITSET_Include( value->value.bitsetValue,
                               ((tObject)value1.value.enumValue)->SymEnumLiteral.Index );
               return TRUE;
            } else {
               /* error: value out of type range */
               return FALSE;
            } /* endif */
            break;
	 default:
	    break;
         } /* endswitch */
      } /* endif */
      break;
/*------*/
   case kElemRange:
      if ( Evaluate( tree->ElemRange.Next, value, multivalue ) &&
           Evaluate( tree->ElemRange.Lwb, &value1, multivalue ) &&
           Evaluate( tree->ElemRange.Upb, &value2, multivalue ) ) {
         if ( (value1.Kind == eInteger || value1.Kind == eCardinal) &&
              (value2.Kind == eInteger || value2.Kind == eCardinal) ) {
            if ( value1.value.integerValue >= 0 &&
                 value2.value.integerValue <= Types_MaxBitset ) {
               for ( i=value1.value.cardinalValue;
                     i<=value2.value.cardinalValue; i++) {
                  BITSET_Include( value->value.bitsetValue, i );
               } /* endfor */
               return TRUE;
            } else {
               /* error: value out of type range */
               return FALSE;
            } /* endif */
         } else if ( (value1.Kind == eChar || value1.Kind == eStringChar) &&
                     (value2.Kind == eChar || value2.Kind == eStringChar) ) {
            if ( value2.value.charValue <= Types_MaxBitset ) {
               for ( i=value1.value.charValue; i<=value2.value.charValue; i++) {
                  BITSET_Include( value->value.bitsetValue, i );
               } /* endfor */
               return TRUE;
            } else {
               /* error: value out of type range */
               return FALSE;
            } /* endif */
         } else if ( value1.Kind == eEnum && value2.Kind == eEnum ) {
            object1 = value1.value.enumValue;
            object2 = value2.value.enumValue;
            if ( object2->SymEnumLiteral.Index <= Types_MaxBitset ) {
               for ( i=object1->SymEnumLiteral.Index;
                     i<=object2->SymEnumLiteral.Index; i++) {
                  BITSET_Include( value->value.bitsetValue, i );
               } /* endfor */
               return TRUE;
            } else {
               /* error: value out of type range */
               return FALSE;
            } /* endif */
         } /* endif */
      } /* endif */
      break;
/*------*/
   case kActualExpr:
      /* other actuals are not valid */
      return Evaluate( tree->ActualExpr.Expr, value, multivalue );
   } /* endswitch */

   return FALSE;
}


/*======================================================================*/

/*
 * Global variables
 */

/* Variables used for common Parallaxis constants. They're defined only once */
/* in Values_Init(). */

tValue  Values_Error,
        Values_Nil,
        Values_True,
        Values_False,
        Values_MinChar,
        Values_MaxChar,
        Values_Zero,
        Values_MaxCardinal,
        Values_MinInteger,
        Values_MaxInteger,
        Values_MinReal,
        Values_MaxReal,
        Values_Pi,
        Values_EOL;

/*======================================================================*/

/*
 * global function implementation
 */

/*------- ComputeConstant --------------------------------------------------*/

/*****************************************************************************
 *  Synopsis:       BOOLEAN Values_ComputeConstant( ADDRESS tree,
 *                                       ADRESS pEnv, tValue *value )
 *
 *  Parameters:     tree:  (sub-) tree containing expression
 *                  pEnv:  current environment to search identifiers in 
 *                  value: value in question
 *                            
 *  Result:	    TRUE:  expression has constant value
 *                  FALSE: expression cannot be evaluated (e.g. variables
 *                         are only known at runtime)
 *
 *  Description:
 *  -----------
 *  Evaluate the expression given by a subtree of the abstract syntax tree
 *  and return the result in value. If it is a constant expression, return
 *  TRUE, else return false. This is used to compute upper and lower bounds
 *  of subranges, label ranges and configurations as well as the value of
 *  constants.
 *****************************************************************************/

BOOLEAN Values_ComputeConstant( ADDRESS /* tTree */ tree,
                                ADDRESS /* tEnv */ pEnv, tValue *value )
{
   tValue dummy; /* only used by multiple expressions */

   env = (tEnv)pEnv;
   return( Evaluate( (tTree)tree, value, &dummy ) );
}

/*------- TypeOfValue ------------------------------------------------------*/

/*****************************************************************************
 *  Synopsis:       ADDRESS Values_TypeOfValue( tValue *value )
 *
 *  Parameters:     value: value in question
 *                            
 *  Result:	    type of the value (tType)
 *
 *  Description:
 *  ------------
 *  Return the type of the given value.
 *****************************************************************************/

ADDRESS /* tType */ Values_TypeOfValue( tValue *value )
{
   switch ( value->Kind ) {
   case eCardinal:   return TypeIntCard;
   case eInteger:    return TypeIntCard;
   case eReal:       return TypeREAL;
   case eBoolean:    return TypeBOOLEAN;
   case eChar:       return TypeCHAR;
   case eStringChar: return TypeStringChar;
   case eString:     return TypeString;
   case eBitset:     return TypeBITSET;
   case eEnum:       return( Symbols_GetType( value->value.enumValue ) );
   case eNilType:    return TypeNil;
   case eType:       return value->value.typeValue;
   default:          return NoType;
   } /* endswitch */
}

/*------- ValueToInt -------------------------------------------------------*/

/*****************************************************************************
 *  Synopsis:       LONG Values_ValueToInt( tValue *value )
 *
 *  Parameters:     value:  value to be converted
 *                            
 *  Result:	    (signed) integer value
 *
 *  Description:
 *  ------------
 *  Convert the value to an integer value. This may cause problems on
 *  CARDINAL (overflow). Convertable types are: CARDINAL, INTEGER, CHAR,
 *  BOOLEAN and enumeration constants. Every other value is returned as 0.
 *****************************************************************************/

LONG Values_ValueToInt( tValue *value )
{
   tObject object;

   switch ( value->Kind ) {
   case eCardinal:   return( value->value.cardinalValue );
   case eInteger:    return( value->value.integerValue );
   case eBoolean:    return( value->value.booleanValue );
   case eChar:
   case eStringChar: return( (LONG)value->value.charValue );
   case eEnum:       object = value->value.enumValue;
                     return( (LONG)object->SymEnumLiteral.Index );
   default:          return 0L;
   } /* endswitch */
}

/*------- ValueToInt -------------------------------------------------------*/

/*****************************************************************************
 *  Synopsis:       LONG Values_ValueAsString( tValue *value, char *buffer )
 *
 *  Parameters:     value:  value to be converted
 *                  buffer: Buffer where to put the string
 *                            
 *  Result:	    buffer
 *
 *  Description:
 *  ------------
 *  Depending on the type of value, the value is returned as a string in the
 *  given buffer.
 *****************************************************************************/

char *Values_ValueAsString( tValue *value, char *buffer )
{
   switch ( value->Kind ) {
   case eCardinal:   
	sprintf(buffer, "%lu", value->value.cardinalValue );
        break;  
   case eInteger:    
	sprintf(buffer, "%ld", value->value.integerValue );
	break;
   case eBoolean:
	sprintf(buffer, "%s", value->value.booleanValue == FALSE ? "FALSE" :
		"TRUE");
	break;
   case eChar:
   case eStringChar: 
	sprintf(buffer, "%d", (int)value->value.charValue );
	break;
   case eEnum:
 	GetString(((tObject)value->value.enumValue)->SymEnumLiteral.Ident, buffer );
	break;
   default:   
	sprintf(buffer, "(?)");
	break;
   } /* endswitch */
   return buffer;
}

/*------- WriteValue -------------------------------------------------------*/

/*****************************************************************************
 *  Synopsis:       void Values_WriteValue( FILE *f, tValue *value )
 *
 *  Parameters:     f:     file to write to
 *                  value: value to be written
 *                            
 *  Result:	    -
 *
 *  Description:
 *  ------------
 *  Write the given value to the given file. Depending on the value type the
 *  value is written in the appropriate format. This is mainly used for
 *  testing purposes (command line option -Zs of p3/p3c).
 *****************************************************************************/

void Values_WriteValue( FILE *f, tValue *value )
{
   tObject object;

   switch ( value->Kind ) {
   case eNoValue:
      fprintf( f, "NoValue" );
      break;
   case eCardinal:
      fprintf( f, "Cardinal, %lu", value->value.cardinalValue );
      break;
   case eInteger:
      fprintf( f, "Integer,  %ld", value->value.integerValue );
      break;
   case eReal:
      fprintf( f, "Real,     %f", value->value.realValue );
      break;
   case eBoolean:
      fprintf( f, "Boolean,  %s", value->value.booleanValue ? "True" : "False" );
      break;
   case eChar:
   case eStringChar:
      fprintf( f, "Char,     %c", value->value.charValue );
      break;
   case eString:
      fprintf( f, "String,   %s", (char *)value->value.stringValue +2);
      break;
   case eBitset:
      fprintf( f, "Bitset,   %08lx", value->value.bitsetValue );
      break;
   case eEnum:
      object = value->value.enumValue;
      fputs( "enum,     ", f ); WriteIdent( f, object->SymEnumLiteral.Ident );
      break;
   case eNilType:
      fprintf( f, "NIL" );
      break;
   case eStdProc:
      fprintf( f, "<StdProc>" );
      break;
   case eType:
      fprintf( f, "<Type>" );
      break;
   } /* endswitch */
}

/*------- Init -------------------------------------------------------*/

/*****************************************************************************
 *  Synopsis:       void Values_Init( void )
 *
 *  Parameters:     -
 *                            
 *  Result:	    -
 *
 *  Description:
 *  ------------
 *  Initialize the module. Create value nodes of type tValue for some 
 *  constants.
 *****************************************************************************/

void Values_Init( void )
{
   if ( !isInitialised ) {
      isInitialised = TRUE;

      BigSet_Init();
      Types_Init();


      Values_Error.Kind = eNoValue;	                 /* Error */
      Values_Nil.Kind = eNilType;	                 /* NIL */
      Values_True.Kind= eBoolean;	                 /* TRUE */
      Values_True.value.booleanValue = TRUE;
      Values_False.Kind= eBoolean;	                 /* FALSE */
      Values_False.value.booleanValue = FALSE;
      Values_MinChar.Kind = eChar;                       /* CHR(0) */
      Values_MinChar.value.charValue = Types_MinChar;
      Values_MaxChar.Kind = eChar;                       /* CHR(255) */
      Values_MaxChar.value.charValue = Types_MaxChar;
      Values_Zero.Kind = eCardinal;                      /* MIN(CARDINAL) */
      Values_Zero.value.cardinalValue = 0L;
      Values_MaxCardinal.Kind = eCardinal;               /* MAX(CARDINAL) */
      Values_MaxCardinal.value.cardinalValue = Types_MaxCardinal;
      Values_MinInteger.Kind = eInteger;                 /* MIN(INTEGER) */
      Values_MinInteger.value.integerValue = Types_MinInteger;
      Values_MaxInteger.Kind = eInteger;                 /* MAX(INTEGER) */
      Values_MaxInteger.value.integerValue = Types_MaxInteger;
      Values_MinReal.Kind = eReal;                       /* MIN(REAL) */
      Values_MinReal.value.realValue = Types_MinReal;
      Values_MaxReal.Kind = eReal;                       /* MAX(REAL) */
      Values_MaxReal.value.realValue = Types_MaxReal;
      Values_Pi.Kind = eReal;                            /* Pi */
      Values_Pi.value.realValue = 3.14159265;
      Values_EOL.Kind = eChar;		                 /* EOLN */
      Values_EOL.value.charValue = '\n';
   } /* endif */
}

