/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name :  BigSet.c
 * desc :  sets of integer values, unlimited in size
 * rcs  :  $Author: keller $
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "DynArray.h"			  /* Reuse-Library */
#include "BigSet.h"

static BOOLEAN isInitialised = FALSE;

#define LOG2BITSPERBITSET     5
#define MASKBITSPERBITSET     0x0000001F

#define NOCARD                 -1
#define ONE                    0x80000000

/*
 * local variables
 */
static FILE *g;

/*
 * local function prototypes
 */
static void PrintElement( ULONG elem );

/*
 * local function implementation
 */
static void PrintElement( ULONG elem )
{
   fprintf( g, " %ld", elem );
}

/*
 * Global variables
 */

/*
 * global function implementation
 */
/*------- MakeSet ----------------------------------------------------------*/

void BITSET_WriteSet( FILE *f, BITSET set )
{
   ULONG i;

   for (i=0; i<BITSPERBITSET; i++) {
      if ( BITSET_IsElement(set,i) ) {
         fprintf( f, "%ld ", i );
      } /* endif */
   } /* endfor */
}

/*------- MakeSet ----------------------------------------------------------*/

tBigSet *BigSet_MakeSet( tBigSet *set, ULONG maxsize )
{
   ULONG elemcount;

   elemcount = (maxsize + BITSPERBITSET - (maxsize & MASKBITSPERBITSET)) >> LOG2BITSPERBITSET;
   MakeArray((char **)&set->bitsetPtr, &elemcount, (ULONG)sizeof(BITSET) );
   set->maxElem = maxsize;
   set->lastBitset = elemcount - 1;
   BigSet_AssignEmpty( set );

   return set;
}

/*------- Union ------------------------------------------------------------*/

tBigSet *BigSet_Union( tBigSet *set1, tBigSet *set2 )
{
   ULONG    i  = set1->lastBitset + 1;
   BITSET  *s1 = set1->bitsetPtr;
   BITSET  *s2 = set2->bitsetPtr;

   do {
      *s1++ |= *s2++;
   } while ( --i ); /* enddo */

   set1->card      = NOCARD;
   set1->firstElem = min( set1->firstElem, set2->firstElem );
   set1->lastElem  = max( set1->lastElem, set2->lastElem );

   return set1;
}

/*------- Difference -------------------------------------------------------*/

tBigSet *BigSet_Difference( tBigSet *set1, tBigSet *set2 )
{
   ULONG    i  = set1->lastBitset + 1;
   BITSET  *s1 = set1->bitsetPtr;
   BITSET  *s2 = set2->bitsetPtr;

   do {
      *s1++ &= ~*s2++;
   } while ( --i ); /* enddo */

   set1->card      = NOCARD;

   return set1;
}

/*------- Intersection -----------------------------------------------------*/

tBigSet *BigSet_Intersection( tBigSet *set1, tBigSet *set2 )
{
   ULONG    i  = set1->lastBitset + 1;
   BITSET  *s1 = set1->bitsetPtr;
   BITSET  *s2 = set2->bitsetPtr;

   do {
      *s1++ &= *s2++;
   } while ( --i ); /* enddo */

   set1->card      = NOCARD;
   set1->firstElem = max( set1->firstElem, set2->firstElem );
   set1->lastElem  = min( set1->lastElem, set2->lastElem );

   return set1;
}

/*------- SymDifference ----------------------------------------------------*/

tBigSet *BigSet_SymDifference( tBigSet *set1, tBigSet *set2 )
{
   ULONG    i  = set1->lastBitset + 1;
   BITSET  *s1 = set1->bitsetPtr;
   BITSET  *s2 = set2->bitsetPtr;

   do {
      *s1++ ^= *s2++;
   } while ( --i ); /* enddo */

   set1->card      = NOCARD;
   set1->firstElem = min( set1->firstElem, set2->firstElem );
   set1->lastElem  = max( set1->lastElem, set2->lastElem );

   return set1;
}

/*------- Complement -------------------------------------------------------*/

tBigSet *BigSet_Complement( tBigSet *set )
{
   ULONG    i = set->lastBitset + 1;
   BITSET  *s = set->bitsetPtr;

   do {
      *s = ~*s;
      s++;
   } while ( i-- ); /* enddo */

   if ( set->card != NOCARD ) {
      set->card = set->maxElem + 1 - set->card;
   }
   set->firstElem = 0;
   set->lastElem  = set->maxElem;

   return set;
}

/*------- Include ----------------------------------------------------------*/

tBigSet *BigSet_Include( tBigSet *set, ULONG elem )
{
   set->bitsetPtr[elem >> LOG2BITSPERBITSET] |= (ONE >> (elem & MASKBITSPERBITSET));
   set->card      = NOCARD;
   set->firstElem = min( set->firstElem, elem );
   set->lastElem  = max( set->lastElem, elem );

   return set;
}

/*------- Exclude ----------------------------------------------------------*/

tBigSet *BigSet_Exclude( tBigSet *set, ULONG elem )
{
   set->bitsetPtr[elem >> LOG2BITSPERBITSET] &= ~(ONE >> (elem & MASKBITSPERBITSET));
   set->card      = NOCARD;
   if ( elem == set->firstElem && elem < set->maxElem ) {
      set->firstElem++;
   } /* endif */
   if ( elem == set->lastElem && elem > 0 ) {
      set->lastElem--;
   } /* endif */

   return set;
}

/*------- Card -------------------------------------------------------------*/

ULONG BigSet_Card( tBigSet *set )
{
   ULONG i;
   if ( set->card == NOCARD ) {
      set->card = 0;
      for (i=set->firstElem; i<=set->lastElem; i++) {
         if ( (set->bitsetPtr[i >> LOG2BITSPERBITSET] &
           (ONE >> (i & MASKBITSPERBITSET))) != 0 ) {
            set->card++;
         } /* endif */
      } /* endfor */
   } /* endif */

   return set->card;
}

/*------- Size -------------------------------------------------------------*/

ULONG BigSet_Size( tBigSet *set )
{
   return set->maxElem;
}

/*------- Minimum ----------------------------------------------------------*/

ULONG BigSet_Minimum( tBigSet *set )
{
   ULONG i;

   for (i=set->firstElem; i<=set->lastElem; i++) {
      if ( (set->bitsetPtr[i >> LOG2BITSPERBITSET] &
           (ONE >> (i & MASKBITSPERBITSET))) != 0 ) {
         set->firstElem = i;
         return i;
      }
   }

   return 0L;
}

/*------- Maximum ----------------------------------------------------------*/

ULONG BigSet_Maximum( tBigSet *set )
{
   ULONG i;

   for (i=set->lastElem; i>=set->firstElem; i--) {
      if ( (set->bitsetPtr[i >> LOG2BITSPERBITSET] &
           (ONE >> (i & MASKBITSPERBITSET))) != 0 ) {
         set->lastElem = i;
         return i;
      }
   }

   return 0L;
}

/*------- Select -----------------------------------------------------------*/

ULONG BigSet_Select( tBigSet *set )
{
   return BigSet_Minimum( set );
}

/*------- Extract ----------------------------------------------------------*/

ULONG BigSet_Extract( tBigSet *set )
{
   ULONG elem = BigSet_Minimum( set );

   BigSet_Exclude( set, elem );
   return elem;
}

/*------- IsSubset ---------------------------------------------------------*/

BOOLEAN BigSet_IsSubset( tBigSet *set1, tBigSet *set2 )
{
   ULONG    i  = set1->lastBitset + 1;
   BITSET  *s1 = set1->bitsetPtr;
   BITSET  *s2 = set2->bitsetPtr;

   do {
      if ( *s1++ &= ~*s2++ ) {
         return FALSE;
      } /* endif */
   } while ( --i ); /* enddo */

   return TRUE;
}

/*------- IsStrictSubset ---------------------------------------------------*/

BOOLEAN BigSet_IsStrictSubset( tBigSet *set1, tBigSet *set2 )
{
   return( BigSet_IsSubset( set1, set2 ) && !BigSet_IsEqual( set1, set2 ) );
}

/*------- IsEqual ----------------------------------------------------------*/

BOOLEAN BigSet_IsEqual( tBigSet *set1, tBigSet *set2 )
{
   ULONG    i  = set1->lastBitset + 1;
   BITSET  *s1 = set1->bitsetPtr;
   BITSET  *s2 = set2->bitsetPtr;

   do {
      if ( *s1++ != *s2++ ) {
         return FALSE;
      } /* endif */
   } while ( --i ); /* enddo */

   return TRUE;
}

/*------- IsNotEqual -------------------------------------------------------*/

BOOLEAN BigSet_IsNotEqual( tBigSet *set1, tBigSet *set2 )
{
   return( !BigSet_IsEqual( set1, set2 ) );
}

/*------- IsElement --------------------------------------------------------*/

BOOLEAN BigSet_IsElement( tBigSet *set, ULONG elem )
{
   return( (set->bitsetPtr[elem >> LOG2BITSPERBITSET] &
           (ONE >> (elem & MASKBITSPERBITSET))) != 0 );
}

/*------- IsEmpty ----------------------------------------------------------*/

extern BOOLEAN BigSet_IsEmpty( tBigSet *set )
{
   if ( set->card == NOCARD ) {
      BigSet_Card( set );
   } /* endif */

   return( set->card == 0L );
}

/*------- Forall -----------------------------------------------------------*/

BOOLEAN BigSet_Forall( tBigSet *set, tBigSetTestFunc func )
{
   ULONG i;

   for (i=set->firstElem; i<=set->lastElem; i++) {
      if ( ((set->bitsetPtr[i >> LOG2BITSPERBITSET] &
           (ONE >> (i & MASKBITSPERBITSET))) != 0) && !func( i ) ) {
         return FALSE;
      } /* endif */
   } /* endfor */

   return TRUE;
}

/*------- Exists -----------------------------------------------------------*/

BOOLEAN BigSet_Exists( tBigSet *set, tBigSetTestFunc func )
{
   ULONG i;

   for (i=set->firstElem; i<=set->lastElem; i++) {
      if ( ((set->bitsetPtr[i >> LOG2BITSPERBITSET] &
           (ONE >> (i & MASKBITSPERBITSET))) != 0) && func( i ) ) {
         return TRUE;
      } /* endif */
   } /* endfor */

   return FALSE;
}

/*------- Exists1 ----------------------------------------------------------*/

BOOLEAN BigSet_Exists1( tBigSet *set, tBigSetTestFunc func )
{
   ULONG i,
         n = 0;


   for (i=set->firstElem; i<=set->lastElem && n<=1; i++) {
      if ( ((set->bitsetPtr[i >> LOG2BITSPERBITSET] &
           (ONE >> (i & MASKBITSPERBITSET))) != 0) && func( i ) ) {
         n++;
      } /* endif */
   } /* endfor */

   return( n == 1 );
}

/*------- Assign -----------------------------------------------------------*/

tBigSet *BigSet_Assign( tBigSet *set1, tBigSet *set2 )
{
   ULONG    i  = set1->lastBitset + 1;
   BITSET  *s1 = set1->bitsetPtr;
   BITSET  *s2 = set2->bitsetPtr;

   do {
      *s1++ = *s2++;
   } while ( --i ); /* enddo */

   set1->card      = set2->card;
   set1->firstElem = set2->firstElem;
   set1->lastElem  = set2->lastElem;

   return set1;
}

/*------- AssignElement ----------------------------------------------------*/

tBigSet *BigSet_AssignElement( tBigSet *set, ULONG elem )
{
   BigSet_AssignEmpty( set );
   BigSet_Include( set, elem );
   set->card      = 1;
   set->firstElem = elem;
   set->lastElem  = elem;

   return set;
}

/*------- AssignEmpty ------------------------------------------------------*/

tBigSet *BigSet_AssignEmpty( tBigSet *set )
{
   ULONG    i = set->lastBitset + 1;
   BITSET  *s = set->bitsetPtr;

   do {
      *s++ = 0L;
   } while ( --i ); /* enddo */

   set->card      = 0L;
   set->firstElem = set->maxElem;
   set->lastElem  = 0L;

   return set;
}

/*------- ForallDo ---------------------------------------------------------*/

void BigSet_ForallDo( tBigSet *set, tBigSetDoFunc func )
{
   ULONG i;

   for (i=set->firstElem; i<=set->lastElem; i++) {
      if ( (set->bitsetPtr[i >> LOG2BITSPERBITSET] &
           (ONE >> (i & MASKBITSPERBITSET))) != 0 ) {
         func( i );
      } /* endif */
   } /* endfor */
}

/*------- ReadSet ----------------------------------------------------------*/

void BigSet_ReadSet( FILE *f, tBigSet *set )
{
   ULONG i, card = 0L;
   int   c;
   while ( (c = fgetc( f )) != EOF && c != '{' )
      ;
   if ( c != EOF ) {
      BigSet_AssignEmpty( set );
      while ( TRUE ) {
         if ( (c = fgetc( f )) == EOF || c == '}' ) {
            break;
         } /* endif */
         fscanf( f, "%ld", &i );
         BigSet_Include( set, i );
         card++;
      } /* endwhile */
      set->card = card;
   } /* endif */
}

/*------- WriteSet ---------------------------------------------------------*/

void BigSet_WriteSet( FILE *f, tBigSet *set )
{
   g = f;

   fprintf( f, "{" );
   BigSet_ForallDo( set, PrintElement );
   fprintf( f, "}" );
}

/*------- ReleaseSet -------------------------------------------------------*/

void BigSet_ReleaseSet( tBigSet *set )
{
   ULONG elemcount;

   elemcount = set->lastBitset + 1;
   ReleaseArray((char **)&set->bitsetPtr, &elemcount, (ULONG)sizeof(BITSET) );
}

void BigSet_Init( void )
{
   if ( !isInitialised ) {
      isInitialised = TRUE;
   } /* endif */
}

