/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  Types.c
 * desc  :  general handling of types and type compatibility
 * author:  lampke
 *
 ************************************************************************/

#include <stdio.h>
#include <math.h>
#include "BigSet.h"
#include "Errors.h"
#include "Tree.h"
#include "Symbols.h"
#include "StdProcs.h"
#include "Values.h"
#include "Types.h"
#include "Code.h"
#include "Options.h"

static BOOLEAN isInitialised = FALSE;

#define NoSize          0

#define SizeCHAR        1
#define AlignCHAR       1
#define SizeINTEGER     4
#define AlignINTEGER    4
#define SizeCARDINAL    4
#define AlignCARDINAL   4
#define SizeREAL        8
#define AlignREAL       8
#define SizePOINTER     4
#define AlignPOINTER    4
#define SizeENUM        4
#define AlignENUM       4

/*
 * local function prototypes
 */
static BOOLEAN Coercible( tType t1, tType t2 );
static BOOLEAN ParameterCoercible( tType t1, tType t2 );
static BOOLEAN ConfigCoercible( tType vt1, tType vt2 );
static BOOLEAN ConfigParameterCoercible( tType vt1, tType vt2 );
static tType   Operable( INDEX operator, tType t1, tType t2 );
static ULONG   TypeSize( tType type, ULONG *align );
static ULONG   TypeSizeFields( tObjects fields, ULONG *align );

/*
 * local function implementation
 */
/*************************************************************************
 * name :  Coercible
 * in   :  two types
 * out  :  TRUE if two types are coercible, e.g. t1 can be assigned to t2
 *         FALSE otherwise
 ************************************************************************/
static BOOLEAN Coercible( tType t1, tType t2 )
{
  if (t1 == t2 || t1 == NoType || t2 == NoType) {
    return TRUE;
  }

  if ((t1 == TypeString || t1 == TypeStringChar) &&
       t2->Kind == kSymArray && t2->SymArray.ElemType == TypeCHAR) {
    /* string to array of char */ 
    return TRUE;
  } 

  if (t2->Kind == kSymSubrange) {
    t2 = t2->SymSubrange.Type;
    /* basetype of subrange is also allowed without typecast */
  }
  switch (t1->Kind) {
  case kSymChar      : return t2 == TypeCHAR || t2 == TypeStringChar;
  case kSymStringChar: return t2 == TypeStringChar || t2 == TypeCHAR;
  case kSymInteger   : return t2 == TypeINTEGER || t2 == TypeCARDINAL ||
      t2 == TypeIntCard;
  case kSymCardinal  :
    return t2 == TypeCARDINAL || t2 == TypeINTEGER || t2 == TypeADDRESS ||
      t2 == TypeIntCard;
  case kSymIntCard   : return t2 == TypeCARDINAL || t2 == TypeINTEGER ||
      t2 == TypeIntCard;
  case kSymReal      : return t2 == TypeREAL;
  case kSymAddress   : return t2 == TypeCARDINAL || t2->Kind == kSymPointer;
  case kSymNil       : return t2 == TypeADDRESS || t2->Kind == kSymPointer;
  case kSymSet       :
    return t2 == TypeBITSET ||
    (t2->Kind == kSymSet && Coercible(t1->SymSet.Type, t2->SymSet.Type));
  case kSymBitset    :
    return t2 == TypeBITSET ||
    (t2->Kind == kSymSet && Coercible(TypeINTEGER, t2->SymSet.Type));
  case kSymSubrange  : return Coercible(t1->SymSubrange.Type, t2);
    /* basetype of subrange is also allowed without typecast */
  case kSymProcType  : 
    /* procedure types are coercible if all parameters are coercible and the */
    /* return types are coercible */
    {
	 BOOLEAN coerc;
	 tType param1, param2;

	 if (t2->Kind != kSymProcType) return FALSE;

	 /* check return types */
	 coerc = Types_IsReturnCompatible(t2->SymProcType.Type, t1->SymProcType.Type);

	 /* check parameter types */
	 param1 = t1->SymProcType.Types;
	 param2 = t2->SymProcType.Types;
	 while(coerc && ((param1 != NoTypes) || (param2 != NoTypes)))
	 {
	      if (param1 == NULL) return FALSE;	/* different number of */
	      if (param2 == NULL) return FALSE;	/* parameters */
	      coerc = coerc && Types_IsParameterCompatible(param2->SymTypes.Type, param1->SymTypes.Type);
	      param1 = param1->SymTypes.Next;
	      param2 = param2->SymTypes.Next;
	 }
	 return coerc;
    }
  } /* endswitch */
#if 0
  printf("expr: %d, lval: %d\n",t1->Kind,t2->Kind);
#endif
  return FALSE;
}

/*************************************************************************
 * name :  ParameterCoercible
 * in   :  two types
 * out  :  TRUE if t1 can be passed to t2
 *         FALSE otherwise
 ************************************************************************/
static BOOLEAN ParameterCoercible( tType t1, tType t2 )
{
  /*  coercible, if assign coercible or an array is passed to an open array
   *  and the types of their elements are assign coercible
   */
  return
    Coercible(t1, t2) ||
    (t1->Kind == kSymArray && t2->Kind == kSymArray && t2->SymArray.IsOpen &&
     Coercible(t1->SymArray.ElemType, t2->SymArray.ElemType));
}

/*************************************************************************
 * name :  ConfigCoercible
 * in   :  two vector types
 * out  :  TRUE if configurations are coercible
 *         FALSE otherwise
 ************************************************************************/
static BOOLEAN ConfigCoercible(tType vt1, tType vt2)
{
  /* two vector types are coercible if
   * both configurations are identical, or
   * both configurations belong to the same configuration group, or
   * both configurations are open
   */
  return vt1->Kind == kSymVector && vt2->Kind == kSymVector &&
         ((vt1->SymVector.Config == vt2->SymVector.Config)   ||
          (vt1->SymVector.Config->Kind == kSymConfig &&
           vt2->SymVector.Config->Kind == kSymConfig &&
           vt1->SymVector.Config->SymConfig.ConfigGroup ==
           vt2->SymVector.Config->SymConfig.ConfigGroup)     ||
          (vt1->SymVector.IsOpen && vt2->SymVector.IsOpen));
}

/*************************************************************************
 * name :  ConfigParameterCoercible
 * in   :  two vector types
 * out  :  TRUE if vt1 can be passed to vt2
 *         FALSE otherwise
 ************************************************************************/
static BOOLEAN ConfigParameterCoercible(tType vt1, tType vt2)
{
  /* vector of type vt1 can be passed as argument to type vt2 if
   * both configurations are identical, or
   * both configurations belong to the same configuration group, or
   * both configurations are open, or
   * vt1 has a concrete configuration and vt2 is an open configuration, or
   * vt1 is respecified from vt2, or
   * vt1 and vt2 are generic
   */
  if ((vt1->Kind == kSymVector && vt1->SymVector.Config == NULL) ||
      (vt2->Kind == kSymVector && vt2->SymVector.Config == NULL)) {
    return TRUE;
  } else {
  return vt1->Kind == kSymVector && vt2->Kind == kSymVector &&
         ((vt1->SymVector.Config == vt2->SymVector.Config)   ||
          (vt1->SymVector.Config->Kind == kSymConfig &&
           vt2->SymVector.Config->Kind == kSymConfig &&
           vt1->SymVector.Config->SymConfig.ConfigGroup ==
           vt2->SymVector.Config->SymConfig.ConfigGroup)     ||
          (vt1->SymVector.IsOpen && vt2->SymVector.IsOpen)   ||
          (!vt1->SymVector.IsOpen && vt2->SymVector.IsOpen)  ||
          (vt1->SymVector.Config->Kind == kSymConfig &&
           vt1->SymVector.Config->SymConfig.IsRespecified &&
           vt2->SymVector.Config->Kind == kSymConfig &&
           vt2->SymVector.Config->SymConfig.IsGeneric)       ||
          (vt1->SymVector.Config->Kind == kSymConfig &&
           vt1->SymVector.Config->SymConfig.IsGeneric &&
           vt2->SymVector.Config->Kind == kSymConfig &&
           vt2->SymVector.Config->SymConfig.IsGeneric)
         );
  }
}

/*************************************************************************
 * name :  Operable
 * in   :  operator and types to be operated
 * out  :  result type of operation
 *         or error type if operation is invalid
 ************************************************************************/
static tType Operable(INDEX operator, tType t1, tType t2)
{
  if (t1 == NoType || t2 == NoType) {
    return NoType;
  } /* endif */
  /* Record-types are only Operable if the Operator is OpEqual or OpNotEqual */
  if (((t1->Kind == kSymRecord) || (t1->Kind == kSymRecord))&&((operator!=OpEqual) &&(operator!= OpNotEqual)))
    return mSymError(xxIllegalOperator, operator, -1L);
  switch (operator) {
  case OpLess        :
  case OpGreater     :
  case OpEqual       :
  case OpNotEqual    :
  case OpLessEqual   :
  case OpGreaterEqual:
    if (t1 == t2 || Coercible(t2, t1) || Coercible(t1, t2)) {
      return TypeBOOLEAN;
    }
    break;
  case OpIn:
    switch (t2->Kind) {
    case kSymSet:
      if (Coercible(t1, t2->SymSet.Type)) {
        return TypeBOOLEAN;
      } else {
        return mSymError(xxOperandIncomp, -1L, -1L);
      }
      break;
    case kSymBitset:
      if (Coercible(t1, TypeINTEGER)) {
        return TypeBOOLEAN;
      } else {
        return mSymError(xxOperandIncomp, -1L, -1L);
      }
      break;
    } /* endswitch */
    break;
  case OpOr:
  case OpAnd:
    if (t1 == TypeBOOLEAN && t2 == TypeBOOLEAN) {
      return TypeBOOLEAN;
    } /* endif */
    break;
  case OpNot:
    if (t1 == TypeBOOLEAN) {
      return t1;
    }
    break;
  case OpPlus:
  case OpMinus:
  case OpTimes:
    if ((Symbols_IsIntType(t1) ||
         Symbols_IsRealType(t1) ||
         Symbols_IsSetType(t1)) &&
        (Symbols_IsIntType(t2) ||
         Symbols_IsRealType(t2) ||
         Symbols_IsSetType(t2))) {
      if (t1 == t2 || Coercible(t2, t1)) {
        return t1;
      } else if (Coercible(t1, t2)) {
        return t2;
      } else {
        return mSymError(xxOperandIncomp, -1L, -1L);
      } /* endif */
    } /* endif */
    break;
  case OpDivide:
    if ((Symbols_IsRealType(t1) || Symbols_IsSetType(t1)) &&
        (Symbols_IsRealType(t2) || Symbols_IsSetType(t2))) {
      if (t1 == t2 || Coercible(t2, t1)) {
        return t1;
      } else if (Coercible(t1, t2)) {
        return t2;
      } else {
        return mSymError(xxOperandIncomp, -1L, -1L);
      } /* endif */
    } /* endif */
    break;
  case OpDiv:
  case OpMod:
    if (Symbols_IsIntType(t1) && Symbols_IsIntType(t2)) {
      if (t1 == t2 || Coercible(t2, t1)) {
        return t1;
      } else if (Coercible(t1, t2)) {
        return t2;
      } else {
        return mSymError(xxOperandIncomp, -1L, -1L);
      } /* endif */
    } /* endif */
    break;
  case OpPower:
    if ((Symbols_IsIntType(t1) || Symbols_IsRealType(t1)) &&
        (Symbols_IsIntType(t2) || Symbols_IsRealType(t2))) {
      if (t1 == t2 || Coercible(t2, t1)) {
        return t1;
      } else if (Coercible(t1, t2)) {
        return t2;
      } else {
        return mSymError(xxOperandIncomp, -1L, -1L);
      } /* endif */
    } /* endif */
    break;
  } /* endswitch */
  return mSymError(xxIllegalOperator, operator, -1L);
}

/*************************************************************************
 * name :  TypeSize
 * in   :  type and pointer to alignment
 * out  :  size and alignment of type
 ************************************************************************/
static ULONG TypeSize( tType type, ULONG *align )
{
  ULONG    localsize;

  if ( type->Kind == kSymSubrange ) {
    type = type->SymSubrange.Type;
  } /* endif */

  switch (type->Kind) {
  case kSymChar       :
  case kSymBoolean    : *align = AlignCHAR;     return SizeCHAR;
  case kSymInteger    : *align = AlignINTEGER;  return SizeINTEGER;
  case kSymCardinal   : *align = AlignCARDINAL; return SizeCARDINAL;
  case kSymReal       : *align = AlignREAL;     return SizeREAL;
  case kSymBitset     :
  case kSymSet        : *align = AlignCARDINAL; return SizeCARDINAL;
  case kSymProc       : *align = AlignPOINTER;  return SizePOINTER;
  case kSymWord       : *align = AlignCHAR;     return SizeCHAR;
  case kSymAddress    : *align = AlignPOINTER;  return SizePOINTER;
  case kSymOpaqueType : *align = AlignPOINTER;  return SizePOINTER;
  case kSymArray      :
    localsize = TypeSize(type->SymArray.ElemType, align);
    switch ( type->SymArray.IndexType->Kind ) {
/*
    case kSymQualident:
         return( localSize * ??? );
 */
    case kSymEnumeration:
         return( localsize *
                 (type->SymArray.IndexType->SymEnumeration.MaxValue + 1) );
    case kSymSubrange:
         return( localsize *
            (Values_ValueToInt( &(type->SymArray.IndexType->SymSubrange.Upb) ) -
            Values_ValueToInt( &(type->SymArray.IndexType->SymSubrange.Lwb) ) + 1) );
    } /* endswitch */
    return NoSize;
  case kSymEnumeration: *align = AlignENUM;     return SizeENUM;
  case kSymPointer    : *align = AlignPOINTER;  return SizePOINTER;
  case kSymProcType   : *align = AlignPOINTER;  return SizePOINTER;
  case kSymRecord     : return TypeSizeFields(Symbols_GetFields(type), align);
  default             : /* error */             return NoSize;
  }
}

/*************************************************************************
 * name :  TypeSizeFields
 * in   :  fieldlist and pointer to alignment
 * out  :  size and alignment of fieldlist
 ************************************************************************/
static ULONG TypeSizeFields( tObjects fields, ULONG *align )
{
  ULONG    localsize1, localsize2, localalign, localadjust;

  switch (fields->Kind) {
  case kSymElem:
    localsize1 = 0;
    do {
         localsize2 = TypeSize( fields->SymElem.Object, &localalign );
         *align = max( *align, localalign );
         localadjust = localsize1 % localalign;
         if ( localadjust != 0 ) {
            localadjust = localalign - localadjust;
         } /* endif */
         localsize1 += localsize2 + localadjust;
    } while ( fields != NoObjects ); /* enddo */
    break;
  case kSymUnion:
    localsize1 = TypeSizeFields( fields->SymUnion.Objects1, align );
    localsize2 = TypeSizeFields( fields->SymUnion.Objects1, &localalign );
    *align = max( *align, localalign );
    return( max( localsize1, localsize2 ) );
  } /* endswitch */
  return NoSize;
}

/*
 * Global variables
 */

/*
 * global function implementation
 */

/*------- IsAssignCompatible -----------------------------------------------*/

BOOLEAN Types_IsAssignCompatible(tType lvalue, tType expr)
{
  if (lvalue == expr) {
    return TRUE;
  }

  if (lvalue->Kind == kSymVector && expr->Kind == kSymVector ) {
    /* lvalue and expr are of vector type */
    return ConfigCoercible(expr, lvalue) &&
           Coercible(expr->SymVector.Type, lvalue->SymVector.Type);
  } else if (lvalue->Kind == kSymVector) {
    /* only lvalue is of vector type */
    return Coercible(expr, lvalue->SymVector.Type);
  } else if (expr->Kind == kSymVector) {
    /* only expr is of vector type: ERROR ! */
    return FALSE;
  } else {
    return Coercible(expr, lvalue);
  }
}

/*------- IsParameterCompatible --------------------------------------------*/

BOOLEAN Types_IsParameterCompatible_old_and_not_yet_removed(tType lvalue, tType expr) 
{
  if (lvalue == expr) {
    return TRUE;
  }

  if (lvalue->Kind == kSymVector && expr->Kind == kSymVector) {
    /* lvalue and expr are of vector type */
    return ConfigParameterCoercible(expr, lvalue) &&
           ParameterCoercible(expr->SymVector.Type, lvalue->SymVector.Type);
  } else if (lvalue->Kind == kSymVector) {
    return FALSE;
    /* AT THE MOMENT: assign of scalar type to vector type is not allowed
     * LEAVE FOR POSSIBLE FURTHER CHANGE
     * only lvalue is of vector type: lvalue must not be an open vector!
     * return !lvalue->SymVector.IsOpen &&
     *        ParameterCoercible(expr, lvalue->SymVector.Type);
     */
  } else if (expr->Kind == kSymVector) {
    /* only expr is of vector type: ERROR ! */
    return FALSE;
  } else {
    return ParameterCoercible(expr, lvalue);
  }
}

BOOLEAN Types_IsParameterCompatible(tType lvalue, tType expr)
{
  BOOLEAN lvalIsVec, exprIsVec;
  lvalIsVec=(lvalue->Kind == kSymVector);
  exprIsVec=(expr->Kind == kSymVector);
  return (lvalue==expr 
	  || (lvalIsVec && exprIsVec && ConfigParameterCoercible(expr, lvalue) &&
	     ParameterCoercible(expr->SymVector.Type, lvalue->SymVector.Type))
	  /* AT THE MOMENT: assign of scalar type to vector type is not allowed
	     LEAVE FOR POSSIBLE FURTHER CHANGE
	     only lvalue is of vector type: lvalue must not be an open vector!
	     || (lvalIsVec  && !exprIsVec && !lvalue->SymVector.IsOpen &&
	      ParameterCoercible(expr, lvalue->SymVector.Type))  
	  */
	  || (!lvalIsVec && !exprIsVec && ParameterCoercible(expr, lvalue))
	 );
}






/*------- IsReturnCompatible -----------------------------------------------*/

BOOLEAN Types_IsReturnCompatible(tType lvalue, tType expr)
{
  if (lvalue == expr) {
    return TRUE;
  }

  if (lvalue->Kind == kSymVector && expr->Kind == kSymVector) {
    /* lvalue and expr are of vector type */
    return ConfigCoercible(expr, lvalue) &&
           Coercible(expr->SymVector.Type, lvalue->SymVector.Type);
  } else if (lvalue->Kind == kSymVector) {
    /* only lvalue is of vector type: ERROR ! */
    return FALSE;
  } else if (expr->Kind == kSymVector) {
    /* only expr is of vector type: ERROR ! */
    return FALSE;
  } else {
    return Coercible(expr, lvalue);
  }
}

/*------- ConfigsCompatible ------------------------------------------------*/

BOOLEAN Types_VectorsCompatible(tType vt1, tType vt2)
{
  return ConfigCoercible(vt1, vt2);
}

/*------- ResultType -------------------------------------------------------*/

tType Types_ResultType(INDEX operator, tType operand1, tType operand2)
{
  if (operand1->Kind == kSymVector && operand2->Kind == kSymVector) {
    /* both operands are of vector type: return Vector of scalar result type */
    if (ConfigCoercible(operand1, operand2)) {
      return Symbols_ExtendToVectorType(operand1, Operable(operator,
        operand1->SymVector.Type, operand2->SymVector.Type));
    } else {
      return mSymError( xxConfigIncomp, -1L, -1L );
    } /* endif */
  } else if (operand1->Kind == kSymVector) {
    /* only operand1 is of vector type */
    return Symbols_ExtendToVectorType(operand1, 
      Operable(operator, operand1->SymVector.Type, operand2));
  } else if (operand2->Kind == kSymVector) {
    /* only operand2 is of vector type */
    return Symbols_ExtendToVectorType(operand2, 
      Operable(operator, operand1, operand2->SymVector.Type));
  } else { /* both are scalar types */
    return Operable(operator, operand1, operand2);
  } /* endif */
}

/*------- StdProcResultType ------------------------------------------------*/

tType Types_StdProcResultType( tType proc, tTypes params )
{
  INDEX    argCount = Symbols_Count( params );
  tType    argType1,
           argType2,
           argType3;
  BOOLEAN  isVar1,
           isVar2,
           isVar3;
  BITSET   expectedParams;

  if (StdProcs_ParamCount(proc->SymStdProcType.StdProc, argCount,
    &expectedParams)) {
    switch (argCount) {
    case 0:
      switch (proc->SymStdProcType.StdProc) {
      case eSystemHalt       :
      case eSystemCloseOutput:
      case eSystemCloseInput :
      case eSystemWriteLn    : return TypeVoid;
      case eSystemCardRandom : return TypeCARDINAL;
      case eSystemIntRandom  : return TypeINTEGER;
      case eSystemRealRandom : return TypeREAL;
      case eSystemBoolRandom : return TypeBOOLEAN;
      case eSystemCharRandom : return TypeCHAR;
      case eSystemArgc       : return TypeCARDINAL;
      } /* endswitch case 0 stdproc */
      break; /* case 0 */
    case 1:
      isVar1 = Symbols_Head(params, &argType1);
      if (argType1->Kind == kSymSubrange) {
        argType1 = argType1->SymSubrange.Type;
      } /* endif */
      switch (proc->SymStdProcType.StdProc) {
      case eSystemAbs:
	 if (Symbols_IsIntType(argType1))
	   /* return Symbols_ExtendToVectorType(argType1, TypeCARDINAL);*/
	      return argType1;
	 else if (Symbols_IsRealType(argType1)) 
	      return argType1;
	 break;
      case eSystemCap:
        if (Symbols_IsCharType(argType1)) {
          return Symbols_ExtendToVectorType(argType1, TypeCHAR);
        } /* endif */
        break;
      case eSystemChr:
        if (Symbols_IsIntType(argType1)) {
          return Symbols_ExtendToVectorType(argType1, TypeCHAR);
        } /* endif */
        break;
      case eSystemId:
        if (Symbols_IsVectorType(argType1)) {
          return Symbols_ExtendToVectorType(argType1, TypeCARDINAL);
        } /* endif */
        break;
      case eSystemLen:
      case eSystemRank:
        if (Symbols_IsVectorType(argType1)) {
          return TypeCARDINAL;
        }
        break;
      case eSystemMax:
      case eSystemMin:
        return argType1;
      case eSystemEven:
      case eSystemOdd:
        if (Symbols_IsIntType(argType1) || Symbols_IsCharType(argType1)) {
          return Symbols_ExtendToVectorType(argType1, TypeBOOLEAN);
        } /* endif */
        break;
      case eSystemFloat:
        if (Symbols_IsIntType(argType1)) {
          return Symbols_ExtendToVectorType(argType1, TypeREAL);
        } /* endif */
        break;
      case eSystemHigh:
        if (Symbols_IsArrayType(argType1)) {
          return Symbols_GetIndexType(argType1);
	} /* endif */
        break;
      case eSystemOrd:
        if (Symbols_IsBoolType(argType1) ||
            Symbols_IsIntType(argType1)  ||
            Symbols_IsCharType(argType1) ||
            Symbols_IsEnumType(argType1) ) {
          return Symbols_ExtendToVectorType(argType1, TypeCARDINAL);
        } /* endif */
        break;
      case eSystemTrunc:
      case eSystemCeiling:
      case eSystemFloor:
      case eSystemRound:
        if (Symbols_IsRealType(argType1)) {
          return Symbols_ExtendToVectorType(argType1, TypeIntCard);
        } /* endif */
        break;
      case eSystemSize:
      case eSystemTSize:
        return Symbols_ExtendToVectorType(argType1, TypeIntCard);
      case eSystemAdr:
        return Symbols_ExtendToVectorType(argType1, TypeADDRESS);
      case eSystemDec:
      case eSystemInc:
        if (Symbols_IsIntType(argType1)  ||
            Symbols_IsCharType(argType1) ||
            Symbols_IsEnumType(argType1)) {
          return TypeVoid;
        } /* endif */
        break;
      case eSystemNew:
      case eSystemDispose:
        return TypeVoid;
      case eSystemSqrt:
      case eSystemExponent:
      case eSystemLn:
      case eSystemSin:
      case eSystemArcSin:
      case eSystemCos:
      case eSystemArcCos:
      case eSystemTan:
      case eSystemArcTan:
        if (Symbols_IsRealType(argType1)) {
          return argType1;
        } /* endif */
        break;
      case eSystemCardRandom:
        if (Symbols_IsVectorType(argType1)) {
          return Symbols_ExtendToVectorType(argType1, TypeCARDINAL);
        } /* endif */
        break;
      case eSystemIntRandom:
        if (Symbols_IsVectorType(argType1)) {
          return Symbols_ExtendToVectorType(argType1, TypeINTEGER);
        } /* endif */
        break;
      case eSystemRealRandom:
        if (Symbols_IsVectorType(argType1)) {
          return Symbols_ExtendToVectorType(argType1, TypeREAL);
        } /* endif */
        break;
      case eSystemBoolRandom:
        if (Symbols_IsVectorType(argType1)) {
          return Symbols_ExtendToVectorType(argType1, TypeBOOLEAN);
        } /* endif */
        break;
      case eSystemCharRandom:
        if (Symbols_IsVectorType(argType1)) {
          return Symbols_ExtendToVectorType(argType1, TypeCHAR);
        } /* endif */
        break;
      case eSystemOpenOutput:
      case eSystemOpenInput:
      case eSystemWriteString:
        if (Symbols_IsStringType(argType1)) {
          return TypeVoid;
        } /* endif */
        break;
      case eSystemWrite:
        if (Symbols_IsCharType(argType1)) {
          return TypeVoid;
        } /* endif */
        break;
      case eSystemRead:
        if (Symbols_IsCharType(argType1)) {
          return TypeVoid;
        } /* endif */
        break;
      case eSystemReadInt:
      case eSystemReadCard:
        if (Symbols_IsIntType(argType1)) {
          return TypeVoid;
        } /* endif */
        break;
      case eSystemReadReal:
        if (Symbols_IsRealType(argType1)) {
          return TypeVoid;
        } /* endif */
        break;
      case eSystemReadBoolean:
        if (Symbols_IsBoolType(argType1)) {
          return TypeVoid;
        } /* endif */
        break;
      case eSystemReadString:
        if (Symbols_IsStringType(argType1)) {
          return TypeVoid;
        } /* endif */
        break;
      } /* endswitch case 1 stdproc */
      return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1 );
    case 2:
      isVar1 = Symbols_Head(params, &argType1);
      if (argType1->Kind == kSymSubrange) {
        argType1 = argType1->SymSubrange.Type;
      } /* endif */
      params = Symbols_Tail(params);
      isVar2 = Symbols_Head(params, &argType2);
      if (argType2->Kind == kSymSubrange) {
        argType2 = argType2->SymSubrange.Type;
      } /* endif */
      switch ( proc->SymStdProcType.StdProc ) {
      case eSystemDim:
        if (Symbols_IsVectorType(argType1)) {
          if (Symbols_IsInt(argType2)) {
            return Symbols_ExtendToVectorType(argType1, TypeINTEGER);
	  } else {
            return mSymError(xxArgumentTypeIncomp, 2L,(ULONG) argType2);
	  }
	} else {
          return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
        } /* endif */
      case eSystemLen:
        if (Symbols_IsVectorType(argType1)) {
          if (Symbols_IsInt(argType2)) {
            return TypeCARDINAL;
	  } else {
            return mSymError(xxArgumentTypeIncomp, 2L,(ULONG) argType2);
	  }
	} else {
          return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
        } /* endif */
      case eSystemLower:
      case eSystemUpper:
        if (Symbols_IsVectorType(argType1)) {
          if (Symbols_IsInt(argType2)) {
            return TypeINTEGER;
	  } else {
            return mSymError(xxArgumentTypeIncomp, 2L,(ULONG) argType2);
	  }
	} else {
          return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
        } /* endif */
      case eSystemDim2Id:
        if (Symbols_IsVectorType(argType1)) {
          if (Symbols_IsArrayOfInt(argType2)) {
            return Symbols_ExtendToVectorType(argType2, TypeCARDINAL);
          } else {
            return mSymError(xxArgumentTypeIncomp, 2L,(ULONG) argType2);
	  }
        } else {
          return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
        } /* endif */
      case eSystemVal:
        if (Symbols_IsIntType(argType1)  ||
            Symbols_IsRealType(argType1) ||
            Symbols_IsCharType(argType1) ||
            Symbols_IsEnumType(argType1)) {
          if (Symbols_IsInt(argType2)) {
            return argType1;
          } else {
            return mSymError(xxArgumentTypeIncomp, 2L, (ULONG) argType2);
          } /* endif */
        } else {
          return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
        } /* endif */
      case eSystemDec:
      case eSystemInc:
        if (Symbols_IsIntType(argType1)  ||
            Symbols_IsCharType(argType1) ||
            Symbols_IsEnumType(argType1)) {
          if (Symbols_IsIntType(argType2)  ||
	      Symbols_IsCharType(argType2) ||
	      Symbols_IsEnumType(argType2)) {
            return TypeVoid;
          } else {
            return mSymError(xxArgumentTypeIncomp, 2L, (ULONG) argType2);
          } /* endif */
        } else {
          return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
        } /* endif */
      case eSystemExcl:
      case eSystemIncl:
        if (argType1->Kind == kSymBitset) {
          if (Coercible(TypeIntCard, argType2)) {
            return TypeVoid;
          } else {
            return mSymError(xxArgumentTypeIncomp, 2L, (ULONG) argType2);
          } /* endif */
        } else if (argType1->Kind == kSymSet) {
          if (Coercible(argType1->SymSet.Type, argType2)) {
            return TypeVoid;
          } else {
            return mSymError(xxArgumentTypeIncomp, 2L, (ULONG) argType2);
          } /* endif */
        } else {
          return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
        } /* endif */
      case eSystemArcTan2:
        if (argType1->Kind == kSymReal) { /* Real */
          if (Symbols_IsRealType(argType2)) { /* Vector or Real*/
            return argType2; /* Vector or Real */
          } else {
            return mSymError( xxArgumentTypeIncomp, 2L, (ULONG) argType2 );
          }
        } else {
          if (Symbols_IsRealType(argType1)) { /* Vector */
            if (Symbols_IsRealType(argType2)) { /* Vector or Real */
              return argType1; /* Vector */
            } else {
              return mSymError(xxArgumentTypeIncomp, 2L, (ULONG) argType2);
            }
          } else {
            return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
          }
	} /* endif */
      case eSystemWriteCard:
      case eSystemWriteInt:
        if (Symbols_IsIntType(argType1)) {
          if (Symbols_IsInt(argType2)) {
            return TypeVoid;
          } else {
            return mSymError(xxArgumentTypeIncomp, 2L, (ULONG) argType2);
          } /* endif */
        } else {
          return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
        } /* endif */
      case eSystemWriteReal:
        if (Symbols_IsRealType(argType1)) {
          if (Symbols_IsInt(argType2)) {
            return TypeVoid;
          } else {
            return mSymError( xxArgumentTypeIncomp, 2L, (ULONG) argType2 );
          } /* endif */
        } else {
           return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
        } /* endif */
      case eSystemWriteBoolean:
        if (Symbols_IsBoolType(argType1)) {
          if (Symbols_IsInt(argType2)) {
            return TypeVoid;
          } else {
            return mSymError(xxArgumentTypeIncomp, 2L, (ULONG) argType2);
          } /* endif */
        } else {
           return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
        } /* endif */
        break;
      case eSystemArgv:
        if (Coercible(argType1, TypeCARDINAL)) {
          if (argType2->Kind == kSymArray &&
              argType2->SymArray.ElemType->Kind == kSymChar) {
            return TypeVoid;
          } else {
            return mSymError(xxArgumentTypeIncomp, 2L, (ULONG) argType2);
          } /* endif */
        } else {
          return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
        } /* endif */
      } /* endswitch case 2 stdproc */
      break; /* case 2 */
    case 3:
      isVar1 = Symbols_Head( params, &argType1 );
      if (argType1->Kind == kSymSubrange) {
        argType1 = argType1->SymSubrange.Type;
      } /* endif */
      params = Symbols_Tail(params);
      isVar2 = Symbols_Head(params, &argType2);
      if (argType2->Kind == kSymSubrange) {
        argType2 = argType2->SymSubrange.Type;
      } /* endif */
      params = Symbols_Tail(params);
      isVar3 = Symbols_Head(params, &argType3);
      if (argType3->Kind == kSymSubrange) {
        argType3 = argType3->SymSubrange.Type;
      } /* endif */
      switch ( proc->SymStdProcType.StdProc ) {
      case eSystemWriteFixPt:
        if (Symbols_IsRealType(argType1)) {
          if (Symbols_IsInt(argType2)) {
            if (Symbols_IsInt(argType3)) {
              return TypeVoid;
            } else {
              return mSymError(xxArgumentTypeIncomp, 3L, (ULONG) argType3);
            } /* endif */
          } else {
            return mSymError(xxArgumentTypeIncomp, 2L, (ULONG) argType2);
          } /* endif */
        } else {
          return mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
        } /* endif */
      case eSystemId2Dim:
        if (Symbols_IsVectorType( argType1 ) ) {
          if (argType2->Kind == kSymVector && argType3->Kind == kSymVector) {
            if (Symbols_IsInt(argType2->SymVector.Type)) {
              if (Symbols_IsArrayOfInt(argType3->SymVector.Type)) {
                return TypeVoid;
              } else {
                return mSymError(xxArgumentTypeIncomp, 3L, (ULONG) argType3);
              }
            } else {
              return mSymError(xxArgumentTypeIncomp, 2L, (ULONG) argType2);
            }
          } else {
            if (Symbols_IsInt(argType2)) {
              if (Symbols_IsArrayOfInt(argType3)) {
                return TypeVoid;
              } else {
                return mSymError(xxArgumentTypeIncomp, 3L, (ULONG) argType3);
              }
            } else {
              return mSymError(xxArgumentTypeIncomp, 2L, (ULONG) argType2);
            }
          }
        } else {
          mSymError(xxArgumentTypeIncomp, 1L, (ULONG) argType1);
	}
      } /* endswitch case 3 stdproc*/
      break; /* case 3 */
    } /* endswitch */
  } /* endif ParamCount */
  return mSymError( xxIllegalArgCount, argCount, expectedParams );
}

/*------- Types_WriteType --------------------------------------------------*/

CHAR *Types_WriteType( tType type, CHAR *buffer )
{
  char localBuffer[1024];

  switch ( type->Kind ) {
  case kSymNoType     : strcpy(buffer, "'Unknown Type'");     break;
  case kSymInteger    : strcpy(buffer, "INTEGER");            break;
  case kSymCardinal   : strcpy(buffer, "CARDINAL");           break;
  case kSymReal       : strcpy(buffer, "REAL");               break;
  case kSymBoolean    : strcpy(buffer, "BOOLEAN");            break;
  case kSymChar       : strcpy(buffer, "CHAR");               break;
  case kSymBitset     : strcpy(buffer, "BITSET");             break;
  case kSymProc       : strcpy(buffer, "PROC");               break;
  case kSymWord       : strcpy(buffer, "WORD");               break;
  case kSymAddress    : strcpy(buffer, "ADDRESS");            break;
  case kSymIntCard    : strcpy(buffer, "'Integer'");          break;
  case kSymNil        : strcpy(buffer, "NIL");                break;
  case kSymString     : strcpy(buffer, "'String'");           break;
  case kSymStringChar : strcpy(buffer, "'String Character'"); break;
  case kSymVoid       : strcpy(buffer, "'void'");             break;
  case kSymStdProcType: strcpy(buffer, "Standard procedure"); break;
  case kSymOpaqueType:
    strcpy(buffer, "opaque type '");
    GetString(Symbols_GetTypeIdent(type), localBuffer);
    strcat(buffer, localBuffer);
    strcat(buffer, "'");
    break;
  case kSymVector:
    if (type->SymVector.IsOpen) {
      strcpy(buffer, "OPEN VECTOR OF " );
    } else {
      strcpy(buffer, "VECTOR '" );
      GetString(type->SymVector.Config->SymObject.Ident, localBuffer);
      strcat(buffer, localBuffer);
      strcat(buffer, "' OF ");
    } /* endif */
    strcat(buffer, Types_WriteType(Symbols_GetVectorBaseType(type),
      localBuffer));
    break;
  case kSymArray:
    if (type->SymArray.IsOpen) {
      strcpy(buffer, "OPEN ARRAY OF ");
    } else {
      strcpy(buffer, "ARRAY OF ");
    } /* endif */
    strcat(buffer, Types_WriteType(Symbols_GetElemType(type), localBuffer));
    break;
  case kSymEnumeration:
    strcpy(buffer, "ENUMERATION '");
    GetString( Symbols_GetTypeIdent(type), localBuffer);
    strcat(buffer, localBuffer);
    strcat(buffer, "'");
    break;
  case kSymPointer:
    strcpy(buffer, "POINTER TO ");
    strcat(buffer, Types_WriteType(Symbols_GetTargetType(type), localBuffer));
    break;
  case kSymRecord:
    strcpy(buffer, "RECORD '");
    GetString( Symbols_GetTypeIdent(type), localBuffer);
    strcat(buffer, localBuffer );
    strcat(buffer, "'" );
    break;
  case kSymSet:
    strcpy(buffer, "SET OF ");
    strcat(buffer, Types_WriteType(Symbols_GetSetBaseType(type), localBuffer));
    break;
  case kSymSubrange:
    strcpy(buffer, "SUBRANGE OF ");
    strcat(buffer, Types_WriteType(Symbols_GetSubrangeBaseType(type),
      localBuffer));
    break;
  case kSymProcType:
     { 
	  tType types;

	  strcpy(buffer, "PROCEDURE(");
	  types = type->SymProcType.Types;
	  while(types!=NoTypes)
	  {
	       strcat(buffer, Types_WriteType(types->SymTypes.Type, localBuffer));
	       types = types->SymTypes.Next;
	       if (types != NoTypes) strcat(buffer, ", ");
	  }
	  strcat(buffer, ")");
	  if ((type->SymProcType.Type != NoType) &&
	      (type->SymProcType.Type->Kind != kSymVoid))
						     
	  {
	       strcat(buffer,": ");
	       strcat(buffer,Types_WriteType(type->SymProcType.Type, localBuffer));
	  }
	  break;
     }
  default          :
     *buffer = '\0';
  } /* endswitch */
  return( buffer );
}

/*------- TypeSize ---------------------------------------------------------*/

ULONG Types_TypeSize( tType type )
{
   ULONG dummy;

   return TypeSize( type, &dummy );
}

/*------- GetLwb -----------------------------------------------------------*/

BOOLEAN Types_GetLwb( tType type, tValue *value )
{
  switch ( type->Kind ) {
  case kSymInteger : *value = Values_MinInteger;     break;
  case kSymCardinal: *value = Values_Zero;           break;
  case kSymReal    : *value = Values_MinReal;        break;
  case kSymBoolean : *value = Values_False;          break;
  case kSymChar    : *value = Values_MinChar;        break;
  case kSymEnumeration:
    value->Kind = eEnum;
    value->value.enumValue =
      Symbols_GetLiteral(type->SymEnumeration.Objects, 0);
    break;
  case kSymSubrange: *value = type->SymSubrange.Lwb; break;
  default          : *value = Values_Error;          return FALSE;
  } /* endswitch */
  return TRUE;
}

/*------- GetUpb -----------------------------------------------------------*/

BOOLEAN Types_GetUpb( tType type, tValue *value )
{
  switch ( type->Kind ) {
  case kSymInteger : *value = Values_MaxInteger;     break;
  case kSymCardinal: *value = Values_MaxCardinal;    break;
  case kSymReal    : *value = Values_MaxReal;        break;
  case kSymBoolean : *value = Values_True;           break;
  case kSymChar    : *value = Values_MaxChar;        break;
  case kSymEnumeration:
    value->Kind = eEnum;
    value->value.enumValue = Symbols_GetLiteral(type->SymEnumeration.Objects,
      type->SymEnumeration.MaxValue );
    break;
  case kSymSubrange: *value = type->SymSubrange.Upb; break;
  default          : *value = Values_Error;          return FALSE;
  } /* endswitch */
  return TRUE;
}

/*------- Cast -------------------------------------------------------------*/

BOOLEAN Types_Cast(INDEX Operator, tType t1, tType t2 )
{
  if (t1->Kind == kSymSubrange ) {
    t1 = t1->SymSubrange.Type;
  }
  if (t2->Kind == kSymSubrange ) {
    t2 = t2->SymSubrange.Type;
  }

  if (t1->Kind == kSymOpaqueType && t1->SymOpaqueType.Type != NoType &&
      t2->Kind == kSymOpaqueType && t2->SymOpaqueType.Type != NoType ) {
    t1 = t1->SymOpaqueType.Type;
    t2 = t2->SymOpaqueType.Type;
  }

  switch (Operator) {
  case Code_Assign:
    return (t1 == TypeADDRESS && t2 != TypeADDRESS) ||
           (t2 == TypeADDRESS && t1 != TypeADDRESS) ||
           (t1->Kind == kSymOpaqueType && t2->Kind != kSymOpaqueType) ||
           (t2->Kind == kSymOpaqueType && t1->Kind != kSymOpaqueType);
  case Code_PassAddress:
    return( (t1 != t2) && Options_IsOptionSet(Options_eCasts) );
  case Code_PassValue:
    if (t2 == TypeStringChar && t1 == TypeCHAR) {
      return FALSE;
    } else {
      return (t1 != t2) && Options_IsOptionSet(Options_eCasts);
    }
  default: return FALSE;
  }
}

void Types_Init(void)
{
  if (!isInitialised) {
    isInitialised = TRUE;
    BigSet_Init();
    Values_Init();
  } /* endif */
}

