/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  UniqueIds.c
 * desc  :  generate unique ids, resolve name clashes
 * author:  lampke
 *
 ************************************************************************/

#include "DynArray.h"			  /* Reuse-Library */
#include "Idents.h"			  /* Reuse-Library */
#include <string.h>
#include "BigSet.h"
#include "Errors.h"
#include "UniqueIds.h"
#include "FileIO.h"

static BOOLEAN isInitialised = FALSE;

/*
 * local constants
 */

#define UniqueIds_InitialSize  4
#define MaxKeyword 121

/*
 * local variables
 */

static tIdent KeywordTable[MaxKeyword]; /* keywords */

static tBigSet ModulaIds,    /* keywords and all Parallaxis identifiers
                                (used in the generated C program(s) ) */
               PervasiveIds; /* keywords, type and constant identifiers */

static long StackPtr, StackSize;
/* IdStack[0] procedure identifiers, global variable identifiers,
 * and variable identifiers declared in local modules
 * IdStack[StackPtr] local variable identifiers of current proc.
 */

static tBigSet *IdStack;

static long KeywordCount;

static unsigned long SetSize;

/*
 * local function prototypes
 */
static void DefineKeyword( char *Name );

/*
 * local function implementation
 */
/*------- DefineKeywords ---------------------------------------------------*/

static void DefineKeyword( char *Name )
{
  KeywordCount++;
  if ( KeywordCount > MaxKeyword ) {
    fprintf( stderr, "UniqueIds.DefineKeywords: assertion violation\n");
    (void) FileIO_CloseAll();
    exit(1);
  }
  KeywordTable[KeywordCount] = MakeIdent( Name, strlen( Name ) );
}

/*
 * Global variables
 */

/*
 * global function implementation
 */
/*------- NameConflict -----------------------------------------------------*/

BOOLEAN UniqueIds_NameConflict( tIdents Idents, tIdentClass IdentClass,
  tIdent Ident )
/* Check if a name conflict with an identifier which has been
 * already declared would result from using the identifier
 * 'Ident' of class 'IdentClass' in the generated C program(s)
 */
{
  if ( Ident > SetSize ) {
    Message( "UniqueIds.NameConflict", xxFatal, NoPosition );
  } /* endif */
  switch ( IdentClass ) {
  case eFields:
    return BigSet_IsElement( &PervasiveIds, Ident );
  case eConsts:
  case eTypes:
    return BigSet_IsElement( &ModulaIds, Ident );
  case eProcs:
    return( BigSet_IsElement( &PervasiveIds, Ident ) ||
            BigSet_IsElement( &IdStack[0], Ident ) );
  case eVars:
    return( BigSet_IsElement( &PervasiveIds, Ident ) ||
            BigSet_IsElement( &IdStack[StackPtr], Ident ) );
  case eModuleVars:
    return( BigSet_IsElement( &PervasiveIds, Ident ) ||
            BigSet_IsElement( &IdStack[StackPtr], Ident ) ||
            BigSet_IsElement( &IdStack[0], Ident ) );
  case eKeywords:
    return TRUE;
  } /* endswitch */
  return FALSE;
}

/*------- DeclareIdent -----------------------------------------------------*/

tIdents UniqueIds_DeclareIdent( tIdents Idents, tIdentClass IdentClass,
  tIdent Ident )
/* Declare the identifier 'ident' of class 'identClass' as used
 * in the generated C program(s)
 */
{
  if ( Ident > SetSize ) {
    Message( "UniqueIds.NameConflict", xxFatal, NoPosition );
  } /* endif */
  BigSet_Include( &ModulaIds, Ident );
  switch ( IdentClass ) {
  case eFields:
    break;
  case eKeywords:
  case eConsts:
  case eTypes:
    BigSet_Include( &PervasiveIds, Ident );
    break;
  case eProcs:
    BigSet_Include( &IdStack[0], Ident );
    break;
  case eVars:
    BigSet_Include( &IdStack[StackPtr], Ident );
    break;
  case eModuleVars:
    BigSet_Include( &IdStack[StackPtr], Ident );
    BigSet_Include( &IdStack[0], Ident );
    break;
  } /* endswitch */
  return Idents;
}

/*------- Begin ------------------------------------------------------------*/

void UniqueIds_Begin( void )
/* Initialize module UniqueIds
 */
{
  long i;
  tIdents Idents;

  SetSize = MaxIdent();
  BigSet_MakeSet( &ModulaIds, SetSize );
  BigSet_MakeSet( &PervasiveIds, SetSize );

  StackPtr = 0;
  StackSize = UniqueIds_InitialSize;
  MakeArray( (char **) &IdStack, &StackSize, sizeof(tBigSet) );
  for (i=0; i<StackSize; i++) {
    BigSet_MakeSet( &IdStack[i], SetSize );
  } /* endfor */

  BigSet_AssignEmpty( &ModulaIds );
  BigSet_AssignEmpty( &PervasiveIds );
  BigSet_AssignEmpty( &IdStack[StackPtr] );

  /* declare keywords
   */
  for (i=0; i<=KeywordCount; i++) {
    Idents = UniqueIds_DeclareIdent( Idents, eKeywords,
      KeywordTable[i] );
  } /* endfor */
}

/*------- Close ------------------------------------------------------------*/

void UniqueIds_Close( void )
/* Finalize module UniqueIds
 */
{
   long i;

   BigSet_ReleaseSet( &ModulaIds );
   BigSet_ReleaseSet( &PervasiveIds );

   for (i=0; i<StackSize; i++) {
      BigSet_ReleaseSet( &IdStack[i] );
   } /* endfor */
   ReleaseArray( (char **) &IdStack, &StackSize, sizeof(tBigSet) );
}

/*------- EnterProc --------------------------------------------------------*/

tIdents UniqueIds_EnterProc( tIdents Idents )
/* Enter local declarations of a procedure
 */
{
  long OldSize, i;

  StackPtr++;

  if ( StackPtr >= StackSize ) {
      OldSize = StackSize;
      ExtendArray( (char **) &IdStack, &StackSize, sizeof(tBigSet) );
      for (i=OldSize; i<StackSize; i++) {
         BigSet_MakeSet( &IdStack[i], SetSize );
      } /* endfor */
  } /* endif */

  BigSet_AssignEmpty( &IdStack[StackPtr] );

  return Idents;
}

/*------- LeaveProc --------------------------------------------------------*/

tIdents UniqueIds_LeaveProc( tIdents Idents )
/* Leave local declarations of a procedure
 */
{
  StackPtr--;

  return Idents;
}

void UniqueIds_Init( void )
{
  if (!isInitialised) {
    isInitialised = TRUE;

    KeywordCount = -1;

    DefineKeyword( "auto" );
    DefineKeyword( "break" );
    DefineKeyword( "case" );
    DefineKeyword( "char" );
    DefineKeyword( "const" );
    DefineKeyword( "continue" );
    DefineKeyword( "default" );
    DefineKeyword( "define" );
    DefineKeyword( "do" );
    DefineKeyword( "double" );
    DefineKeyword( "else" );
    DefineKeyword( "endif" );
    DefineKeyword( "enum" );
    DefineKeyword( "extern" );
    DefineKeyword( "float" );
    DefineKeyword( "for" );
    DefineKeyword( "goto" );
    DefineKeyword( "if" );
    DefineKeyword( "ifdef" );
    DefineKeyword( "ifndef" );
    DefineKeyword( "include" );
    DefineKeyword( "int" );
    DefineKeyword( "line" );
    DefineKeyword( "long" );
    DefineKeyword( "main" );
    DefineKeyword( "register" );
    DefineKeyword( "return" );
    DefineKeyword( "short" );
    DefineKeyword( "signed" );
    DefineKeyword( "sizeof" );
    DefineKeyword( "static" );
    DefineKeyword( "struct" );
    DefineKeyword( "switch" );
    DefineKeyword( "typedef" );
    DefineKeyword( "undef" );
    DefineKeyword( "union" );
    DefineKeyword( "unsigned" );
    DefineKeyword( "void" );
    DefineKeyword( "volatile" );
    DefineKeyword( "while" );

   /* standard constants
    */
    DefineKeyword( "TRUE" );
    DefineKeyword( "FALSE" );
    DefineKeyword( "NIL" );
    DefineKeyword( "EOL" );
    DefineKeyword( "pi" );

   /* basic types
    */
    DefineKeyword( "INTEGER" );
    DefineKeyword( "CARDINAL" );
    DefineKeyword( "BOOLEAN" );
    DefineKeyword( "CHAR" );
    DefineKeyword( "REAL" );

   /* standard types
    */
    DefineKeyword( "BITSET" );
    DefineKeyword( "PROC" );
    DefineKeyword( "BYTE" );
    DefineKeyword( "WORD" );
    DefineKeyword( "ADDRESS" );

   /* parallaxis types
    */
    DefineKeyword( "CONFIGURATION" );

   /* standard functions
    */
    DefineKeyword( "ABS" );
    DefineKeyword( "ABSI" );
    DefineKeyword( "ABSC" );
    DefineKeyword( "ABSR" );
    DefineKeyword( "CAP" );
    DefineKeyword( "CHR" );
    DefineKeyword( "FLOAT" );
    DefineKeyword( "ORD" );
    DefineKeyword( "TRUNC" );
    DefineKeyword( "VAL" );
    DefineKeyword( "ODD" );

   /* parallaxis standard functions
    */
    DefineKeyword( "argc" );
    DefineKeyword( "LEN" );
    DefineKeyword( "LEN1" );
    DefineKeyword( "LOWER" );
    DefineKeyword( "RANK" );
    DefineKeyword( "UPPER" );

   /* standard procedures
    */
    DefineKeyword( "INC" );
    DefineKeyword( "INC1" );
    DefineKeyword( "DEC" );
    DefineKeyword( "DEC1" );
    DefineKeyword( "EXCL" );
    DefineKeyword( "INCL" );
    DefineKeyword( "ADR" );
    DefineKeyword( "ADR1" );

   /* parallaxis standard procedures
    */
    DefineKeyword( "argv" );

   /* compiler types
    */
    DefineKeyword( "OPAQUE" );

   /* standard variables
    */
    DefineKeyword( "termCH" );
    DefineKeyword( "Done" );

   /* compiler functions
    */
    DefineKeyword( "CaseError" );
    DefineKeyword( "ReturnError" );

   /* flag for conditional compilation
    */
    DefineKeyword( "StackAlloc" );

   /* (some) compiler generated identifiers
    */
    DefineKeyword( "A" );
    DefineKeyword( "dummy" );

   /* parallaxis idents, SYSTEM_C.h */
    DefineKeyword("PE");
    DefineKeyword("BOOLEAN_ELEM");
    DefineKeyword("BOOLEAN_ELEM_PTR");
    DefineKeyword("ACTIVE_STACK");
    DefineKeyword("ACTIVE_STACK_PTR");
    DefineKeyword("ELSE_STACK");
    DefineKeyword("ELSE_STACK_PTR");
    DefineKeyword("LINK");
    DefineKeyword("PARAMETER_LINK");
    DefineKeyword("CONNECTION");
    DefineKeyword("PARAMETER_CONNECTION");
    DefineKeyword("GENERIC_CONNECTIONS");
    DefineKeyword("GENERIC_PARAMETER_CONNECTIONS");
    DefineKeyword("CONFIGURATION");

    DefineKeyword("srandom");
    DefineKeyword("random");
    DefineKeyword("printf");
    DefineKeyword("fprintf");
    DefineKeyword("strcat");
    DefineKeyword("strcmp");
    DefineKeyword("strcpy");

   /* global identifiers in SYSTEM_PVM.h
    */
    DefineKeyword( "my_tid" );
    DefineKeyword( "parent_tid" );
    DefineKeyword( "io_tid" );
    DefineKeyword( "n_tasks" );
    DefineKeyword( "slavetids" );
    DefineKeyword( "me" );
    DefineKeyword( "PE" );
    DefineKeyword( "message_tag" );
    DefineKeyword( "SOURCE_ARRAY" );
    DefineKeyword( "DESTINATION_ARRAY" );
    DefineKeyword( "CONFIG_1" );
  }
}
