/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  GetDefMods.c
 * desc  :  read in all definition modules which the current compilation
 *          unit depends on an dadd them to syntaxtree
 * author:  lampke
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "Idents.h"			  /* Reuse-Library */
#include "AssocTab.h"
#include "GeneralIO.h"
#include "Scanner.h"
#include "Parser.h"
#include "Errors.h"
#include "Positions.h"
#include "Tree.h"
#include "Symbols.h"
#include "GetDefMods.h"

static BOOLEAN isInitialised = FALSE;

/*
 * local types
 */
typedef enum {
  eNotAnalyzed, eAnalyzed, eError
} tMarkClass, *tMarkClassPtr;

/* An association table is used to keep track of each definition
 * module, whether it has been read in and whether it's import
 * statements have been completely analyzed:
 *
 *      Mark = GetAssoc( ModuleName );
 *
 *      if ( Mark == NULL) {
 *        definition module 'ModuleName' not read in
 *      } else if ( *Mark == eNotAnalyzed ) {
 *        definition module 'ModuleName' read in, but it's import
 *        statements are not completely analyzed
 *      } else if ( *Mark == eAnalyzed ) {
 *        definition module 'ModuleName' read in and it's import
 *        statements have been completely analyzed
 *      } else if ( *Mark == eError ) {
 *        definition module 'ModuleName' not found or
 *        definition module 'ModuleName' contains syntactic errors or
 *        module 'ModuleName' is not a definition modules
 *      }
 */

/*
 * local variables
 */
static tTree Tree;

/*
 * local function prototypes
 */
static tMarkClassPtr MakeMark( tMarkClass Class );
static void TraverseImports( tTree Module );
static void GetDefinitionModule( tIdent ModuleName, tPosition Pos );

/*
 * local function implementation
 */
/*************************************************************************
 * name :  MakeMark
 * in   :
 * out  :
 ************************************************************************/
static tMarkClassPtr MakeMark(tMarkClass Class)
{
  tMarkClassPtr markPtr;

  markPtr = (tMarkClassPtr) calloc(1, sizeof(tMarkClass));
  *markPtr = Class;
  return markPtr;
}

/*************************************************************************
 * name :  TraverseImports
 * in   :
 * out  :
 ************************************************************************/
static void TraverseImports(tTree Module)
{
  tTree Import, Ids;
  tMarkClassPtr markPtr;

  switch (Module->Kind) {
  case kDefMod:  Import = Module->DefMod.Imports;  break;
  case kImplMod: Import = Module->ImplMod.Imports; break;
  } /* endswitch */
  while (Import->Kind != kNoImport) {
    if ( Import->Kind == kFrom ) { /* FROM mod IMPORT ids; */
      AssocTab_GetAssoc(Import->From.Ident, (void **) &markPtr);
      if ( markPtr == NULL ) { /* module not read in */
        GetDefinitionModule(Import->From.Ident, Import->From.Pos);
      } else if ( *markPtr == eNotAnalyzed ) { /* cyclic dependence */
        ErrorMessage(xxCyclicDefMods, xxFatal, Import->From.Pos);
      }
    } else { /* IMPORT ids; */
      Ids = Import->Objects.ImpIds;
      while (Ids->Kind != kNoImpId) {
        AssocTab_GetAssoc(Ids->ImpId.Ident, (void **) &markPtr);
        if (markPtr == NULL) { /* module not read in */
          GetDefinitionModule( Ids->ImpId.Ident, Ids->ImpId.Pos );
        } else if (*markPtr == eNotAnalyzed) { /* cyclic dependence */
          ErrorMessage(xxCyclicDefMods, xxFatal, Ids->ImpId.Pos);
        }
        Ids = Ids->ImpId.Next;
      }
    }
    Import = Import->Import.Next;
  }
  if (Module->Kind != kImplMod) {
    /* note that the Import statements of definition module
     * 'ModuleName' have been completely analyzed
     */
    AssocTab_GetAssoc(Module->DefMod.Ident, (void **) &markPtr);
    if (markPtr == NULL) {
      markPtr = MakeMark(eError);
      ErrorMessageI(xxFileNameMismatch, xxError, Module->DefMod.Pos,
        xxIdent, &Module->DefMod.Ident);
    } else {
      *markPtr = eAnalyzed;
    } /* endif */
    AssocTab_PutAssoc( Module->DefMod.Ident, markPtr );
  }
  Module->CompUnit.Next = Tree;
  Tree = Module;
}

/*************************************************************************
 * name :  GetDefinitionModule
 * in   :
 * out  :
 ************************************************************************/
static void GetDefinitionModule(tIdent ModuleName, tPosition Pos)
{
  char fileName[256];

  if (GeneralIO_CheckDefFile(ModuleName, fileName)) {
    BeginFile(fileName);
    (void) Parser();
    if (!ErrorGetCount()) {
      if (TreeRoot->Kind == kImplMod) {
        /* not a definition module */
        AssocTab_PutAssoc(ModuleName, MakeMark(eError));
        ErrorMessageI(xxModNotFound, xxError, Pos, xxIdent, &ModuleName);
      } else {
        /* mark definition module as read in and
         * analyze it's import statements
         */
        AssocTab_PutAssoc(ModuleName, MakeMark(eNotAnalyzed));
        TraverseImports(TreeRoot);
      }
    } else {
      /* syntactic errors */
      AssocTab_PutAssoc(ModuleName, MakeMark(eError));
    } /* endif */
  } else {
    /* module not found */
    AssocTab_PutAssoc(ModuleName, MakeMark(eError));
    ErrorMessageI(xxModNotFound, xxError, Pos, xxIdent, &ModuleName);
  } /* endif */
}

/*
 * global function implementation
 */
/*------- GetDefModules ----------------------------------------------------*/

tTree GetDefMods_GetDefModules(tTree CompUnit)
{
  Tree = mNoCompUnit();
  AssocTab_Begin();
  /* mark pseudo module SYSTEM as read in and analyzed */
  AssocTab_PutAssoc( IdentSYSTEM, MakeMark( eAnalyzed ) );
  if (CompUnit->Kind == kImplMod) {
    if (CompUnit->ImplMod.Kind == ModTypeImp) {
      /* get corresponding definition module */
      GetDefinitionModule(CompUnit->ImplMod.Ident, CompUnit->ImplMod.Pos);
    }
    /* analyze import statements of current compilation unit */
    TraverseImports(CompUnit);
  } else {
     /* mark definition module as read in and analyze it's import statements */
     AssocTab_PutAssoc(CompUnit->DefMod.Ident, MakeMark(eNotAnalyzed));
     TraverseImports(CompUnit);
  }
  AssocTab_Close();
  return mROOT(ReverseTree(Tree));
}

void GetDefMods_Init( void )
{
  if (!isInitialised) {
    isInitialised = TRUE;
    GeneralIO_Init();
    Errors_Init();
    Positions_Init();
  } /* endif */
}

