#define DATE "Date: Wed Jan  8 14:44:46 1997"
/********************************************************************************
 *  Projectname:	Parallaxis-III
 *  Filename:		p3.c
 *  Version:		0.4
 *  Last change:	Wed Jan  8 14:44:46 1997
 *  Author:  		Hartmut Keller
 *  Status:		working version (no MPL-compilation)
 ********************************************************************************
 *  IMPORTANT:          This source text requires a compiler capable of
 *                      assigning a whole structure (e.g. ANSI-C, gcc). If
 *                      a and b are structures of the same type, then b=a; 
 *                      has to copy the whole content from a to b.
 ********************************************************************************
 *
 *  Description:
 *  ------------
 *  This program is a user interface program for the Parallaxis compiler p3c. It
 *  provides similar features as gcc does for the GNU C-compiler. That's:
 *
 *   - compile several files in one p3 call
 *   - call and control the appropriate compiler steps according to the file
 *     extensions
 *   - process the overall options and provide the single compiler steps
 *     with their own correct options
 *
 *
 *  Errors:
 *  -------
 *  - option -k does only work for options, you can't quote a filename
 *  - unknown options with an argument aren't passed together with their
 *    argument. The argument is mistaken as filename. You have to quote.
 *  - "p3 xyz"   with only a file xyz.pd results in a call to gcc with xyz. 
 *  - Option -c is passed to the backend compiler. gcc gives a warning about 
 *    unused Libraries in the call generated by p3. --> -c should be
 *    recognized by p3 and the libs should be omitted then.
 *
 *  Version history:
 *  ----------------
 *   0.1:     First working version.
 *   0.2:     Option -o added, option -vvv removed, changed meaning
 *            of -vv and -v
 *   0.3:     Added "-lm" to STANDARDLIB. The backend compiler is only
 *            called if there are files to compile (e.g. if you only give 
 *            .pd-Files on the commandline, no backend compiler is called).
 *   0.4:     General redesign of the code. The command line for the called
 *            compilers is now generated as an argument list, not as a
 *            string anymore. Several enhancements were added. Some changes
 *            are visible to the user:
 *            - Option -r changed to "Don't generate range checks".
 *            - Option -c now recognized by p3, Libraries are only set in
 *              the backend compiler commandline if no -c is given.
 *            - New environment variable P3OPT; variable P3MPLCC and option
 *              -mplcc removed.
 *	      - '@' anywhere in P3INC or P3LIB includes the default paths in
 *              the list at that place. Pathnames may be seperated by one of
 *              `,', `;', ` ', or `:' (was only `:').
 *            - Pathnames may contain environment variables which are
 *              replaced (e.g. setenv P3INC '@:$PVM_ROOT/include')
 *            - The check for ".pd" and ".pm" files takes also place if the
 *              filename ends in ".". This allows filename completion in
 *              Shells (e.g. tcsh). 
 *            - Lots of comments added.
 *            - Site specific settings are exported to a file "p3config.h"
 *   0.5:     Report an error when compiler process cannot be started (e.g.
 *            if executable cannot be found)
 *
 *   at work:
 *   --------
 *   0.6: 
 *            Delete .px-files before compiling
 *            Create all intermediate files in the TEMP-directory
 *
 *
 *
 *   Ideas and things to do:
 *   -----------------------
 *  - delete .c-files created from .pm-files if option -p is not set
 *  - add options to allow usage of alloca instead of malloc in generated
 *    code. Probably more optimization switches.
 *  - add MPL-compilation pass
 *  - Change in options layout due to extra MPL compiler pass in large Model
 *    (topic postponed)
 *  - add Options -pd, -pm to force a file to be Definition/Implementation Module
 *    even if the extension doesn't match. (e.g. -pm file.par)
 *  - change man-page and Makefile (include P3OPT; new '@'-feature of
 *    path-lists; option -c; last of {-C, -MPL, -PVM} wins; $-feature of
 *    paths, keywords: "not recursive", "no further check for
 *    cycles or other problems", "undefined vars end up as empty string")
 *  - add option -X for X-Windows-support (libraries, includes)
 ********************************************************************************/

#if ((!defined(C_VERSION)) && (!defined(PVM_VERSION)) && (!defined(MPL_VERSION)))
#define PVM_VERSION
#define C_VERSION
#undef  MPL_VERSION
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <time.h>			  /* clock() */
#include <sys/wait.h>			  /* wait() */
#include <unistd.h>			  /* fork() */
#include <ctype.h>			  /* isalnum() */
#include "language.h"		/* for Errormessages in different Languages */
#define TRUE 1				  /* Just to be sure */
#define FALSE 0

/****************************************************************************
 *  Version of this program
 ****************************************************************************/

#define VERSION "V0.5"

/*============================================================================*/

/*************** START OF CONFIGURATION PART ********************************/

/* ----- User definable settings ----- */
#include "p3config.h"

/* ----- Compiler dependant settings ----- */

#define EXT ".c"			  /* extension for C-programs */

#ifdef PVM_VERSION
#define EXT_PVM ".c"			  /* extension for PVM programs */
#endif

#ifdef MPL_VERSION
#define EXT_MPL ".m"			  /* extension for MPL programs */
#endif

/* Name of this program if no argv[0] available (Usually argv[0] *is* */
/* available; just to be sure) */ 
#ifndef P3NAME
#define P3NAME "p3"
#endif

/* Parallaxis file name extension used for definition modules */
#ifndef PDEXT
#define PDEXT ".pd"
#endif

/* Parallaxis file name extension used for program/implementation modules */
#ifndef PMEXT
#define PMEXT ".pm"
#endif

/* Parallaxis file name extension used for crossreference files */
#ifndef PXEXT
#define PXEXT ".px"
#endif

/* Definitions for p3c-compiler switches. The options are usually included */
/* in the Parallaxis compiler call if the appropriate flag with the same */
/* name (stripped off the leading "s") is TRUE. */

#define sOPTcasts   "-c"
#define sOPTverbose "-v"
#define sOPTdebug   "-g"
#define sOPTheaders "-i"
#define sOPTmemory  "-m"
#define sOPTrchecks "-r"
#define sOPTtime    "-t"
#define sOPTnowarnings "-w"
#define sOPTindent "-I"
#define sOPTpath   "-P"
				
#ifdef MPL_VERSION			  /* The MPL-version has an extra option */
#define sOPTsmall   "-s"
#endif /* MPL_VERSION */	
				
#ifdef DEVELOP				  /* During development there are some */
#define sOPT_Zw "-Zw"			  /* additional options due to */
#define sOPT_Zs "-Zs"			  /* debugging purposes */
#define sOPT_Zq "-Zq"
#define sOPT_Zc "-Zc"
#define sOPT_Z1 "-Z1"
#define sOPT_Z2 "-Z2"
#endif /* DEVELOP */		



/* The following variables default to FALSE. If the appropriate option is */
/* given, the variable gets TRUE. */
int OPTindent = -1;			  /* -1: no "-indent", else value */

static int OPTcasts      = FALSE;	  /* "-casts" */
static int OPTnolink     = FALSE;	  /* "-c" */
static int OPTp3only     = FALSE;	  /* "-p" */
static int OPTversion    = FALSE;	  /* "-v" or "-vv" */
static int OPTverbose    = FALSE;	  /* "-v" or "-vv"*/
static int OPTveryverbose= FALSE;	  /* "-vv" */
static int OPTdebug      = FALSE;	  /* "-g" */
static int OPTusage      = FALSE;	  /* "-h", "-H", "-\?", "-help" or "-usage" */
static int OPTheaders    = FALSE;	  /* "-headers" */
static int OPTmemory     = FALSE;	  /* "-m" or "-mem" */
static int OPTnocompile  = FALSE;	  /* "-n" or "-nocompile" */
static int OPTrchecks    = FALSE;	  /* "-r" or "-rchecks" */
static int OPTnop3inc    = FALSE;	  /* "-nop3inc" or "-nodefaults" */
static int OPTnop3lib    = FALSE;	  /* "-nop3lib" or "-nodefaults" */
static int OPTtime       = FALSE;	  /* "-t" or "-time" */
static int OPTtotaltime  = FALSE;	  /* "-tt" or "-total" */
static int OPTnowarnings = FALSE;	  /* "-w" or "-ww" */
static int OPTwwnowarnings= FALSE;	  /* "-ww" */
static int OPToutput     = FALSE;	  /* "-o" */

enum codeversion {cversion, pvmversion, mplversion};
static int OPTcodeversion= cversion;	  /* default -C */

#ifdef MPL_VERSION
static int OPTsmall      = FALSE;	  /* "-s" or "-small" */
#endif /* MPL_VERSION */

#ifdef DEVELOP
static int OPT_Zw = FALSE;		  /* "-Zw" */
static int OPT_Zs = FALSE;		  /* "-Zs" */
static int OPT_Zq = FALSE;		  /* "-Zq" */
static int OPT_Zc = FALSE;		  /* "-Zc" */
static int OPT_Z1 = FALSE;		  /* "-Z1" */
static int OPT_Z2 = FALSE;		  /* "-Z2" */
#endif /* DEVELOP */

/*********** END OF CONFIGURATION PART **************************************/

/*============================================================================*/


/****************************************************************************
 *  Now the code part of the program
 ****************************************************************************/

static char *program = P3NAME;		  /* Name of this program */

/* copy some constants strings to own memory to avoid segmentation */
/* violations when modifying them. */

/* Objects valid in C-mode (default) */
static char p3c[]     = P3C;		  /* name of Parallaxis-compiler */
static char p3cc[]    = P3CC;		  /* name of C-compiler */
static char ext[]     = EXT;		  /* file name extension */
static char p3inc[]   = P3INC;		  /* Include directory */
static char p3lib[]   = P3LIB;		  /* Library directory */
static char standardlib[] = STANDARDLIB;

/* Objects valid in PVM-mode */
#ifdef PVM_VERSION
static char p3c_pvm[]     = P3C_PVM;	  /* name of Parallaxis-compiler */
static char p3cc_pvm[]    = P3CC_PVM;	  /* name of PVM-compiler */
static char ext_pvm[]     = EXT_PVM;	  /* file name extension */
static char p3inc_pvm[]   = P3INC_PVM;	  /* Include directory */
static char p3lib_pvm[]   = P3LIB_PVM;	  /* Library directory */
static char standardlib_pvm[] = STANDARDLIB_PVM;
#endif

/* Objects valid in MPL-mode */
#ifdef MPL_VERSION
static char p3c_mpl[]     = P3C_MPL;	  /* name of Parallaxis-compiler */
static char p3cc_mpl[]    = P3CC_MPL;	  /* name of MPL-compiler */
static char ext_mpl[]     = EXT_MPL;	  /* file name extension */
static char p3inc_mpl[]   = P3INC_MPL;	  /* Include directory */
static char p3lib_mpl[]   = P3LIB_MPL;	  /* Library directory */
static char standardlib_mpl[] = STANDARDLIB_MPL;
#endif

/* Objects finally used when compiling */
static char *usedp3c = p3c;		  /* Parallaxis compiler */
static char *usedp3cc = p3cc;		  /* backend compiler */
static char *usedext = ext;		  /* file name extension */
static char *usedp3inc = p3inc;		  /* include directories */
static char *usedp3lib = p3lib;		  /* library directories */
static char *usedstdlib = standardlib;	  /* libraries */

static char *p_p3cc  = NULL;		  /* proposed backend compiler by */
					  /* command line */

static int n;				  /* number of remaining arguments */
					  /* in argv */
static int from, to;			  /* used for copying arguments */

static int nofcfiles;			  /* Number of unknown files found */
					  /* in command line */

static long starttime, p3time, totaltime; /* timing values */


#define STRINGBLOCKSIZE 20		  /* always get strings in blocks of 20 */

typedef struct				  /* array of strings */
{
     int entries;			  /* current number of entries */
     char **strings;			  /* string entries themselves */
} stringarray;


static stringarray p3pathlist    = {0, NULL}; /* list of Parallaxis paths */
static stringarray cpathlist     = {0, NULL}; /* list of C-includepaths */
static stringarray libpathlist   = {0, NULL}; /* list of C-librarypaths */
static stringarray arguments     = {0, NULL}; /* list of all arguments */
					      /* (P3OPT and argv) */

static stringarray p3commandline = {0, NULL}; /* command line for p3c */
static stringarray ccommandline  = {0, NULL}; /* the same for backend compiler */

static stringarray warnings      = {0, NULL}; /* warnings already given */

static char *outfile;			  /* filename of option "-o" if */
					  /* present */


static int check(stringarray *array, char *text); /* forward declaration */



/*-------------Errors, Warnings and Usage-----------------------------------*/

/*****************************************************************************
 *  Function:	static void warning(char *text)
 *
 *  Parameters:	text:  text to be displayed
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Print a warning message.
 *****************************************************************************/

#define language_stmt(deutsch_stmt, english_stmt)\
switch (error_messages_language) { \
 case deutsch: \
   deutsch_stmt; \
   break; \
 default: \
 case english: \
   english_stmt; \
   break; \
 }
static void warning(char *text)
{
   language_stmt(fprintf(stderr, "%s: Warnung! %s\n", program, text), 
                 fprintf(stderr, "%s: Warning! %s\n", program, text));
}

static void len0warning(char *text)
{
     if (!check(&warnings, text))	  /* already reported? */
     {
       language_stmt(fprintf(stderr, "%s: Warnung! Illegale Environmentvariable in "
			     "`%s'.\n", program, text);,
                     fprintf(stderr, "%s: Warning! Illegal environment variable in "
			     "`%s'.\n", program, text);)
     }
   }

static void suspiciouswarning(char *name)
{
     if (!check(&warnings, name))	  /* already reported? */
     {
       language_stmt(
	  fprintf(stderr, "%s: Warnung! Der verdaechtge VAriablenname `$%s' "
		  "wurde nicht ersetzt.\n", program, name);,
	  fprintf(stderr, "%s: Warning! Suspicious variable name `$%s' "
		  "not replaced.\n", program, name);
		     )
     }
}

static void undefinedwarning(char *name)
{
     if (!check(&warnings, name))	  /* already reported? */
     {
       language_stmt(
	  fprintf(stderr, "%s: Warnung! Undefinierte Environmentvariable "
		  "`$%s'.\n", program, name);,
	  fprintf(stderr, "%s: Warning! Undefined environment variable "
		  "`$%s'.\n", program, name));
     }
}

/*****************************************************************************
 *  Function:	static void error(char *text, int retval)
 *
 *  Parameters:	text:  text to be displayed
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Print an error message and exit program.
 *****************************************************************************/

static void error(char *text, int retval)
{
language_stmt(     fprintf(stderr, "%s: Fehler! %s\n", program, text);,
     fprintf(stderr, "%s: Error! %s\n", program, text);)
     exit(retval);
}

static void mallocerror(void)
{
     language_stmt(error("Konnte keinen Speicher fuer einen String z.B Pfad- oder Filename bereitstellen", -1);,
error("Can't malloc space for string, e.g. path or filename.", -1);)
}


#ifndef MPL_VERSION
static void mplerror(char *option)
{
     language_stmt(fprintf(stderr, "%s: Option %s fehlerhaft: %s ist ohne MPL-Support compiliert.\n", program, option, program);,
		   fprintf(stderr, "%s: Error at option %s: %s compiled without MPL support.\n", program, option, program);)
     exit(-1);
}
#endif

#ifndef PVM_VERSION
static void pvmerror(char *option)
{
     language_stmt(fprintf(stderr, "%s: Option %s fehlerhaft: %s wurde ohne PVM-Support compiliert.\n", program, option, program);,
		   fprintf(stderr, "%s: Error at option %s: %s compiled without PVM support.\n", program, option, program);)
     exit(-1);
}
#endif

#ifndef C_VERSION
static void c_error(char *option)
{
     language_stmt(fprintf(stderr, "%s: Option %s fehlerhaft: %s wurde ohne C-simulation compiliert.\n", program, option, program);,
		   fprintf(stderr, "%s: Error at option %s: %s compiled without C-simulation support.\n", program, option, program);)
     exit(-1);
}
#endif



/*****************************************************************************
 *  Function:	static void usage(int returnval)
 *
 *  Parameters:	returnval: value to be passed to exit
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Print usage information on screen and exit with the given value.
 *****************************************************************************/

static void printversion(void)
{
     fprintf(stderr, "%s -- Parallaxis-III Compiler User Interface %s (%s)\n",
	     program, VERSION, DATE+6); 
}

static void usage(int returnval)
{
     printf("NAME\n\t");
     printversion();

     printf("\nDESCRIPTION\n");
     printf("\tCompile some Parallaxis-III programs and call backend compiler.\n");

     printf("\nSYNOPSIS\n");
     printf("\t%s [options] [file] ...\n", program);

     printf("\nOPTIONS\n");
#ifdef C_VERSION
     printf("\t-C             Generate C-code for simulation (default)\n");
#endif /* C_VERSION */
     printf("\t-casts         Generate type casts to make C-programs lint free\n");
     printf("\t-cc name       Name of the backend compiler to use\n");
     printf("\t-g             Generate debug code (also passed to backend compiler)\n");
     printf("\t-h, -H, -help  Print this usage\n");
     printf("\t-headers       Generate header files for imported modules\n");
     printf("\t-Ipath         Add path to import/include list (Parallaxis and backend)\n");
     printf("\t-indent i      Set indent of generated code to i blanks\n");
     printf("\t-koption       Pass option directly to backend compiler\n");
     printf("\t-Lpath         Add path to library path (backend only)\n");
     printf("\t-m, -mem       Print statistics about used memory\n");
#ifdef MPL_VERSION
     printf("\t-MPL, -mpl     Generate MPL-code for MasPar\n");
#endif /* MPL_VERSION */
     printf("\t-n, -nocompile Don't compile, just show comands (implies -v)\n");
     printf("\t-nop3inc       Don't use standard include paths\n");
     printf("\t-nop3lib       Don't use standard library paths\n");
     printf("\t-nodefaults    Same as -nop3inc -nop3lib\n");
     printf("\t-o name        Name of the generated executable\n");
     printf("\t-p             Parallaxis compile only, don't call backend compiler\n");
     printf("\t-c             Parallaxis and backend compile only, don't call linker\n");
     printf("\t-Ppath         Add path to import list (Parallaxis only)\n");
#ifdef PVM_VERSION
     printf("\t-PVM, -pvm     Generate PVM-code for Paragon\n");
#endif /* PVM_VERSION */
     printf("\t-r, -rchecks   Don't generate runtime range checks\n");
#ifdef MPL_VERSION	      
     printf("\t-s, -small     Generate small MPL-only model (max. 128KB)\n");
#endif /* MPL_VERSION */      
     printf("\t-t, -time      Print statistics about used time (Parallaxis only)\n");
     printf("\t-tt, -total    Like -t, but also for backend compiler\n");
     printf("\t-v             Print version of %s and the resulting compiler calls\n", program);
     printf("\t-vv            Like -v, passes also -v to backend compiler\n");
     printf("\t-w             Don't generate warnings\n");
     printf("\t-ww            Like -w, passes also -w to backend compiler\n");
			      
#ifdef DEVELOP
     printf("\nOPTIONS ONLY AVAILABLE DURING DEVELOPMENT\n");
     printf("\t-Zw            Write code tree\n");
     printf("\t-Zs            Write symbols tree\n");
     printf("\t-Zq            Query code tree\n");
     printf("\t-Zc            Check code tree\n");
     printf("\t-Z1            Run parser only, no semantic check\n");
     printf("\t-Z2            Run parser and semantic check only, no code generation\n");
#endif /* DEVELOP */

     printf("\n");
     printf("\tEvery other option is passed unchanged to the backend compiler.\n");

     printf("\nENVIRONMENT\n");
     printf("\tP3CC           Name of the backend compiler\n");
     printf("\tP3INC          \":\"-seperated list of paths where to find sources\n");
     printf("\tP3LIB          \":\"-seperated list of paths where to find libraries\n");
     printf("\tP3OPT          Default options always to set\n");

     exit(returnval);
}



/*--------------String Utility Functions------------------------------*/

/*****************************************************************************
 *  Function:	static void addstring(stringarray *array, char *string)
 *
 *  Parameters:	array:  list of strings
 *              string: string to be appended to the list
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Add the string at the end of the string array. The entries are handled
 *  in blocks of size STRINGBLOCKSIZE. If the string array is full, the
 *  array is expanded by another block.
 *
 *  REMARK: By setting STRINGSIZEBLOCK to 1, the array size is exactly the
 *  number of strings, so no extra space is wasted. But this approach is
 *  probably somehow slower as each insertion results in a call of realloc()
 *  which in turn may copy the whole memory block to another address.
 *****************************************************************************/

static void addstring(stringarray *array, char *string)
{
     if (array->entries == 0)		  /* empty list? get a block */
     {
	  array->strings = (char **) malloc(STRINGBLOCKSIZE * sizeof(char *));
     }
     else if (array->entries % STRINGBLOCKSIZE == 0) /* block full? get */
     {						    /* another block */
	  array->strings = (char **) realloc( array->strings,
					    (array->entries+STRINGBLOCKSIZE)
					    * sizeof(char *));
     }
     array->strings[array->entries++] = string; /* insert string in array */
}



/*****************************************************************************
 *  Function:	static void printstringarray(stringarray *array)
 *
 *  Parameters:	array:  list of strings
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Print the strings of the string array as one output line to stderr. The
 *  single strings are seperated by blanks and the line is ended by a
 *  newline.
 *****************************************************************************/

static void printstringarray(stringarray *array)
{
     int i=0;
     
     while (i<array->entries)
     {
	  fprintf(stderr, "%s", array->strings[i]);
	  i++;
	  if (i<array->entries) 
	       fprintf(stderr, " ");
	  else 
	       fprintf(stderr, "\n");
     }
}



/*****************************************************************************
 *  Function:	static int check(stringarray *array, char *text)
 *
 *  Parameters:	array:  list of strings
 *              text:   string to be checked
 *      
 *  Result:	0: string not found, now added
 *              1: string found, not added
 *
 *  Description:
 *  ------------
 *  Check if the given text is already in the array of strings. If yes, 1 is
 *  returned; if no, 0 is returned and a copy of the string is added to the
 *  array. 
 *****************************************************************************/

static int check(stringarray *array, char *text)
{
     int i;
     char *copy;

     for (i=0; i<array->entries; i++)
     {
	  if (strcmp(array->strings[i], text) == 0)
	       return 1;
     }

     copy = (char *) malloc(strlen(text)+1);
     if (copy==NULL) mallocerror();

     strcpy(copy, text);
     
     addstring(array, copy);
     return 0;
}



/*****************************************************************************
 *  Function:	static void splitline(stringarray *array, char *line)
 *
 *  Parameters:	array:  list of strings
 *              line:   string to be parsed
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Split the given string "line" into single substrings and put them in the
 *  string array. Blanks and tabs are used as word separators.
 *****************************************************************************/

/* Skip blanks and tabulators */
#define SKIPSPACES(p) while(((*p)==' ') || ((*p)=='\t')){p++;}

static void splitline(stringarray *array, char *line)
{
     char *p;

     SKIPSPACES(line);			  /* skip leading whitespace */
     while (*line!='\0')		  /* reached end of text? */
     {
	  p = strchr(line, ' ');	  /* search for blank and tabs */
	  if (p==NULL) p = strchr(line, '\t');
	  if (p==NULL)			  /* nothing found? */
	  {
	       addstring(array, line);	  /* use rest as one argument */
	       line += strlen(line);	  /* go to end of text */
	  }
	  else
	  {
	       *p = '\0';		  /* replace whitespace by \0 */
	       addstring(array, line);	  /* add the string */
	       line = p+1;		  /* go to next word */
	       SKIPSPACES(line);	  /* skip multiple whitespace chars */
	  }
     }
     
}



/*****************************************************************************
 *  Function:	static void setpaths(stringarray *pathlist, char *envvar, 
 *                                                       char *defaultpaths)
 *
 *  Parameters:	pathlist:     list of pathnames
 *              envvar:       string to be parsed
 *              defaultpaths: path names to add when "@" is encountered
 *                            (NULL if empty)
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Split the environment variable envvar into a list of pathnames.
 *  Pathnames in the environment variable are seperated by any character of
 *  VALUESEPARATORS. In the special case of the string "@", the whole list
 *  of pathnames given in defaultpaths is inserted at the current place.
 *****************************************************************************/

static void setpaths(stringarray *pathlist, char *envvar, char
		     *defaultpaths)
{
     char *p, *new;
     unsigned long len;

     while (*envvar != '\0')
     {
	  p = envvar;
	  while ((*p != '\0') && (strchr(VALUESEPARATORS, *p) == NULL))
	       p++;			  /* search end of one pathname */

	  len = p-envvar;		  /* length (without seperator) */

	  if (len>0)
	  {
	       if ((len==1) && (*(envvar)=='@') && (defaultpaths != NULL))
	       {			  /* set defaults by a recursive call */
		    setpaths(pathlist, defaultpaths, NULL);
	       }
	       else
	       {			  /* copy pathname */
		    new = (char *) malloc(len+1); /* get space for string */
		    if (new==NULL) mallocerror();
		    strncpy(new, envvar, len);
		    new[len] = '\0';
		    addstring(pathlist, new);
	       }
	  }
	  if (*p != '\0') p++;		  /* skip separator */
	  envvar = p;
     };
}



/*****************************************************************************
 *  Function:	static void replaceenvvars(stringarray *array)
 *
 *  Parameters:	array:  list of strings to check for variables
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Check every string in the list for environment variables. If the
 *  character '$' is found, every occurrence is replaced by the appropriate
 *  variable. Variable names must obey the usual identifier rules.
 *****************************************************************************/

static void replaceenvvars(stringarray *array)
{
     int i;				  /* argument counter */
     char *env, *end;			  /* start and end of variable name */
     char *name;			  /* name of the environment variable */
     int len;				  /* length of name */
     char *ep;				  /* contents of environment variable */
     int start;				  /* length of leading string part */
					  /* before the environment variable*/
     char *string;			  /* new string after replacement */
     char c;

     for (i=0; i<array->entries; i++)	  /* loop over all entries */
     {
	  env = array->strings[i];
	  while ((env = strchr(env, '$'))) /* loop over all variable names */
	  {				   /* in the string */
	       env++;
	       end = env;
	       while (1)		  /* search end of variable name */
	       {
		    c = *end;
		    if (!isalnum(c) && (c!='_')) break;
		    end++;
	       }

	       len = end-env;
	       if (len<1)		  /* name length=0? illegal name */
	       {
		    len0warning(array->strings[i]); /* give warning */
		    env = end;
		    continue;		  /* and don't replace */
	       }

	       name = (char *) malloc(len+1); /* extract name */
	       if (name==NULL) mallocerror();
	       strncpy(name, env, len);
	       name[len]='\0';

	       if (isdigit(*env))	  /* name starting with digit? */
	       {
		    suspiciouswarning(name); /* give warning */
		    env = end;
		    continue;		  /* and don't replace */
	       }

	       ep = getenv(name);	  /* get environment variable */
	       if (ep==NULL)		  /* not defined? */
	       {
		    undefinedwarning(name); /* give warning */
		    ep = "";		  /* and replace by empty string */
	       }

	       /* Computation of new string length: 
		  old string length:         strlen(array->strings[i])
		  remove var-name incl. '$': -(strlen(name)+1)
		  add var contents length:   +strlen(ep)
		  plus \0 at end:            +1             */
	       string = (char *) malloc(strlen(array->strings[i])+
					strlen(ep)-strlen(name)-1+1);
	       if (string==NULL) mallocerror();
	       
	       /* replace environment variable by its contents */
	       start = (env-1)-array->strings[i];
	       strncpy(string, array->strings[i], start); /* start with */
	       string[start] = '\0';	                  /* leading part */
	       strcat(string, ep);	  /* concatenate contents */
	       env = string+strlen(string);
	       strcat(string, end);	  /* and trailing part */
	       array->strings[i] = string;

	       free(name);		  /* release name */
	  } /* while */
     } /* for */
}


/*------------Files and Processes-------------------------------------------*/

/*****************************************************************************
 *  Function:	static int fileexists(char *basename, stringarray *pathlist)
 *
 *  Parameters:	basename: filename of the file
 *              pathlist: list of directories where to search for the file
 *      
 *  Result:	TRUE:  file exists at least in one directory
 *              FALSE: file does not exist in the given directories
 *
 *  Description:
 *  ------------
 *  Check if the given file exists in the current directory. If not, the
 *  directories in the pathlist are prepended one after the other and the
 *  file is searched there. If the file is found anywhere, TRUE is returned,
 *  else FALSE.
 *****************************************************************************/

static int fileexists(char *basename, stringarray *pathlist)
{
     int i;
     FILE *file;
     char *fullname;
     char pathsepstring[]={PATHSEPARATOR, '\0'};

     file = fopen(basename, "r");	  /* exists file in current directory? */
     if (file!=NULL)
     {
	  fclose(file);			  /* yes, return TRUE */
	  return TRUE;
     }

     /* Try prepending the paths one after the other */
     for (i=0; i<pathlist->entries; i++)
     {
	  /* get space for full filename */
	  fullname = (char *) malloc(strlen(basename)+strlen(pathsepstring)
				     +strlen(pathlist->strings[i])+1);
	  if (fullname==NULL)
	       mallocerror();

	  /* concatenate directory and basefilename */
	  strcpy(fullname, pathlist->strings[i]); /* directory */
	  strcat(fullname, pathsepstring);        /* separator "/" */
	  strcat(fullname, basename);	          /* basename */

	  file = fopen(fullname, "r");	  /* test extistance by opening */
	  free(fullname);
	  if (file!=NULL)		  /* if the file exists, return TRUE */
	  {
	       fclose(file);
	       return TRUE;
	  }
     }
     return FALSE;
}



/*****************************************************************************
 *  Function:	static int execute(stringarray *commandline)
 *
 *  Parameters:	commandline: command line of the program to execute
 *      
 *  Result:	return value of the program after execution
 *
 *  Description:
 *  ------------
 *  Execute the program with the given commandline by forking a new process.
 *  The returned status is analyzed and the return code of the program is
 *  returned.
 *****************************************************************************/

static int execute(stringarray *commandline)
{
     pid_t pid;
     int status;
     
     if (OPTnocompile)			  /* Only execute if not prevented by */
	  return 0;			  /* an option */

     if ((pid = fork())==0)		  /* fork a new process */
     {
	  char **argv;
	  int argc;
	  int returnvalue;

	  argc = commandline->entries;	  /* get arguments list */
	  addstring(commandline, NULL);	  /* add a NULL string */
	  argv = commandline->strings;    /* get address */
#if 0
	  { 
	       int i;
	       /* print resulting argumentlist for debugging purposes */
	       printf("argc: %i\n",argc);
	       for (i=0; argv[i]!=NULL; i++)
		    printf("argv[%i]: %s\n",i,argv[i]);
	  }
	  exit(0);
#else
	  returnvalue = execvp(argv[0], argv);	  /* execute the program */
	  
	  if (returnvalue == -1) 
	  {	       
	       char errheader[1024];
	       sprintf(errheader,"%s (starting %s)",program,argv[0]);
	       perror(errheader);
	  }
	  exit(-99);			  /* only reached if exec failed */
#endif
     }

     if (pid==-1)			  /* fork error? */
	  error("Can't spawn a new process.", -99);

     /* wait for compiler to finish */
     wait(&status);

     /* analyse status gotten from the subprocess */
     if (WIFSIGNALED(status))
	  fprintf(stderr, "Compiler process received signal %i\n", 
		  WTERMSIG(status));

     if (WIFSTOPPED(status))
	  fprintf(stderr, "Compiler process was stopped by signal %i\n", 
		  WSTOPSIG(status));

     if (!WIFEXITED(status))
	  error("Compiler exited abnormally, probably was killed.", -1);

#if 0
     if (WIFEXITED(status) && WEXITSTATUS(status)== (-99 & 0xFF))
	  error("Compiler process could not be started. Program not found.", -99);
#endif

     return WEXITSTATUS(status);
}



/*------------Parallaxis Options Recognition--------------------------------*/

/*****************************************************************************
 *  Function:	static void checkp3options(char *argv[])
 *
 *  Parameters:	argv: list of arguments
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Compare one argument with the option strings. If an option needs
 *  parameters, they are fetched from the subsequent arguments. Recognized
 *  options are deleted from the list.
 *****************************************************************************/

static void checkp3options(char *argv[])
{
     char *a;				  /* its easier than "argv[from]" */

     a = argv[from];
     if (strcmp(a, "-casts")==0)
     {
	  OPTcasts = TRUE;		  /* Option -casts ? */
	  from++;
	  n--;
     }
     else if (strcmp(a, "-c")==0)
     {
	  OPTnolink = TRUE;		  /* Option -c ? */
	  from++;
	  n--;
     }

     else if (strcmp(a, "-p")==0)
     {
	  OPTp3only = TRUE;		  /* Option -p ? */
	  from++;
	  n--;
     }
     else if (strcmp(a, "-v")==0)
     {
	  OPTversion = TRUE;		  /* Option -v ? */
	  OPTverbose = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-vv")==0)
     {
	  OPTveryverbose = TRUE;
	  OPTverbose = TRUE;		  /* Option -vv ? */
	  OPTversion = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-g")==0)
     {
	  OPTdebug = TRUE;		  /* Option -g ? */
	  from++;
	  n--;
     }
     else if ((strcmp(a, "-h")==0) ||
	      (strcmp(a, "-H")==0) ||
	      (strcmp(a, "-?")==0) ||
	      (strcmp(a, "-help")==0) ||
	      (strcmp(a, "-usage")==0))
     {
	  OPTusage = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-headers")==0)
     {
	  OPTheaders = TRUE;
	  from++;
	  n--;
     }
     else if ((strcmp(a, "-m")==0) ||
	      (strcmp(a, "-mem")==0))
     {
	  OPTmemory = TRUE;
	  from++;
	  n--;
     }
     else if ((strcmp(a, "-n")==0) ||
	      (strcmp(a, "-nocompile")==0))
     {
	  OPTnocompile = TRUE;
	  OPTverbose   = TRUE;
	  from++;
	  n--;
     }
     else if ((strcmp(a, "-C")==0))
     {
#ifdef C_VERSION
	  OPTcodeversion = cversion;	  /* Option -C or -ccode? */
	  from++;
	  n--;
#else
	  c_error(a);
#endif /* C_VERSION */
     }
     else if ((strcmp(a, "-pvm")==0) ||
	      (strcmp(a, "-PVM")==0))
     {
#ifdef PVM_VERSION
	  OPTcodeversion = pvmversion;
	  from++;
	  n--;
#else
	  pvmerror(a);
#endif /* PVM_VERSION */
     }
     else if ((strcmp(a, "-r")==0) ||
	      (strcmp(a, "-rchecks")==0))
     {
	  OPTrchecks = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-cc")==0)
     {
	  from++;			  /* get C-compiler name from next */
	  p_p3cc = argv[from++];	  /* argument */
	  n-=2;
     }
     else if (strcmp(a, "-nop3inc")==0)
     {
	  OPTnop3inc = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-nop3lib")==0)
     {
	  OPTnop3lib = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-nodefaults")==0)
     {
	  OPTnop3inc = TRUE;		  /* set both */
	  OPTnop3lib = TRUE;
	  from++;
	  n--;
     }

     else if ((strcmp(a, "-mpl")==0) ||
	      (strcmp(a, "-MPL")==0))
     {
#ifdef MPL_VERSION
	  OPTcodeversion = mplversion;	  /* Option -mpl, -MPL or -mplcode? */
	  from++;
	  n--;
#else
	  mplerror(a);
#endif /* MPL_VERSION */
     }
#ifdef MPL_VERSION
     else if ((strcmp(a, "-s")==0) ||
	      (strcmp(a, "-small")==0))
     {
	  OPTsmall = TRUE;		  /* Option -s or -small? */
	  from++;
	  n--;
     }
#endif /* MPL_VERSION */
     else if ((strcmp(a, "-t")==0) ||
	      (strcmp(a, "-time")==0))
     {
	  OPTtime = TRUE;		  /* Option -t or -time */
	  from++;
	  n--;
     }
     else if ((strcmp(a, "-tt")==0) ||
	      (strcmp(a, "-total")==0))
     {
	  OPTtotaltime = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-w")==0)
     {
	  OPTnowarnings = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-ww")==0)
     {
	  OPTwwnowarnings =  TRUE;
	  from++;
	  n--;
     }

#ifdef DEVELOP
     else if (strcmp(a, "-Zw")==0)
     {
	  OPT_Zw = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-Zs")==0)
     {
	  OPT_Zs = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-Zq")==0)
     {
	  OPT_Zq = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-Zc")==0)
     {
	  OPT_Zc = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-Z1")==0)
     {
	  OPT_Z1 = TRUE;
	  from++;
	  n--;
     }
     else if (strcmp(a, "-Z2")==0)
     {
	  OPT_Z2 = TRUE;
	  from++;
	  n--;
     }
#endif  /* DEVELOP */

     else if (strncmp(a, "-indent", 7)==0)
     {
	  int f = FALSE;		  /* flag for warning state */
	  int indent = 0;

	  a = &(a[7]);
	  from++;
	  n--;
	  if (a[0] == '\0')		  /* Parameter in same argument? */
	  {
	       if ((argv[from][0]>='0') && (argv[from][0]<='9'))
	       {
		    a=argv[from++];	  /* no, it follows in the next */
		    n--;
	       }
	       else
		    f = TRUE;		  /* no number following in next arg */
	  }
	  else if ((a[0] <='0') || (a[0] >='9'))
	       f = TRUE;		  /* no number following -indent */

	  if (!f)
	  {
	       indent = atoi(a);
	       if (indent>0)
		    OPTindent = indent;	  /* Set indent value */
	       else
		    f = TRUE;		  /* Value 0 is impossible */
	  }
	  if (f)			  /* warning state reached? */
	       warning("Option -indent: expected number (>0).");
     }
     else if (strncmp(a, "-o", 2)==0)
     {
	  OPToutput = TRUE;
	  if (a[2]=='\0')		  /* -o alone in one argument? */
	  {
	       from++;
	       outfile = argv[from++];	  /* filename in next argument */
	       n-=2;
	  }
	  else
	  {
	       outfile = &(a[2]);	  /* filename is rest of argument */
	       from++;
	       n--;
	  }
     }
     else if (strncmp(a, "-k", 2)==0)
     {
	  if (a[2]=='\0')		  /* -k alone in one argument? */
	  {
	       from++;			  /* delete the -k */
	       n--;			 
	       argv[to++] = argv[from++]; /* and use the next argument unseen */
	  }
	  else
	  {
	       a[1] = '-';		  /* put the "-" before the parameter */
	       argv[to++] = &(a[1]);	  /* behind the k and use this as */
	       from++;			  /* new argument */
	  }
     }
     else if (strncmp(a, "-P", 2)==0)
     {
	  if (a[2]=='\0')		  /* path in extra argument? */
	  {
	       from++;			  /* yes: add next argument to pathlist */
	       addstring(&p3pathlist, argv[from++]);
	       n-=2;
	  }
	  else
	  {				  /* no: path follows -p immediatly */
	       addstring(&p3pathlist, &(a[2])); /* add it to pathlist */
	       from++;
	       n--;
	  }
     }
     else if (strncmp(a, "-I", 2)==0)
     {
	  if (a[2]=='\0')		  /* path in extra argument? */
	  {
	       from++;			  /* yes: add next argument to the */
	       addstring(&p3pathlist, argv[from]);	/* two pathlists */
	       addstring(&cpathlist, argv[from++]);
	       n-=2;
	  }
	  else
	  {
	       addstring(&p3pathlist, &(a[2])); /* no: add the immediatly following */
	       addstring(&cpathlist, &(a[2])); /* path to the two pathlists */
	       from++;
	       n--;
	  }
     }
     else if (strncmp(a, "-L", 2)==0)
     {
	  if (a[2]=='\0')		  /* path in extra argument? */
	  {
	       from++;			  /* yes: add next argument to the*/
	       addstring(&libpathlist, argv[from]); /* pathlist */
	       n--;
	  }
	  else
	  {				  /* no: path follows -L immediatly */
	       addstring(&libpathlist, &(a[2])); /* add it to pathlist */ 
	       from++;
	       n--;
	  }
     }

     else				  /* unknown option */
     {
	  argv[to++] = argv[from++];	  /* leave unmodified */
     }
}



/*****************************************************************************
 *  Function:	static void getp3options(stringarray *arguments)
 *
 *  Parameters:	arguments: list of command line arguments
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Parse the given command line and extract all known p3-options. Delete
 *  them from the argument list and set the appropriate OPT-variables.
 *  Arguments not beginning with "-" are expected to be filenames. Every
 *  unknown option is left in the arguments to be passed to the back-end
 *  compiler.
 *****************************************************************************/

static void getp3options(stringarray *arguments)
{
     from = 0;
     to = 0;
     n = arguments->entries;		  /* n counts current number of */
					  /* valid entries (decreases as */
					  /* options are deleted from the list) */

     while (from < arguments->entries)
     {
	  if (arguments->strings[from][0]=='-')	/* Option? */
	  {
	       checkp3options(arguments->strings); /* check and delete it if */
						   /* recognized */
	  }
	  else				  /* found filename, leave it unchanged */
	  {				  /* but copy it to the new position */
	       arguments->strings[to++] = arguments->strings[from++];
	  }
     }
     arguments->entries = n;		  /* these arguments are left */
}


/*--------------Parallaxis compilation--------------------------------------*/

/*****************************************************************************
 *  Function:	static void setp3commandline(void)
 *
 *  Parameters:	-
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Generate the basic Parallaxis-compiler command line. Later there only
 *  has to be appended the filename of the appropriate source file.
 *****************************************************************************/

static void setp3commandline(void)
{
     char *str;
     static char indentstring[20];
     int i;

     /* Generate Parallaxis command line: 1. start with command name */
     addstring(&p3commandline, usedp3c);

     /* 2. add options */
     if (OPTcasts) addstring(&p3commandline, sOPTcasts);
     if (OPTveryverbose) addstring(&p3commandline, sOPTverbose);
     if (OPTdebug) addstring(&p3commandline, sOPTdebug);
     if (OPTheaders) addstring(&p3commandline, sOPTheaders);
     if (OPTmemory) addstring(&p3commandline, sOPTmemory);
     if (OPTrchecks) addstring(&p3commandline, sOPTrchecks);
#ifdef MPL_VERSION
     if (OPTsmall) addstring(&p3commandline, sOPTsmall);
#endif /* MPL_VERSION */
     if (OPTtime) addstring(&p3commandline, sOPTtime);
     if (OPTnowarnings) addstring(&p3commandline, sOPTnowarnings);
#ifdef DEVELOP
     if (OPT_Zw) addstring(&p3commandline, sOPT_Zw);
     if (OPT_Zs) addstring(&p3commandline, sOPT_Zs);
     if (OPT_Zq) addstring(&p3commandline, sOPT_Zq);
     if (OPT_Zc) addstring(&p3commandline, sOPT_Zc);
     if (OPT_Z1) addstring(&p3commandline, sOPT_Z1);
     if (OPT_Z2) addstring(&p3commandline, sOPT_Z2);
#endif /* DEVELOP */
     if (OPTindent>=0) 
     {
	  sprintf(indentstring, "%s%d", sOPTindent, OPTindent);
	  addstring(&p3commandline, indentstring);
     }

     /* 3. add paths where to find definition modules */
     for (i=0; i<p3pathlist.entries; i++)
     {
	  str = (char *) malloc(strlen(sOPTpath)+strlen(p3pathlist.strings[i])+1);
	  if (str==NULL) mallocerror();
	  strcpy(str, sOPTpath);
	  strcat(str, p3pathlist.strings[i]);
	  addstring(&p3commandline, str);
     }
}



/*****************************************************************************
 *  Function:	static void p3compile(char *filename)
 *
 *  Parameters:	filename:  sourcefile to compile
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Add the given filename to the Parallaxis compiler command line and call
 *  the compiler.
 *****************************************************************************/

static void p3compile(char *filename)
{
     int returncode;

#if 0
     printf("Compiling %s ...\n", filename);
#endif

     /* Append filename (temporarily) to arguments */
     addstring(&p3commandline, filename);

     /* Print command line if requested */
     if (OPTverbose)
	  printstringarray(&p3commandline);

     /* Call the Parallaxis compiler */
     returncode = execute(&p3commandline);

     /* Remove the filename again from the arguments list */
     p3commandline.entries--;

     if ((returncode!=0)) /* exit on error */
	  exit(returncode);
}



/*****************************************************************************
 *  Function:	static void p3compilepm(int index, char *argv[])
 *
 *  Parameters:	index: argument index number in the arguments list
 *              argv:  arguments list; argv[index] is the filename to
 *                     compile 
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Compile the given file via p3compile() and then rename the file by
 *  exchanging the extension to ".c" (or ".m" in MPL-mode).
 *
 *  REMARK: This routine assumes that the given filename definitely has an
 *  extension (usually ".pm"). So no extra checks for the correctness of the
 *  extension are made.
 *****************************************************************************/

static void p3compilepm(int index, char *argv[])
{
     char *ext;
     char *new;
     unsigned long len;
     

     p3compile(argv[index]);
     ext = strrchr(argv[index], '.');
     len = ext-argv[index];

     new = (char *) malloc(len + strlen(usedext)+1);
     if (new==NULL) mallocerror();

     strncpy(new, argv[index], len);
     new[len] = '\0';
     strcat(new, usedext);
     argv[index] = new;
}



/*****************************************************************************
 *  Function:	static void compilep3files(stringarray *arguments)
 *
 *  Parameters:	arguments: command line arguments
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Scan argument list for Parallaxis files. Files ending in ".pd" or ".pm"
 *  are sent directly to the compiler, files without any extension are
 *  compiled if a file with extension ".pm" or ".pd" exists. Files with
 *  ".pd" are deleted from the argument list, files with ".pm" are renamed
 *  to ".c" (or ".m" in MPL-mode) and left in the argument list for the
 *  back-end compiler.
 *
 *  REMARK:
 *  Files ending in a single "." are treated the same way as having no
 *  extension at all. This is to allow filename completion in shells which
 *  usually end with the "." when both ".pm" and ".pd" (or even ".c" and
 *  ".px") files exist. As a result there can be no files really ending with
 *  a ".". But this is bad style anyway.
 *****************************************************************************/

static void compilep3files(stringarray *arguments)
{
     char *newpd, *newpm;		  /* new filename with extension */
     int  flag;
     char *ext;				  /* existing extension */

     char **argv;
     int argc;
     from = 0;
     to = 0;

     argv = arguments->strings;
     argc = arguments->entries;
     n = argc;

     while (from<argc)
     {
	  if (argv[from][0]=='-')	  /* Options stay unchanged */
	  {
	       argv[to++]=argv[from++];
	  }
	  else				  /* found Filename */
	  {
	       ext = strrchr(argv[from], '.'); /* has it an extension? */
	       if ((ext != NULL) && (strchr(ext, PATHSEPARATOR)!=NULL))
		    ext = NULL;

	       if ((ext!=NULL) && (strcmp(ext, ".")==0))
	       {			  /* handle filename ending in "." */
		    *ext = '\0';	  /* like a filename without extension */
		    ext = NULL;
	       }
	       if (ext==NULL)
	       {			  /* no extension */
		    newpd = (char *) malloc(strlen(argv[from])+strlen(PDEXT)+1);
		    if (newpd==NULL)
			 mallocerror();

		    strcpy(newpd, argv[from]); /* append ".pd" */
		    strcat(newpd, PDEXT);
		    
		    flag = fileexists(newpd, &p3pathlist);
		    if (flag)		  /* if file exists, compile it */
		    {
			 p3compile(newpd);
			 flag = 1;
		    }

		    newpm = (char *) malloc(strlen(argv[from])+strlen(PMEXT)+1);
		    if (newpm==NULL)
			 mallocerror();

		    strcpy(newpm, argv[from]); /* append ".pm" */
		    strcat(newpm, PMEXT);
		    
		    if (fileexists(newpm, &p3pathlist)) /* if file exists, */
		    {
			 free(newpd);	  /* free pd-name (not needed) */
			 argv[from]=newpm; /* save pm-name in argument */
			 p3compilepm(from, argv); /* and compile */
			 argv[to++] = argv[from++]; /* keep c-file-name */
		    }
		    else
		    {
			 if (!flag)
			 {
		            fprintf(stderr, "Warning! Neither `%s' nor `%s' "
				    "exists => `%s' skipped.\n", 
				    newpd, newpm, argv[from]);
			 }
			 free(newpd);
			 free(newpm);
#if 1
			 from++;	  /* delete this filename from arguments */
			 n--;
#else
			 argv[to++] = argv[from++]; /* keep original name */
#endif
		    }
	       }
	       else if (strcmp(ext, PDEXT)==0) /* extension ".pd"? */
	       {			 
		    p3compile(argv[from]); /* Compile definition module */
		    from++;		  /* and delete it from arguments */
		    n--;
	       }
	       else if (strcmp(ext, PMEXT)==0) /* extension ".pm"? */
	       {
		    p3compilepm(from, argv); /* Compile impl./prog. module */
		    argv[to++] = argv[from++]; /* leave c-file in arguments */
	       }
	       else if (strcmp(ext, PXEXT)==0) /* extension ".px"? */
	       {
		    fprintf(stderr, "Warning! Crossreference file %s ignored.\n",
			    argv[from]);
		    from++;
		    n--;
	       }
	       else
	       {
		    argv[to++] = argv[from++]; /* leave name for backend compiler */
	       }
	  }
     }
     arguments->entries = n;
}


/*----------------Back-end Compilation--------------------------------------*/

/*****************************************************************************
 *  Function:	static void compilecfiles(stringarray *arguments)
 *
 *  Parameters:	arguments: command line arguments for backend compiler
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Generate the command line for the back-end compiler and execute the
 *  compilation process. If the number of files to compile is 0, no compiler
 *  is called. If option "-c" is set, the linkage stuff (libraries and
 *  library paths) are omitted. Every unknown option of the original p3
 *  command line goes into this back-end command line.
 *****************************************************************************/

static void compilecfiles(stringarray *arguments)
{
     int returncode;
     int i;
     char *str;

     /* Count number of files of resulting commandline */
     nofcfiles = 0;
     for (i=0; i<arguments->entries; i++)
	  if (arguments->strings[i][0]!='-') nofcfiles++;

     if (nofcfiles == 0)		  /* if no files there, don't call */
	  return;			  /* the backend compiler */

     /* Generate command line as argument list. 1. start with command name */
     addstring(&ccommandline, usedp3cc);

     /* 2. add options already recognized by p3 */
     if (OPTnolink)       addstring(&ccommandline, sOPTnolink);
     if (OPTveryverbose)  addstring(&ccommandline, sOPTcverbose);
     if (OPTdebug)        addstring(&ccommandline, sOPTcdebug);
     if (OPTwwnowarnings) addstring(&ccommandline, sOPTcnowarnings);
     if (OPToutput)
     {
	  addstring(&ccommandline, sOPToutput);
	  addstring(&ccommandline, outfile);
     }

     /* 3. add include paths */
     for (i=0; i<cpathlist.entries; i++)
     {
	  str = (char *) malloc(strlen(sOPTinclude)+strlen(cpathlist.strings[i])+1);
	  if (str==NULL) mallocerror();
	  strcpy(str, sOPTinclude);
	  strcat(str, cpathlist.strings[i]);
	  addstring(&ccommandline, str);
     }

     /* 4. add library paths (only when linking is not switched off) */
     if (!OPTnolink)
     {
	  for (i=0; i<libpathlist.entries; i++)
	  {
	       str = (char *) malloc(strlen(sOPTlibpath)
				     +strlen(libpathlist.strings[i])+1);
	       if (str==NULL) mallocerror();
	       strcpy(str, sOPTlibpath);
	       strcat(str, libpathlist.strings[i]);
	       addstring(&ccommandline, str);
	  }
     }

     /* 5. add remaining arguments including file names */
     for (i=0; i<arguments->entries; i++)
     {
	  splitline(&ccommandline, arguments->strings[i]);
     }

     /* 6. add standard libraries (only when linking is not switched off) */
     if ((!OPTnop3lib) && (!OPTnolink))
     {
	  splitline(&ccommandline, usedstdlib);
     }

     /* Print resulting command line if requested */
     if (OPTverbose)
	  printstringarray(&ccommandline);

     /* Execute backend compiler */
     returncode = execute(&ccommandline);
     if (returncode!=0)			  /* exit on error */
	  exit(returncode);
}



/*--------------Version dependant stuff-------------------------------------*/

/*****************************************************************************
 *  Function:	static void setusedobjects(void)
 *
 *  Parameters:	-
 *      
 *  Result:	-
 *
 *  Description:
 *  ------------
 *  Determine the compilers, paths, libraries, etc. really used after
 *  checking the target system (C, PVM, MPL) the environment and the command
 *  line.
 *****************************************************************************/

static void setusedobjects(void)
{
     switch (OPTcodeversion)
     {
#ifdef PVM_VERSION
       case pvmversion:			  /* PVM */
	  usedp3c = p3c_pvm;
	  usedp3cc = p3cc_pvm;
	  usedext = ext_pvm;
	  usedp3inc = p3inc_pvm;
	  usedp3lib = p3lib_pvm;
	  usedstdlib = standardlib_pvm;
	  break;
#endif /* PVM_VERSION */
#ifdef MPL_VERSION
       case mplversion:			  /* MPL */
	  usedp3c = p3c_mpl;
	  usedp3cc = p3cc_mpl;
	  usedext = ext_mpl;
	  usedp3inc = p3inc_mpl;
	  usedp3lib = p3lib_mpl;
	  usedstdlib = standardlib_mpl;
	  break;
#endif /* MPL_VERSION */
#ifdef C_VERSION
       case cversion:			  /* C */
#endif /* C_VERSION */
       default:				  /* default */
	  usedp3c = p3c;
	  usedp3cc = p3cc;
	  usedext = ext;
	  usedp3inc = p3inc;
	  usedp3lib = p3lib;
	  usedstdlib = standardlib;
	  break;
     }
     if (p_p3cc) usedp3cc = p_p3cc;	  /* overwrite default by given */
					  /* argument */
}


/* Main program -------------------------------------------------- */

/*****************************************************************************
 *  Function:	int main(int argc, char *argv[], char *env[])
 *
 *  Parameters:	argc:   number of command line arguments
 *              argv:   command line arguments
 *      
 *  Result:	return value of program
 *
 *  Description:
 *  ------------
 *  Compile programs. Scan options, build commandline for Parallaxis
 *  compiler, compile every module, compile and link generated modules by a
 *  call to the backend compiler.
 *****************************************************************************/

int main(int argc, char *argv[], char *env[])
{
     char *ep;				  /* pointer to environment */
     int i;
     
     if (argc>0)
     {
	  argc--;
	  if (argv[0]!=NULL)
	       program = argv[0];	  /* get program name */
	  argv = &argv[1];		  /* skip program name */
     }

     if (argc==0)			  /* give usage if no arguments given */
	  usage(0);

     ep = getenv("P3CC");		  /* get default backend compiler name */
     if (ep!=NULL) p_p3cc = ep;		  /* of environment variable */

     ep = getenv("P3OPT");		  /* if environment variable P3OPT */
     if (ep!=NULL)			  /* is set, break it into arguments */
     {
	  splitline(&arguments, ep);
     }

     for (i=0; i<argc; i++)		  /* append command line arguments */
	  addstring(&arguments, argv[i]); /* to argument list */

     getp3options(&arguments);		  /* parse argument list for options */

     if (OPTusage)			  /* should usage be printed? */
	  usage(0);

     if (OPTtotaltime)
	  starttime = clock();		  /* beginning time */

     if (OPTversion)			  /* should version be printed? */
	  printversion();

     setusedobjects();			  /* get compiler and things really */
					  /* used (different for C, PVM, MPL) */

     if (!OPTnop3inc)			  /* add default include paths? */
     {
	  ep = getenv("P3INC");
	  if (ep!=NULL)
	  {
	       setpaths(&p3pathlist, ep, usedp3inc);
	       setpaths(&cpathlist, ep, usedp3inc);
	  }
	  else
	  {
	       setpaths(&p3pathlist, usedp3inc, NULL);
	       setpaths(&cpathlist, usedp3inc, NULL);
	  }
     }
     if (!OPTnop3lib)			  /* add default library paths? */
     {
	  ep = getenv("P3LIB");
	  if (ep!=NULL)
	       setpaths(&libpathlist, ep, usedp3lib);
	  else
	       setpaths(&libpathlist, usedp3lib, NULL);
     }

     replaceenvvars(&p3pathlist);	  /* replace environment variables */
     replaceenvvars(&cpathlist);	  /* in the path lists (e.g. */
     replaceenvvars(&libpathlist);	  /* $PVM_ROOT or $PVM_ARCH) */

     setp3commandline();		  /* create p3c commandline */
     compilep3files(&arguments);	  /* compile all Parallaxis files */

     if (OPTtotaltime)
	  p3time = clock();		  /* time to Parallaxis compile */

     if (!OPTp3only)
     {
	  compilecfiles(&arguments);	  /* call C-compiler */

#ifdef MPL_VERSION
	  /* for large model call C- or MPL-compiler here (2nd pass) */
#endif
     }

     if ((OPTtotaltime) && (!OPTnocompile))
     {
	  totaltime = clock();
	  fprintf(stderr, "Time for Parallaxis compiler %s: %8.2f s\n",
		  usedp3c, (float)(p3time-starttime)/CLOCKS_PER_SEC);
	  fprintf(stderr, "Time for backend compiler %s: %8.2f s\n",
		  p3cc, (float)(totaltime-p3time)/CLOCKS_PER_SEC);
	  fprintf(stderr, "==> Total time for compiling: %8.2f s\n",
		  (float)(totaltime-starttime)/CLOCKS_PER_SEC);
     }

#if 0
     {
	  printf("Remaining arguments:\n");
	  printstringarray(&arguments);

	  printf("Default include path: %s\n", p3inc);
	  printf("Default library path: %s\n", p3lib);

     }
#endif

     return 0;				  /* successfully finished */
}

