/************************************************************************/
/*									*/
/*	If the library function alloca is available and the C programs	*/
/*	are compiled with the flag -DStackAlloc, then the memory space	*/
/*	for open array value parameters will be allocated in the stack	*/
/*	frame of the corresponding procedure. This temporary space	*/
/*	will be freed automatically when the procedure returns.		*/
/*	Otherwise, malloc and free will be used to allocate and		*/
/*	deallocate memory space for open array value parameters.	*/
/*									*/
/*	By default open array value parameters are copied using the	*/
/*	library function memcpy. This can be switched to bcopy by	*/
/*	compiling with the flag -DBCOPPKY on systems without memcpy.	*/
/*									*/
/************************************************************************/

#ifndef _SYSTEM_PVM_h
#define _SYSTEM_PVM_h


#include <math.h>			  /* e.g. M_PI */
#include <string.h>			  /* e.g. strncpy(), memcpy() */
#include <stdlib.h>			  /* e.g. exit() */
#include <limits.h>			  /* e.g. DBL_MIN */
/* communication-signals (WriteCardSig, ...) and PVM-macros */
#include "shared_macros.h"
/* include PVM-Messagepassing and Task-Library */
/* #include <pvm3.h> is already done in shared_macros! */


#ifdef __sun__			          /* Our SUNs don't have headers */
extern double drand48(void);		  /* which define the rand48() */
extern long   mrand48(void);		  /* functions. :-( */
extern void   srand48(long);		  /* Usually they are in stdlib.h */
#endif


/* Definition of library functions **************************************/

#ifdef StackAlloc
#include <alloca.h>			  /* alloca() */
#else
#include <malloc.h>			  /* malloc(), free() */
#endif

# ifdef __STDC__
# define ARGS(parameters)       parameters
# else
# define ARGS(parameters)       ()
# endif

/* Definition of standard constants *************************************/

#ifndef FALSE
#define FALSE           ((BOOLEAN) 0)
#endif
#ifndef TRUE
#define TRUE            ((BOOLEAN) 1)
#endif
#define NIL             0L
#define EOL             '\n'
#ifdef M_PI
#define pi              M_PI   /*3.141592653589793116*/
#else
#define pi              3.141592653589793116
#endif

/* Definition of basic types ********************************************/

typedef long            INTEGER;
typedef unsigned long	CARDINAL;
typedef unsigned char   BOOLEAN;
typedef unsigned char   CHAR;
typedef double          REAL;

/* Definition of standard types *****************************************/

typedef unsigned long   BITSET;
typedef void            (*PROC)();
typedef unsigned char   WORD;
typedef WORD            BYTE;
#ifdef __STDC__
typedef void           *ADDRESS;
#else
typedef char           *ADDRESS;
#endif

/* Definition of parallaxis types ***************************************/

typedef struct S_1 {
  struct S_1 *Previous;
  CHAR       *Elem;
} BOOLEAN_ELEM, *BOOLEAN_ELEM_PTR;

typedef struct {
  CARDINAL ElemSize;          /* number of bytes needed for one StackElem */
  BOOLEAN_ELEM_PTR StackElem;
} ACTIVE_STACK, *ACTIVE_STACK_PTR, ELSE_STACK, *ELSE_STACK_PTR;

typedef struct {
  CARDINAL start_ID;
  CARDINAL end_ID;
  CARDINAL num_PEs;     /* = end_ID - start_ID + 1 */
  CARDINAL num_breaks;  /* number of linebreaks in ID-range when printing */
} TASK_DATA;

typedef struct {
  CARDINAL NumConnections;
  CARDINAL *PEnumbers;
} TASK_LINKS;

typedef struct {
  TASK_LINKS *ToTask;
  TASK_LINKS *FromTask;
} LINK_ARRAY;

typedef struct S_2 {
  struct S_2 *Next;
  CARDINAL SourceID;
  CARDINAL DestinationID;
} LINK;

typedef struct {
  struct S_6 *SourceConfig;
  struct S_6 *DestinationConfig;
  void *Links;            /* void because it has to point to different types */
                          /* (first to LINK, then to TASK_LINK               */
} CONNECTION;

typedef struct S_3 {
  struct S_3 *Next;
  INTEGER Parameter;
  void *Links;            /* void because it has to point to different types */
                          /* (first to LINK, then to TASK_LINK               */
} PARAMETER_LINK;

typedef struct {
  struct S_6 *SourceConfig;
  struct S_6 *DestinationConfig;
  PARAMETER_LINK *ParameterLinks;
} PARAMETER_CONNECTION;

typedef struct S_4 {
  struct S_4 *Next;
  char ConnectionName[256];
  struct S_6 *Configuration;
  void *Links;            /* void because it has to point to different types */
                          /* (first to LINK, then to TASK_LINK               */
} GENERIC_CONNECTIONS, *GENERIC_CONNECTIONS_PTR;

typedef struct S_5 {
  struct S_5 *Next;
  char ConnectionName[256];
  struct S_6 *Configuration;
  PARAMETER_LINK *ParameterLinks;
} GENERIC_PARAMETER_CONNECTIONS, *GENERIC_PARAMETER_CONNECTIONS_PTR;

typedef struct S_6 {
  char Name[256];
  CARDINAL Len_conf;    /* number of PEs                                     */
  CARDINAL Rank;        /* number of Dimensions                              */
  CARDINAL *Len;        /* array containing length of each Dimension         */
  INTEGER  *Lower;      /* array containig the lower bound of each Dimension */
  INTEGER  *Upper;      /*                     upper                         */
  CARDINAL u_proc;      /* number of processors used for this Configuration  */
  TASK_DATA *ID_range;  /* array containing ID-Range of each Task           */
  ACTIVE_STACK_PTR ActiveStack; /* pointer to active-stack */
  ELSE_STACK_PTR ElseStack;
  struct GENERIC_CONNECTIONS *Generics;
  struct GENERIC_PARAMETER_CONNECTIONS *GenericsParam;
} CONFIGURATION;

/* Definition of standard functions *************************************/

#define ABS(x)		((x) < 0 ? -(x) : (x))
#define ABSC(x)	        ((CARDINAL) (x))
#define CHR(x)		((CHAR) (x))
#define FLOAT(x)	((REAL) (x))
#define ORD(x)		((CARDINAL) (x))
#define TRUNC(x)	((CARDINAL) (x))
#define VAL(T,x)	((T) (x))
#define MOD(x, y)       (((x) % (y) < 0) ? (x) % (y) + ((y) < 0 ? -(y) : (y)) : (x) % (y))


extern INTEGER ABSI ARGS((INTEGER x));
extern REAL ABSR ARGS((REAL x));
extern CHAR CAP ARGS((CHAR ch));
extern CARDINAL MOD1 ARGS((INTEGER x, INTEGER y));

/* INTEGER limits */
#ifdef LONG_MIN
#define MIN_INTEGER     LONG_MIN
#else
#define MIN_INTEGER	(-2147483648L)
#endif
#ifdef LONG_MAX
#define MAX_INTEGER     LONG_MAX
#else
#define MAX_INTEGER	2147483647L
#endif

/* CARDINAL limits */
#define MIN_CARDINAL	((CARDINAL)0L)
#ifdef ULONG_MAX
#define MAX_CARDINAL    ULONG_MAX
#else
#define MAX_CARDINAL	((CARDINAL)4294967295L)
#endif

/* BOOLEAN limits */
#define MIN_BOOLEAN	FALSE
#define MAX_BOOLEAN	TRUE

/* CHAR limits */
#define MIN_CHAR	((CHAR)'\0')
#define MAX_CHAR	((CHAR)'\377')

/* REAL limits */
#ifdef DBL_MIN
#define MIN_REAL        DBL_MIN
#else
#define MIN_REAL	4.94065645841246544e-324
#endif
#ifdef DBL_MAX
#define MAX_REAL        DBL_MAX
#else
#define MAX_REAL	1.79769313486231470e+308
#endif

/*********** Definition of Parallaxis standard functions ***********/

#define ODD(x)		((BOOLEAN)((x) & 01))

#define EVEN(x)         (!((BOOLEAN)((x) & 01)))
#define LEN(x)          (x).Len_conf
#define LEN1(x,y)       (x).Len[(x).Rank - (y)]
#define RANK(x)         (x).Rank
#define LOWER(x,y)      (x).Lower[(x).Rank - (y)]
#define UPPER(x,y)      (x).Upper[(x).Rank - (y)]

extern CARDINAL argc ARGS((void));

#define round(r)        ((INTEGER) ((r) + 0.5))
#define RandomCard1     ((CARDINAL) mrand48())
#define RandomInt1      ((INTEGER) mrand48())
#define RandomReal1     ((REAL) drand48())
#define RandomBool1     ((BOOLEAN) ((mrand48() & 01000) == 0))
#define RandomChar1     ((CHAR) (mrand48() & 0xff))
/* scalar random-functions */
extern CARDINAL RandomCard ARGS((void));
extern INTEGER  RandomInt  ARGS((void));
extern REAL     RandomReal ARGS((void));
extern BOOLEAN  RandomBool ARGS((void));
extern CHAR     Random ARGS((void));

extern CARDINAL DIM2ID ARGS((CONFIGURATION Config, INTEGER *Dim));
extern INTEGER *DIM ARGS((CONFIGURATION Config, CARDINAL Dim));
extern CARDINAL *ID ARGS((CONFIGURATION Config));

extern REAL ArcTan2 ARGS((REAL x, REAL y));

/* Definition of standard procedures ************************************/

#define INC(x)		(x)++
#define INC1(x,n)	x += n
#define DEC(x)		(x)--
#define DEC1(x,n)	x -= n

#define EXCL(s,i)	s &= ~(0X1L << (i))
#define INCL(s,i)	s |= 0X1L << (i)

#define ADR(x)		((ADDRESS) &(x))
#define ADR1(x)		((ADDRESS) (x))

/* Definition of parallaxis standard procedures *************************/

extern void argv ARGS((CARDINAL index, CHAR arg[]));

extern void CloseInput ARGS((void));
extern void CloseOutput ARGS((void));
extern void OpenInput ARGS((CHAR *s));
extern void OpenOutput ARGS((CHAR *s));

extern void Read ARGS((CHAR *c));
extern void ReadCard ARGS((CARDINAL *c));
extern void ReadInt ARGS((INTEGER *c));
extern void ReadReal ARGS((REAL *r));
extern void ReadBool ARGS((BOOLEAN *b));
extern void ReadString ARGS((CHAR *s));
extern void Read1 ARGS((CONFIGURATION Config, CHAR *char_value));
extern void ReadCard1 ARGS((CONFIGURATION Config, CARDINAL *card_value));
extern void ReadInt1  ARGS((CONFIGURATION Config, INTEGER  *int_value));
extern void ReadReal1 ARGS((CONFIGURATION Config, REAL *real_vect));
extern void ReadBool1 ARGS((CONFIGURATION Config, BOOLEAN *bool_vect));
extern void ReadString1 ARGS((CONFIGURATION Config, CHAR **string_vect));

extern void WriteLn     ARGS((void));
extern void WriteCard   ARGS((CARDINAL card_value, CARDINAL length));
extern void WriteInt    ARGS((INTEGER int_value, CARDINAL length));
extern void WriteReal   ARGS((REAL real_value, CARDINAL length));
extern void WriteFixPt  ARGS((REAL real, CARDINAL length, CARDINAL decimals));
extern void WriteBool   ARGS((BOOLEAN bool_value, CARDINAL length));
extern void Write       ARGS((CHAR char_value));
extern void WriteString ARGS((CHAR *string_value));

extern void WriteCard1 ARGS((CONFIGURATION Config, CARDINAL card_value[], CARDINAL length));
extern void WriteInt1 ARGS((CONFIGURATION Config, INTEGER int_value[], CARDINAL length));
extern void WriteReal1 ARGS((CONFIGURATION Config, REAL real_value[], CARDINAL l));
extern void WriteFixPt1 ARGS((CONFIGURATION Config, REAL real_value[], CARDINAL length, CARDINAL decimals));
extern void WriteBool1 ARGS((CONFIGURATION Config, BOOLEAN bool_value[], CARDINAL length));
extern void Write1 ARGS((CONFIGURATION Config, CHAR char_value[]));
extern void WriteString1 ARGS((CONFIGURATION Config, CHAR **string_value));
/*
extern void WriteStringChar1 ARGS((CONFIGURATION Config, CHAR s[]));
*/


/* Definition of (some) set operators ***********************************/

#define SYSTEM_MaxSet		(sizeof (unsigned long) * 8 - 1)

#define SET_ELEM(el)		(0X1UL << (el))
#define SET_cRNG(lo,hi) \
	((lo) <= (hi) ? ~0X0UL >> (lo) << (lo) + SYSTEM_MaxSet - (hi) >> SYSTEM_MaxSet - (hi) : 0X0UL)
#define SET_RANGE(lo,hi)	SET_RANGE1((CARDINAL)(lo), (CARDINAL)(hi))

extern unsigned long SET_RANGE1 ARGS((CARDINAL lo, CARDINAL hi));

#define SET_DIFF(s1,s2)		((s1) & ~(s2))
#define IN(x,s)			((BOOLEAN)((s) >> (x) & 0X1UL))
#define SET_IS_SUBSET1(s1,s2)	((BOOLEAN)!((s1) & ~(s2)))
#define SET_IS_SUBSET2(s1,s2)	((BOOLEAN)!((s2) & ~(s1)))

/* Definition of compiler constants *************************************/

#define SYSTEM_ALIGN		8
#define SYSTEM_MASK             (~(SYSTEM_ALIGN - 1))

/* Definition of compiler types *****************************************/

typedef ADDRESS		OPAQUE;

/* Definition of compiler variables *************************************/

extern unsigned int SYSTEM_argc;
extern char **SYSTEM_argv;
extern char **SYSTEM_envp;

extern char termCH;
extern BOOLEAN Done;

extern CARDINAL PE;        /* counter over all local PEs             */
extern INTEGER *SOURCE_ARRAY;      /* for computation of connections */
extern INTEGER *DESTINATION_ARRAY; /* dito                           */
extern CONFIGURATION CONFIG_1;     /* for REDUCE-functions           */

extern int my_tid;         /* own tid                                */
extern int parent_tid;     /* tid of parent-process                  */
extern int io_tid;         /* tid of io-process                      */
extern CARDINAL n_tasks;   /* number of elements in task_ids-array   */
extern int *task_ids;      /* array containing tids of all tasks     */
extern int me;             /* index of own tid in task_ids           */
extern int message_tag;    /* tag used for data-exchange-messages    */
extern CARDINAL MaxDiff;   /* maximum difference of PEs of two tasks */
extern CARDINAL MinPEs;    /* minimum number of PEs of one task (not granted)*/
/* for an more detailed explanation of MaxDiff and MinPEs look at the header */
/* of INIT_ID_RANGE */

/* Definition of compiler macros and functions **************************/

#ifdef StackAlloc

#define OPEN_ARRAY_LOCALS	char *FREE_POINTER;
#define ALLOC_OPEN_ARRAYS(size, arrays) \
	FREE_POINTER = alloca((int)((size) + (arrays) * (SYSTEM_ALIGN - 1)));
#define FREE_OPEN_ARRAYS

#else

#define OPEN_ARRAY_LOCALS	char *BLOCK_POINTER, *FREE_POINTER;
#define ALLOC_OPEN_ARRAYS(size, arrays)	\
	BLOCK_POINTER = FREE_POINTER = \
		malloc((unsigned)((size) + (arrays) * (SYSTEM_ALIGN - 1)));
#define FREE_OPEN_ARRAYS	free(BLOCK_POINTER);

#endif

#ifdef BCOPY
#define COPY_OPEN_ARRAY(array, elems, type) \
	{ int ARRAY_SIZE = elems * sizeof(type); \
	  bcopy((char *)array, FREE_POINTER, ARRAY_SIZE); \
	  array = FREE_POINTER; \
	  FREE_POINTER += (ARRAY_SIZE + (SYSTEM_ALIGN - 1)) & SYSTEM_MASK; \
	}
#else
#define COPY_OPEN_ARRAY(array, elems, type) \
	{ int ARRAY_SIZE = elems * sizeof(type); \
	  array = (type *)memcpy(FREE_POINTER, (char *)array, ARRAY_SIZE); \
	  FREE_POINTER += (ARRAY_SIZE + (SYSTEM_ALIGN - 1)) & SYSTEM_MASK; \
	}
#endif

#define FOR_LIMIT_UP(last, step, min) \
	((last) < (min) + ((step) - 1) ? (min) : (last) - ((step) - 1))

#define FOR_LIMIT_DOWN(last, step, max) \
	((last) > (max) + ((step) + 1) ? (max) : (last) - ((step) + 1))

extern void CaseError ARGS((char file[], int line));
extern void ReturnError ARGS((char file[], int line));
extern INTEGER CHECK_PE_NUMBER ARGS((CONFIGURATION Config, INTEGER PeNumber, char file[], int line));

/*Definition of parallaxis compiler macros and functions ***************/

extern void INIT_MODULE ARGS((int argc, char *argv[], char *envp[]));
extern void EXIT_MODULE ARGS((void));

/* shortcut for the local number of PEs */
#define LOCAL_NUM_PEs(Config) (Config).ID_range[me].num_PEs 

/* loop over all local PEs which are active */
/* example: FOR_ALL_LOCAL_ACTIVES(grid) { action(PE); } */
/* uses the global Variable PE as loop-counter */
#define FOR_ALL_LOCAL_ACTIVES(Config)                    \
        if (me < Config.u_proc)                          \
          for (PE = 0; PE < LOCAL_NUM_PEs(Config); PE++) \
            if (LOCAL_IS_ACTIVE(Config, PE))

/* loop over all local PES even the inactive ones */
/* example: FOR_ALL_LOCAL_PEs(grid) { action(PE); } */
/* uses the local Variable PE as loop-counter */
#define FOR_ALL_LOCAL_PEs(Config)                        \
        if (me < Config.u_proc)                          \
          for (PE = 0; PE < LOCAL_NUM_PEs(Config); PE++)

/* check for configuration bounds */
extern BOOLEAN IS_IN_CONFIG ARGS((CONFIGURATION Config, INTEGER *IndexArray));

#define MAX(x, y) (x) > (y) ? (x) : (y)
#define MIN(x, y) (x) < (y) ? (x) : (y)

extern CARDINAL TASK_INDEX_HOLDING_ID ARGS((CARDINAL id, CONFIGURATION Config));

/* 2 functions for LOAD */
extern CARDINAL NUM_ACTIVES_OF_PREDS ARGS((CONFIGURATION Config));
extern CARDINAL NUM_ELEMS_LOADED ARGS((CONFIGURATION Config, CARDINAL
				       n_elems, BOOLEAN send_counter));

/* 2 procedures for STORE */
extern void PREPARE_UNPACKING ARGS((CARDINAL *num_unpacked, CARDINAL *num_data,
				    CARDINAL *t_ind, BOOLEAN *values_available,
				    unsigned int type_size,
				    CONFIGURATION Config));
extern void FREE_MSG_BUFFERS ARGS((CARDINAL t_ind, CONFIGURATION Config));

/* configuration */

extern void INIT_CONFIGURATION ARGS((char Name[], CONFIGURATION *Config, CARDINAL Rank));
/* Initializes a Configuration with the value of Rank. */
/* Also allocates memory for the Lower-, Upper- and Len-array. */
/* Then it receives the start_ID, end_ID and num_PEs for each */
/* of the Config->u_proc tasks of this Configuration. */

extern void SET_BOUNDS ARGS((CONFIGURATION *Config, CARDINAL Dim, INTEGER Lwb, INTEGER Upb));
/* Initializes one field of the Lower-, Upper- and Len-array of Config. */
/* with the Index Dim. The value of Len_conf is updated. */
/* Take care that the first dimension has the index 0! */

extern void INIT_ID_RANGE ARGS((CONFIGURATION *Config, CARDINAL num_real, CARDINAL MaxDiff, CARDINAL MinPEs));
/* This procedure distributes the start_ID, end_ID and num_PEs to each      */
/* of the num_real tasks (except that there are less than num_real PEs      */
/* in this Configuration.                                                   */
/* The Configuration must already be initialized (uses Config->Len_conf). */
/* MaxDiff is the maximum difference in the number of PEs that are simulated */
/*         from two different Tasks. Is MUST be >= 1 !!! */
/* MinPEs  is the minmal number of PEs that are simulated by one Task. */
/*         MinPEs MUST be less than MaxDiff! If MinPEs is less than */
/*         MaxDiff/2, then it is granted that one task simulates at least */
/*         MinPEs, otherwise it may be less. */

extern void COPY_ID_RANGE ARGS((CONFIGURATION *DestConfig, CONFIGURATION
				SourceConfig, CARDINAL num_real));
/* is used instead of INIT_ID_RANGE when several Configuration share the */
/* same PEs                                                              */

extern void INIT_STACKS ARGS((CONFIGURATION *Config));
/* Inits all required Stacks of Config (active-stack) */

extern void COPY_STACKS ARGS((CONFIGURATION *DestConfig, CONFIGURATION SourceConfig));
/* is used instead of INIT_STACKS when several Configurations share the */
/* same PEs                                                             */

extern void FREE_CONFIGURATION ARGS((CONFIGURATION *Config, BOOLEAN FreeStacks));
/* frees the allocated memory of the active-stack, the Lower-, Upper- */
/* and Len-array                                                      */


/* connection without index */

extern void INIT_CONNECTION ARGS((CONNECTION *Connection, CONFIGURATION *SourceConfig, CONFIGURATION *DestinationConfig));
extern void REGISTER_LINK ARGS((CONNECTION *Connection, INTEGER *SourceArray, INTEGER *DestinationArray));
extern LINK_ARRAY *GET_LINKS ARGS((CONNECTION Connection));
extern void FREE_CONNECTION ARGS((CONNECTION *Connection));

/* connection with index */

extern void INIT_PARAMETER_CONNECTION ARGS((PARAMETER_CONNECTION *Connection, CONFIGURATION *SourceConfig, CONFIGURATION *DestinationConfig));
extern void REGISTER_PARAMETER_LINK ARGS((PARAMETER_CONNECTION *Connection, INTEGER Index, INTEGER *SourceArray, INTEGER *DestinationArray));
extern LINK_ARRAY *GET_PARAMETER_LINKS ARGS((PARAMETER_CONNECTION Connection, INTEGER Parameter));
extern void FREE_PARAMETER_CONNECTION ARGS((PARAMETER_CONNECTION *Connection));

/* generic connection without index */

extern void INIT_GENERIC_CONNECTION ARGS((char Connection[], CONFIGURATION *Configuration));
extern void REGISTER_GENERIC_LINK ARGS((char ConnectionName[], CONFIGURATION *Configuration, INTEGER *SourceArray, INTEGER *DestinationArray));
extern LINK_ARRAY *GET_GENERIC_LINKS ARGS((char Connection[], CONFIGURATION *Configuration));
extern void FREE_GENERIC_CONNECTION ARGS((char Connection[], CONFIGURATION *Configuration));

/* generic connection with index */

extern void INIT_GENERIC_PARAMETER_CONNECTION ARGS((char Connection[], CONFIGURATION *Configuration));
extern void REGISTER_GENERIC_PARAMETER_LINK ARGS((char ConnectionName[], CONFIGURATION *Configuration, INTEGER Parameter, INTEGER *SourceArray, INTEGER *DestinationArray));
extern LINK_ARRAY *GET_GENERIC_PARAMETER_LINKS ARGS((char Connection[], CONFIGURATION *Configuration, INTEGER Parameter));
extern void FREE_GENERIC_PARAMETER_CONNECTION ARGS((char Connection[], CONFIGURATION *Configuration));

/* procedures to organize connections in a way that messages to each task */
/* can be easily produced */
extern void REORGANIZE_CONNECTION ARGS((CONNECTION *Connect));
extern void REORGANIZE_PARAMETER_CONNECTION ARGS((PARAMETER_CONNECTION *Connect));
extern void REORGANIZE_GENERIC_CONNECTION ARGS((char ConnectionName[], CONFIGURATION *Configuration));
extern void REORGANIZE_GENERIC_PARAMETER_CONNECTION ARGS((char ConnectionName[], CONFIGURATION *Configuration));

/* check for configuration bounds */

extern BOOLEAN IS_IN_LOCAL_ID_RANGE ARGS((CONFIGURATION Config, CARDINAL id));

/* active stack */
extern void PUSH_ACTIVE_ALL ARGS((CONFIGURATION Config));
extern void PUSH_ACTIVE ARGS((CONFIGURATION Config));
extern void POP_ACTIVE ARGS((CONFIGURATION Config));
extern void TOP_ACTIVE ARGS((CONFIGURATION Config));
extern BOOLEAN_ELEM_PTR READ_ACTIVE ARGS((CONFIGURATION Config));
extern void RESET_ACTIVE ARGS((CONFIGURATION Config, BOOLEAN_ELEM_PTR Pointer));
extern void SET_INACTIVE ARGS((CONFIGURATION Config, CARDINAL local_Processor));
/* sets the status of one Processor of Config to inactive */
/* the Parameter local_Processor has the range            */
/* from 0 to Config.ID_range[me].num_PEs - 1 (!)          */

extern void EXIT_ALL_ACTIVES ARGS((CONFIGURATION Config, BOOLEAN_ELEM_PTR ActivePtr, BOOLEAN_ELEM_PTR ElsePtr));
extern BOOLEAN LOCAL_IS_ACTIVE ARGS((CONFIGURATION Config, 
				     CARDINAL local_Processor));
/* returns the active-status of one Processor of Config */
/* the Parameter local_Processor has the range           */
/* from 0 to Config.ID_range[me].num_PEs - 1 (!)         */
extern BOOLEAN IS_ANY_ACTIVE ARGS((CONFIGURATION Config));

/* wird nicht gebraucht, koennte man (Hartmut) aber mal einbauen,  */
/* weil es effizienter ist. */
/*
extern void INVERT_ACTIVE ARGS((CONFIGURATION Config));
*/
/* inverts the active status of these PEs which are now active or */
/* inactive and were active before */

/* else stack */

extern void INIT_ELSE_STACK ARGS((CONFIGURATION *Config));
extern void FREE_ELSE_STACK ARGS((CONFIGURATION *Config));
extern void PUSH_ELSE ARGS((CONFIGURATION Config));
extern void POP_ELSE ARGS((CONFIGURATION Config));
extern BOOLEAN_ELEM_PTR READ_ELSE ARGS((CONFIGURATION Config));
extern void RESET_ELSE ARGS((CONFIGURATION Config, BOOLEAN_ELEM_PTR Pointer));
extern void COPY_ELSE_TO_ACTIVE ARGS((CONFIGURATION Config));
extern void UNSET_ELSE ARGS((CONFIGURATION Config, CARDINAL Processor));


#endif
