/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  BigSet.h
 * desc  :  sets of integer values, unlimited in size
 * Author:  lampke
 *
 ************************************************************************/

#ifndef BigSet_h
#define BigSet_h

#include <stdio.h>
#include "Sets.h"			  /* Reuse-Library */
#include "parallaxis.h"

/*
 * definition bitset
 */
/* defined in Sets.h
 * typedef ULONG BITSET;
 */
#define BITSPERBITSET   BITSPERLONG
#define EMPTYBITSET     0L

#define BITSET_Include(set,elem)           ((set) |= (0x00000001 << (elem)))
#define BITSET_Exclude(set,elem)           ((set) &= ~(0x00000001 << (elem)))
#define BITSET_IsElement(set,elem)         ((set) & (0x00000001 << (elem)))
#define BITSET_Complement(set)             (~(set))
#define BITSET_Union(set1,set2)            ((set1) | (set2))
#define BITSET_Intersect(set1,set2)        ((set1) & (set2))
#define BITSET_Diff(set1,set2)             ((set1) & BITSET_Complement(set2))
#define BITSET_SymDiff(set1,set2)          ((set1) ^ (set2))
#define BITSET_IsEqual(set1,set2)          ((set1) == (set2))
#define BITSET_IsSubset(set1,set2)         (BITSET_Diff((set1),(set2)) == 0)
#define BITSET_IsStrictSubset(set1,set2)   (BITSET_IsSubset((set1),(set1)) &&\
                                            !BITSET_IsEqual((set1),(set1)))

typedef struct  {
   ULONG    maxElem;
   ULONG    lastBitset;
   ULONG    firstElem;
   ULONG    lastElem;
   BITSET  *bitsetPtr;
   SHORT    card;
} tBigSet;

typedef BOOLEAN (*tBigSetTestFunc)( ULONG elem );
typedef void (*tBigSetDoFunc)( ULONG elem );

/*************************************************************************
 * name :  WriteSet
 * in   :  pointer to open file and BITSET
 * out  :
 ************************************************************************/
extern void BITSET_WriteSet( FILE *f, BITSET set );

/*************************************************************************
 * name :  MakeSet
 * in   :  pointer to set
 *         maximale number of elements
 * out  :  pointer to initialised empty set of size maxsize
 ************************************************************************/
extern tBigSet *BigSet_MakeSet( tBigSet *set, ULONG maxsize );

/*************************************************************************
 * name :  Union
 * in   :  pointer to two sets
 * out  :  pointer to first set, which is the union of the two sets
 ************************************************************************/
extern tBigSet *BigSet_Union( tBigSet *set1, tBigSet *set2 );

/*************************************************************************
 * name :  Difference
 * in   :  pointer to two sets
 * out  :  pointer to first set, which is the difference of the two sets
 ************************************************************************/
extern tBigSet *BigSet_Difference( tBigSet *set1, tBigSet *set2 );

/*************************************************************************
 * name :  Intersection
 * in   :  pointer to two sets
 * out  :  pointer to first set, which is the intersection of the two sets
 ************************************************************************/
extern tBigSet *BigSet_Intersection( tBigSet *set1, tBigSet *set2 );

/*************************************************************************
 * name :  SymDifference
 * in   :  pointer to two sets
 * out  :  pointer to first set, which is the symetric difference
 *         of the two sets
 ************************************************************************/
extern tBigSet *BigSet_SymDifference( tBigSet *set1, tBigSet *set2 );

/*************************************************************************
 * name :  Complement
 * in   :  pointer to set
 * out  :  pointer to same set, which is complemented
 ************************************************************************/
extern tBigSet *BigSet_Complement( tBigSet *set );

/*************************************************************************
 * name :  Include
 * in   :  pointer to set
 *         element to insert
 * out  :  pointer to same set, now containing element elem
 ************************************************************************/
extern tBigSet *BigSet_Include( tBigSet *set, ULONG elem );

/*************************************************************************
 * name :  Exclude
 * in   :  pointer to set
 *         element to remove
 * out  :  pointer to same set, with element elem removed
 ************************************************************************/
extern tBigSet *BigSet_Exclude( tBigSet *set, ULONG elem );

/*************************************************************************
 * name :  Card
 * in   :  pointer to set
 * out  :  number of elements actually in set
 ************************************************************************/
extern ULONG BigSet_Card( tBigSet *set );

/*************************************************************************
 * name :  Size
 * in   :  pointer to set
 * out  :  maximal size of set
 ************************************************************************/
extern ULONG BigSet_Size( tBigSet *set );

/*************************************************************************
 * name :  Minimum
 * in   :  pointer to set
 * out  :  minimal element in set
 ************************************************************************/
extern ULONG BigSet_Minimum( tBigSet *set );

/*************************************************************************
 * name :  Maximum
 * in   :  pointer to set
 * out  :  maximum element in set
 ************************************************************************/
extern ULONG BigSet_Maximum( tBigSet *Set );

/*************************************************************************
 * name :  Select
 * in   :  pointer to set
 * out  :  selects an arbitrary element in set
 ************************************************************************/
extern ULONG BigSet_Select( tBigSet *set );

/*************************************************************************
 * name :  Extract
 * in   :  pointer to set
 * out  :  selects and removes an arbitrary element in set
 ************************************************************************/
extern ULONG BigSet_Extract( tBigSet *set );

/*************************************************************************
 * name :  IsSubset
 * in   :  pointer to two sets
 * out  :  TRUE, if set1 is subset of set2
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN BigSet_IsSubset( tBigSet *set1, tBigSet *set2 );

/*************************************************************************
 * name :  IsStrictSubset
 * in   :  pointer to two sets
 * out  :  TRUE, if set1 is a strict subset of set2
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN BigSet_IsStrictSubset( tBigSet *set1, tBigSet *set2 );

/*************************************************************************
 * name :  IsEqual
 * in   :  pointer to two sets
 * out  :  TRUE, if set1 is equal to set2
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN BigSet_IsEqual( tBigSet *set1, tBigSet *set2 );

/*************************************************************************
 * name :  IsNotEqual
 * in   :  pointer to two sets
 * out  :  TRUE, if set1 is not equal to set2
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN BigSet_IsNotEqual( tBigSet *set1, tBigSet *set2 );

/*************************************************************************
 * name :  IsElement
 * in   :  pointer to set and element 'elem'
 * out  :  TRUE, if element 'elem' is element of set
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN BigSet_IsElement( tBigSet *set, ULONG elem );

/*************************************************************************
 * name :  IsEmpty
 * in   :  pointer to set
 * out  :  TRUE, if set is empty
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN BigSet_IsEmpty( tBigSet *set );

/*************************************************************************
 * name :  Forall
 * in   :  pointer to set and pointer to test function
 * out  :  TRUE, if function 'func' is true for all elements in the set
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN BigSet_Forall( tBigSet *set, tBigSetTestFunc func );

/*************************************************************************
 * name :  Exists
 * in   :  pointer to set and pointer to test function
 * out  :  TRUE, if function 'func' is true at least one element in the set
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN BigSet_Exists( tBigSet *set, tBigSetTestFunc func );

/*************************************************************************
 * name :  Exists1
 * in   :  pointer to set and pointer to test function
 * out  :  TRUE, if function 'func' is true at exact one element in the set
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN BigSet_Exists1( tBigSet *set, tBigSetTestFunc func );

/*************************************************************************
 * name :  Assign
 * in   :  pointer to two set
 * out  :  pointer to first set with second set assigned to it
 ************************************************************************/
extern tBigSet *BigSet_Assign( tBigSet *set1, tBigSet *set2 );

/*************************************************************************
 * name :  AssignElement
 * in   :  pointer to set and element
 * out  :  pointer to set containing only element 'elem'
 ************************************************************************/
extern tBigSet *BigSet_AssignElement( tBigSet *set, ULONG elem );

/*************************************************************************
 * name :  AssignEmpty
 * in   :  pointer to set
 * out  :  pointer to set containing no element
 ************************************************************************/
extern tBigSet *BigSet_AssignEmpty( tBigSet *set );

/*************************************************************************
 * name :  ForallDo
 * in   :  pointer to set
 * out  :
 ************************************************************************/
extern void BigSet_ForallDo( tBigSet *set, tBigSetDoFunc func );

/*************************************************************************
 * name :  ReadSet
 * in   :  pointer to open file and pointer to set
 * out  :
 ************************************************************************/
extern void BigSet_ReadSet( FILE *f, tBigSet *set );

/*************************************************************************
 * name :  WriteSet
 * in   :  pointer to open file and pointer to set
 * out  :
 ************************************************************************/
extern void BigSet_WriteSet( FILE *f, tBigSet *set );

/*************************************************************************
 * name :  ReleaseSet
 * in   :  pointer to set, which is to be destroyed
 * out  :
 ************************************************************************/
extern void BigSet_ReleaseSet( tBigSet *set );

extern void BigSet_Init(void);

# endif
