/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name   : parallaxis.rex
 * desc   : Scanner Specification for Parallaxis
 *          generated scanner recognizes tokens and computes their
 *          attributes
 * Authors: kappel, lampke
 *
 ************************************************************************/

EXPORT {

#include "StringMem.h"			  /* Reuse-Library */
#include "Idents.h"			  /* Reuse-Library */
#include "parallaxis.h"
#include "Positions.h"
#include "Errors.h"
#include "Tree.h"

#if defined __STDC__
#define ARGS(parameters)       parameters
#else
#define ARGS(parameters)       ()
#endif

typedef struct {
  tPosition yyPos;
  tIdent Ident;
} yyIdent;

typedef struct {
  tPosition yyPos;
  unsigned long Integer;
  unsigned long Kind;
} yyIntConst;

typedef struct {
  tPosition yyPos;
  tStringRef String;
} yyRealConst;

typedef struct {
  tPosition yyPos;
  unsigned char Char;
} yyCharConst;

typedef struct {
  tPosition yyPos;
  tStringRef String;
} yyStringConst;

typedef union {
  tPosition Position;
  yyIdent Ident;
  yyIntConst IntConst;
  yyRealConst RealConst;
  yyCharConst CharConst;
  yyStringConst StringConst;
} tScanAttribute;

extern void ErrorAttribute ARGS((int Token, tScanAttribute * pAttribute));
/* Returns in parameter 'Attribute' */
/* default values for the attributes of token 'Token' */

}

/************************************************************************/

GLOBAL {

#include <math.h>
#include <string.h>
#include <ctype.h>

static char Word[256];
static int NestingLevel = 0;

void ErrorAttribute
#if defined __STDC__
(int Token, tScanAttribute *pAttribute)
#else
(Token, pAttribute) int Token; tScanAttribute *pAttribute;
#endif
{
  pAttribute->Position = Attribute.Position;
  switch (Token) {
  case 1 /* Ident */:
    pAttribute->Ident.Ident = NoIdent;
    break;
  case 2 /* IntConst */:
    pAttribute->IntConst.Integer = 0L;
    pAttribute->IntConst.Kind = IntDecimal;
    break;
  case 3 /* RealConst */:
    pAttribute->RealConst.String = PutString("0.0", 3);
    break;
  case 4 /* CharConst */:
    pAttribute->CharConst.Char = '?';
    break;
  case 5 /* StringConst */:
    pAttribute->StringConst.String = PutString("", 0);
    break;
  }
}

}

/************************************************************************/

LOCAL {

char String[256], S[256];

}

/************************************************************************/

DEFAULT {

if ( iscntrl(*TokenPtr) ) {
  sprintf( String, "0x%02X", *TokenPtr );
  MessageI( "illegal character", xxError, Attribute.Position, xxString, String );
} else {
  MessageI( "illegal character", xxError, Attribute.Position, xxCharacter, TokenPtr );
}

}

/************************************************************************/

EOF {

if (yyStartState == Comment) {
  Message( "unclosed comment", xxError, Attribute.Position );
  NestingLevel = 0;
}
if (yyStartState != STD) {
  yyStart(STD);
}

}

/************************************************************************/

DEFINE
  digit  =   {0-9}       .
  letter =   {a-z A-Z _} .
  CmtCh  = - {*(\t\n}    .
  StrCh1 = - {'\t\n}     .
  StrCh2 = - {"\t\n}     .

/************************************************************************/

START
  Comment, Str1, Str2

/************************************************************************/

RULES

#STD, Comment# "(*"     :- {NestingLevel++; yyStart(Comment);}
#Comment#  "*)"         :- {NestingLevel--; if (NestingLevel == 0) yyStart(STD);}
#Comment#  "(" | "*" | CmtCh +
                        :- {}

#STD# digit +           ,
#STD# digit + / ".."    ,
#STD# digit + / BY      : {(void)GetWord(Word);
                           Attribute.IntConst.Integer = atoi(Word);
                           Attribute.IntConst.Kind = IntDecimal;
                           return 2;}
#STD# {0-7} + B         : {(void)sscanf(TokenPtr, "%lo", &Attribute.IntConst.Integer);
                           Attribute.IntConst.Kind = IntOctal;
                           return 2;}

/* Read in char-value as int and convert it to char */
#STD# {0-7} + C         : {(void)sscanf(TokenPtr, "%lo", &Attribute.IntConst.Integer);
			   Attribute.CharConst.Char = (char) Attribute.IntConst.Integer;
                           return 4;}
#STD# digit {0-9 A-F} * H
                        : {(void)sscanf(TokenPtr, "%lx", &Attribute.IntConst.Integer);
                           Attribute.IntConst.Kind = IntHexadecimal;
                           return 2;}
#STD# digit + "." digit * (E {+\-} ? digit +) ?
                        : {(void)GetWord(Word);
                           Attribute.RealConst.String = PutString(Word, strlen(Word));
                           return 3;}

#STD#   '               :  {String[0] = '\0'; yyStart(Str1);}
#Str1#  StrCh1 +        :- {(void)GetWord(S); (void)strcat(String, S);}
#Str1#  '               :- {Attribute.StringConst.String = PutString(String, strlen(String));
                            yyStart(STD); return 5;}

#STD#   \"              :  {String[0] = '\0'; yyStart(Str2);}
#Str2#  StrCh2 +        :- {(void)GetWord(S); (void)strcat(String, S);}
#Str2#  \"              :- {Attribute.StringConst.String = PutString(String, strlen(String));
                            yyStart(STD); return 5;}

#Str1, Str2# \t         :- {(void)strcat(String, "\t"); yyTab;}
#Str1, Str2# \n         :- {Message( "unclosed string", xxError, Attribute.Position );
                            Attribute.StringConst.String = PutString( String, strlen(String) );
                            yyEol(0); yyStart(STD); return 5; }

#STD# "&"               : {return 36;}
#STD# "<>"              : {return 40;}
#STD# "~"               : {return 15;}

 #STD#DEFINITION        : { return 6; }
 #STD#MODULE            : { return 7; }
 #STD#\;                : { return 8; }
 #STD#IMPORT            : { return 9; }
 #STD#\,                : { return 10; }
 #STD#FROM              : { return 11; }
 #STD#CONFIGURATION     : { return 12; }
 #STD#\[                : { return 13; }
 #STD#\+                : { return 14; }
 #STD#\NOT              : { return 15; }
 #STD#\{                : { return 16; }
 #STD#\.\.              : { return 17; }
 #STD#\}                : { return 18; }
 #STD#\.                : { return 19; }
 #STD#\<\<              : { return 20; }
 #STD#\>\>              : { return 21; }
 #STD#\<\:              : { return 22; }
 #STD#\*                : { return 23; }
 #STD#\:\>              : { return 24; }
 #STD#\]                : { return 25; }
 #STD#\^                : { return 26; }
 #STD#REDUCE            : { return 27; }
 #STD#\(                : { return 28; }
 #STD#\)                : { return 29; }
 #STD#MOVE              : { return 30; }
 #STD#RECEIVE           : { return 31; }
 #STD#\*\*              : { return 32; }
 #STD#\/                : { return 33; }
 #STD#DIV               : { return 34; }
 #STD#MOD               : { return 35; }
 #STD#AND               : { return 36; }
 #STD#\-                : { return 37; }
 #STD#OR                : { return 38; }
 #STD#\=                : { return 39; }
 #STD#\#                : { return 40; }
 #STD#\<                : { return 41; }
 #STD#\<\=              : { return 42; }
 #STD#\>                : { return 43; }
 #STD#\>\=              : { return 44; }
 #STD#IN                : { return 45; }
 #STD#CONNECTION        : { return 46; }
 #STD#\:                : { return 47; }
 #STD#\-\>              : { return 48; }
 #STD#\<\-\>            : { return 49; }
 #STD#FOR               : { return 50; }
 #STD#\:\=              : { return 51; }
 #STD#TO                : { return 52; }
 #STD#DO                : { return 53; }
 #STD#END               : { return 54; }
 #STD#CONST             : { return 55; }
 #STD#TYPE              : { return 56; }
 #STD#ARRAY             : { return 57; }
 #STD#OF                : { return 58; }
 #STD#RECORD            : { return 59; }
 #STD#CASE              : { return 60; }
 #STD#\|                : { return 61; }
 #STD#ELSE              : { return 62; }
 #STD#SET               : { return 63; }
 #STD#POINTER           : { return 64; }
 #STD#PROCEDURE         : { return 65; }
 #STD#VAR               : { return 66; }
 #STD#FOREIGN           : { return 67; }
 #STD#\EXPORT           : { return 68; }
 #STD#QUALIFIED         : { return 69; }
 #STD#IMPLEMENTATION    : { return 70; }
 #STD#BEGIN             : { return 71; }
 #STD#IF                : { return 72; }
 #STD#THEN              : { return 73; }
 #STD#ELSIF             : { return 74; }
 #STD#WHILE             : { return 75; }
 #STD#REPEAT            : { return 76; }
 #STD#UNTIL             : { return 77; }
 #STD#LOOP              : { return 78; }
 #STD#BY                : { return 79; }
 #STD#WITH              : { return 80; }
 #STD#EXIT              : { return 81; }
 #STD#RETURN            : { return 82; }
 #STD#SEND              : { return 83; }
 #STD#LOAD              : { return 84; }
 #STD#STORE             : { return 85; }
 #STD#ALL               : { return 86; }
 #STD#\:\#              : { return 88; }
 #STD#VECTOR            : { return 89; }

#STD# letter (letter | digit) *
                        : { Attribute.Ident.Ident = MakeIdent(TokenPtr, TokenLength);
                            return 1; }

\r \n                   :- { yyEol( 0 ); }

