/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  Options.c
 * desc  :  Handling of command line options and arguments
 * author:  lampke, keller
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "BigSet.h"
#include "FileIO.h"
#include "GeneralIO.h"
#include "Options.h"

static BOOLEAN isInitialised = FALSE;

#define DEFAULT_SOURCE_INDENT     2

/*
 * local variables
 */

static char *program = "p3c";		  /* Default if no argv[0] present */

static char SourceFile[FILENAME_MAX];
static unsigned long SourceIndent = DEFAULT_SOURCE_INDENT;

static tBigSet Options;

/*
 * local function prototypes
 */
static void Usage( void );
static void SetIndent( char *code );
#ifdef DEVELOP
static void SetDevelopOptions(char *flags);
#endif

/*
 * local function implementation
 */
/*************************************************************************
 * name :  Usage
 * in   :
 * out  :
 ************************************************************************/
static void Usage(void) {

     printf("NAME\n");
     printf("     %s - Parallaxis-III compiler\n", program);
     printf("SYNOPSIS\n");
     printf("     %s [ -c ] [ -g ] [ -h ] [ -i ] [ -In ] [ -m ] [ -Pdir ]\n", program);
#ifdef MPL_VERSION
     printf("    ");
#endif
     printf("         [ -r ] [ -t ] [ -v ] [ -w ]");
#ifdef MPL_VERSION
     printf(" [ -s ]");
#endif
     printf(" sourcefile["MODFILE_EXT"|"DEFFILE_EXT"]\n");
     printf("\nDESCRIPTION\n");
#ifdef MPL_VERSION
     printf("     This is the Parallaxis-III  compiler. It translates programs\n");
     printf("     written  in the  Parallaxis-III  programming  language  into\n");
     printf("     source programs in C or MPL respectively. Definition modules\n");
     printf("     and foreign modules are translated into  C header files (ex-\n");
     printf("     tension .h), implementation and program modules are transla-\n");
     printf("     ted into MPL program files (extension .m) and also C program\n");
     printf("     files (extension .c) if the option -s is not set.\n");
#else
     printf("     This is the Parallaxis-III compiler. It  translates programs\n");
     printf("     written in the Parallaxis-III programming  language  into  C\n");
     printf("     programs.  Definition  modules  and  foreign   modules   are\n");
     printf("     translated into  C  header  files  with  the  extension  .h,\n");
     printf("     implementation and program modules  are  translated  into  C\n");
     printf("     program files with the extension .c.\n");
#endif
     printf("\nOPTIONS\n");
     printf("     -c        Generate type casts to make c-programs lint free.\n");
     printf("     -g        Generate debugging  code  for  implementation  and\n");
     printf("               program modules.  For  all  statements  in  the  C\n");
     printf("               file references to the corresponding statements in\n");
     printf("               the Parallaxis file are generated. Additionally  a\n");
     printf("               crossreference  file   with  the   extension  "CROSSFILE_EXT",\n");
     printf("               containing all Parallaxis  identifiers  and  their\n");
     printf("               corresponding C identifiers, is generated.\n");
     printf("     -h        Print this help information.\n");
     printf("     -i        Generate header files for imported modules.\n");
     printf("     -In       Change the indent of the  code  of  the  generated\n");
     printf("               files to n blanks. The default setting is 2.\n");
     printf("     -m        Print statistics about memory used.\n");
     printf("     -Pdir     Add dir to the definition modules search path.\n");
     printf("     -r        Don't generate runtime range checks.\n");
#ifdef MPL_VERSION
     printf("     -s        Generate \"small\" version using MPL only.  Then the\n");
     printf("               executable must be smaller than 128 kb.\n");
#endif
     printf("     -t        Print statistics about time needed.\n");
     printf("     -v        Verbose mode. Print version number and be somewhat\n");
     printf("               more informative.\n");
     printf("     -w        Suppress warning diagnostics.\n");
#ifdef DEVELOP
     printf("\nADDITIONAL OPTIONS DURING DEVELOPMENT\n");
     printf("     -Zw       Write code tree.\n");
     printf("     -Zs       Write symbols tree.  This  is  only  available  in\n");
     printf("               conjunction with the option -Zw.\n");
     printf("     -Zq       Query code tree.\n");
     printf("     -Zc       Check code tree.\n");
     printf("     -Z1       Run parser only, no semantic check.\n");
     printf("     -Z2       Run parser and semantic check only,  no code gene-\n");
     printf("               ration.\n");
#endif

  exit(-1);
}

/*************************************************************************
 * name :  SetIndent
 * in   :  indent
 * out  :  sets indent
 ************************************************************************/
static void SetIndent(char *code)
{
  unsigned long indent;

  if (code[0] != '\0') {
    indent = atol( code );
    if ( indent != 0 ) {
      SourceIndent = indent;
      BigSet_Include( &Options, Options_eIndent  );
    } else {
      fprintf( stderr, "Numeric value expected: '-I%s'\n\n", code );
#if 0
      Usage(); /* this point is never reached */
#else
      exit(-1);
#endif
    } /* endif */
  } /* endif */
}

#ifdef DEVELOP

/*************************************************************************
 * name :  SetDevelopOptions
 * in   :  develop options
 * out  :  sets develop options
 ************************************************************************/
static void SetDevelopOptions(char *flags)
{
  unsigned long i;

  for (i = 0; flags[i]; i++) {
    switch (flags[i]) {
    case 'w': BigSet_Include(&Options, Options_eWriteTree);        break;
    case 's': BigSet_Include(&Options, Options_eWriteSymbolsTree); break;
    case 'p': BigSet_Include(&Options, Options_eDebugPuma);        break;
    case 'q': BigSet_Include(&Options, Options_eQueryTree);        break;
    case 'c': BigSet_Include(&Options, Options_eCheckTree);        break;
    case '1':
      BigSet_Exclude(&Options, Options_eEvalOnly);
      BigSet_Include(&Options, Options_eCompileOnly);
      break;
    case '2':
      BigSet_Exclude(&Options, Options_eCompileOnly);
      BigSet_Include(&Options, Options_eEvalOnly);
      break;
    default:
      fprintf(stderr, "Illegal develop option : '-Z%c'\n\n", flags[i]);
    }
  }
}

#endif

/*
 * global function implementation
 */
/*------- CheckArguments ---------------------------------------------------*/

/* Remark: the appropriate option bit is set if the feature is turned on, */
/* not if the command line switch is given. E.g. Options_eChecks is set, if */
/* range checks should be generated and unset, if the switch -r is given. */

BOOLEAN Options_CheckArguments(int argc, char *argv[])
{
  static BOOLEAN sourceRead = FALSE;

  short i, j;

  if (argc>0)
       program = argv[0];

  BigSet_Include(&Options, Options_eChecks);   /* use range-checks by default */
  BigSet_Include(&Options, Options_eWarnings); /* give warnings by default */

  for ( i=1; i<argc; i++ ) {
    if ( argv[i][0] == '-' ) {
      switch (argv[i][1]) {
      case 'I': SetIndent(&argv[i][2]);               break;
      case 'P': GeneralIO_AddSearchPath(&argv[i][2]); break;
#ifdef DEVELOP
      case 'Z': SetDevelopOptions(&argv[i][2]);       break;
#endif
      default:
        for (j=1; argv[i][j]; j++) {
          switch (argv[i][j]) {
          case 'c':			  /* generate type casts */
	       BigSet_Include(&Options, Options_eCasts);     
	       break;
	  case 'd':		/* german Errormessages */
	       BigSet_Include(&Options, Options_edeutsch);     
	       break;
          case 'e':		/* english Errormessages */
	       BigSet_Include(&Options, Options_eenglish);     
	       break;
          case 'f':		/* french Errormessages ( Not available yet ) */
	       BigSet_Include(&Options, Options_efrancais);     
	       break;
          case 'g':			  /* generate debugging info */
	       BigSet_Include(&Options, Options_eDebug);    
	       break;
          case 'h':			  /* print usage */
	       Usage(); 
	       break;			  /* this point is never reached */
          case 'i':			  /* generate header files */
	       BigSet_Include(&Options, Options_eInclude);   
	       break;
          case 'm':			  /* print memory statistics */
	       BigSet_Include(&Options, Options_eMemStats);  
	       break;
          case 'r':			  /* don't generate range-checks */
	       BigSet_Exclude(&Options, Options_eChecks);    
	       break;
          case 't':			  /* print time statistics */
	       BigSet_Include(&Options, Options_eTimeStats); 
	       break;
          case 'w':			  /* don't give warnings */
	       BigSet_Exclude(&Options, Options_eWarnings);  
	       break;
	  case 'v':			  /* give additional info */
	       BigSet_Include(&Options, Options_eVerbose);   
	       break;
          default :
            fprintf( stderr, "Illegal option: '-%c'\n\n", argv[i][j] );
#if 0
            Usage(); /* this point is never reached */
#else
	     exit(-1);
#endif
          } /* endswitch */
        } /* endfor */
      } /* endswitch */
    } else {
      if (!sourceRead) {
        sourceRead = TRUE;
        if (!GeneralIO_CheckModFile(argv[i], SourceFile)) {
          fprintf(stderr, "Source file '%s' could not be opened!\n\n",
            argv[i]);
#if 0
          Usage(); /* this point is never reached */
#else
	  exit(-1);
#endif
        } /* endif */
      } else {
        fprintf(stderr, "Illegal second input file specification: '%s'\n\n",
          argv[i]);
#if 0
        Usage(); /* this point is never reached */
#else
	exit(-1);
#endif
      } /* endif */
    } /* endif */
  } /* endfor */

  if (!sourceRead && !Options_IsOptionSet( Options_eVerbose )) {
    fprintf(stderr, "No input file specified!\n\n");
  } /* endif */
  return sourceRead;
}

/*------- GetSourceFile ----------------------------------------------------*/

char *Options_GetSourceFile(void)
{
  if ( SourceFile[0] == '\0' ) {
    return NULL;
  } else {
    return SourceFile;
  } /* endif */
}

/*------- GetSourceIndent --------------------------------------------------*/

ULONG Options_GetSourceIndent(void)
{
  return SourceIndent;
}

/*------- IsOptionSet ------------------------------------------------------*/

BOOLEAN Options_IsOptionSet( tOptions Option )
{
  return(BigSet_IsElement(&Options, Option));
}

void Options_Init(void)
{
  if (!isInitialised) {
    isInitialised = TRUE;
    BigSet_Init();
    FileIO_Init();
    GeneralIO_Init();
    SourceFile[0] = '\0';
    BigSet_MakeSet( &Options, Options_Options );
  } /* endif */
}

