/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * Name  :  GenIdents.c
 * Author:  lampke, keller
 * Date  :  Tue Sep 27 14:31:39 1994
 *
 * Description:
 * ------------
 * Generate (new) identifiers. These are modified versions of existing
 * identifiers (e.g. in the case of name conflicts) or automatically
 * generated new identifiers like labels or help variables.
 *
 ************************************************************************/

#include "Idents.h"			  /* Reuse-Library */
#include <string.h>
#include "parallaxis.h"
#include "GenIdents.h"

static BOOLEAN isInitialised = FALSE;

/*
 * local variables
 */
static unsigned int RenameCnt    = 0;	  /* for identifier renaming */
static unsigned int LabelCnt     = 0;	  /* for generated labels */
static unsigned int ParamCnt     = 0;	  /* for generated parameter names */
static unsigned int WithCnt      = 0;	  /* for identifiers in WITH-blocks */
static unsigned int StructCnt    = 10;	  /* for generated structure names */
static unsigned int GlobalPtrCnt = 0;	  /* for global pointers during the */
					  /* process of unrolling local */
					  /* procedures  */
static unsigned int LocalPtrCnt  = 0;	  /* dito, but local pointers */
static unsigned int BoundCnt     = 0;	  /* for bounds used in FOR */
static unsigned int ReturnCnt    = 0;	  /* for return-values */
static unsigned int StringCnt    = 0;
static unsigned int HelpVarCnt   = 0;	  /* for scratch variables */

static tIdent IdentEXIT;		  /* Id of "EXIT" prefix for labels */

/*
 * local function prototypes
 */
static tIdent GenIdent1( char letter, unsigned long Nr, tIdent Ident );
static tIdent GenIdent2( char letter, unsigned long Nr );
static tIdent GenIdent3( tIdent Ident, unsigned long Nr );

/*-----------------------------------------------------------------------*/

/*
 * local function implementation
 */

/***********************************************************************
 *  Generate a new identifier by prepending the string of the given
 *  identifier with the letter and the given number: "X_nnn_Ident"
 ***********************************************************************/

tIdent GenIdent1( char letter, unsigned long Nr, tIdent Ident )
{
  char NewId[1024];
  char String[1024];

  GetString( Ident, String );
  sprintf( NewId, "%c_%lu_%s", letter, Nr, String );
#if 0
  return MakeIdent( NewId, strlen( NewId ) );
#else
  return GenIdents_MakeNewIdent( NewId );
#endif
}


/***********************************************************************
 *  Generate a new identifier "X_nnn" with the given letter X and the 
 *  given number.
 ***********************************************************************/

tIdent GenIdent2( char letter, unsigned long Nr )
{
  char NewId[1024];

  sprintf( NewId, "%c_%lu", letter, Nr );
#if 0
  return MakeIdent( NewId, strlen( NewId ) );
#else
  return GenIdents_MakeNewIdent( NewId );
#endif
}


/***********************************************************************
 *  Generate a new identifier by appending the given number to the
 *  string of the given identifier: "Ident_nnn"
 ***********************************************************************/

tIdent GenIdent3( tIdent Ident, unsigned long Nr )
{
  char NewId[1024];
  char String[1024];

  GetString( Ident, String );
  sprintf( NewId, "%s_%lu", String, Nr );
#if 0
  return MakeIdent( NewId, strlen( NewId ) );
#else
  return GenIdents_MakeNewIdent( NewId );
#endif
}

/*-----------------------------------------------------------------------*/

/*
 * global function implementation
 */

/***********************************************************************
 *  Generate a definitely new identifier by calling GenIdents_Rename()
 *  until the name is unambiguous. If the requested name is already 
 *  unique, no modification is done. The semantics itself is very similar
 *  to MakeIdent() of reuse's Idents. But no string length is needed.
 ***********************************************************************/

tIdent GenIdents_MakeNewIdent( char *ReqIdName )
{
     tIdent maxid;
     tIdent newid;

     maxid = MaxIdent();
     newid = MakeIdent(ReqIdName, strlen(ReqIdName));
     if (maxid >= newid) return (GenIdents_Rename(newid));
     return newid;
}


/*****************************************************************
 *  Return the id of a (probably new) identifier whose name is
 *  obtained by concatenating the strings of the first identifier 
 *  (the module name) and the second identifier: "Module_Ident"
 *****************************************************************/

tIdent GenIdents_MakeQualified ( tIdent Module, tIdent Ident )
{
  char QualId[1024];
  char String[1024];

  GetString( Module, QualId );
  strcat( QualId, "_");
  GetString( Ident, String );
  strcat( QualId, String );
  return MakeIdent( QualId, strlen(QualId) );
}


/*************************************************************
 *  Return the id of a (probably new) identifier whose name
 *  is obtained by appending the string of the given Ident to
 *  the prefix "C_0_".
 *************************************************************/

tIdent GenIdents_RenameField( tIdent Ident )
{
  return( GenIdent1( 'C', 0, Ident ) );
}


/*************************************************************
 *  Return the id of a new identifier whose name is obtained
 *  by appending the string of the given Ident to the prefix
 *  "C_nnn_" with a new number.
 *************************************************************/

tIdent GenIdents_Rename( tIdent Ident )
{
  RenameCnt++;
  return( GenIdent1( 'C', RenameCnt, Ident ) );
}


/***************************************************
 *  Return the id of a new identifier "EXIT_nnn" with
 *  a new number.
 ***************************************************/

tIdent GenIdents_Label( void )
{
  LabelCnt++;
  return( GenIdent3( IdentEXIT, LabelCnt ) );
}


/********************************************************
 *  Return the id of a (probably new) identifier "X_nnn" 
 *  with the given letter X ('U' for unions or 'V' for 
 *  structs) and the given number.  
 ***************************************************/

tIdent GenIdents_Selector( CHAR StructOrUnion, ULONG Nr )
{
  return( GenIdent2( StructOrUnion, Nr ) );
}

/***************************************************
 *  Return the id of a new identifier "O_nnn"
 ***************************************************/

tIdent GenIdents_Param( void )
{
  ParamCnt++;
  return( GenIdent2( 'O', ParamCnt ) );
}


/***************************************************
 *  Return the id of a new identifier "W_nnn".
 ***************************************************/

tIdent GenIdents_With( void )
{
  WithCnt++;
  return( GenIdent2( 'W', WithCnt ) );
}


/*********************************************************************
 *  Return the id of a (probably new) identifier by appending the 
 *  string of the given identifier (a module name) and the given 
 *  number: "Module_nnn". If the module name is only one character 
 *  long, add another underscore ("Module__nnn") to avoid name con-
 *  flicts with other automatically generated identifiers. E.g. if
 *  the module name is "S" there could be problems with identifiers
 *  generated for structs (see below).
 *********************************************************************/

tIdent GenIdents_Struct1( tIdent Module, ULONG Nr )
{
  char ModuleId[1024];

  GetString( Module, ModuleId );
  if ( strlen( ModuleId ) > 1 ) {
    return GenIdent3( Module, Nr );
  } else {
    strcat( ModuleId, "_" );
    return GenIdent3( MakeIdent( ModuleId, strlen(ModuleId) ), Nr );
  };
}


/***************************************************
 *  Return the id of a new identifier "S_nnn".
 ***************************************************/

tIdent GenIdents_Struct2( void )
{
  StructCnt++;
  return( GenIdent2( 'S', StructCnt ) );
}


/***************************************************
 *  Return the id of a new identifier "G_nnn_Ident" 
 *  with a new number num.
 ***************************************************/

tIdent GenIdents_GlobalPtr( tIdent Ident )
{
  GlobalPtrCnt++;
  return( GenIdent1( 'G', GlobalPtrCnt, Ident ) );
}


/***************************************************
 *  Return the id of a new identifier "L_nnn".
 ***************************************************/

tIdent GenIdents_LocalPtr( void )
{
  LocalPtrCnt++;
  return( GenIdent2( 'L', LocalPtrCnt ) );
}


/***************************************************
 *  Return the id of a new identifier "B_nnn".
 ***************************************************/

tIdent GenIdents_Bound( void )
{
  BoundCnt++;
  return( GenIdent2( 'B', BoundCnt ) );
}


/***************************************************
 *  Return the id of a new identifier "R_nnn".
 ***************************************************/

tIdent GenIdents_Return( void )
{
  ReturnCnt++;
  return( GenIdent2( 'R', ReturnCnt ) );
}


/***************************************************
 *  Return the id of a new identifier "X_nnn".
 ***************************************************/

tIdent GenIdents_String( void )
{
  StringCnt++;
  return( GenIdent2( 'X', StringCnt ) );
}

/*********************************************************************
 *  Return the id of a (probably new) identifier whose name is a
 *  suffix of the given Identifier. The suffix begins behind the
 *  last underscore. Example: "foo_bar_baz" --> identifier "baz".
 *********************************************************************/

tIdent GenIdents_Opaque( tIdent TypeName )
{
  char NewId[1024];
  char QualId[1024];
  unsigned int Index;

  GetString( TypeName, QualId);
  Index = strlen( QualId ) - 1;
  while ( QualId[Index] != '_' ) {
    Index--;
  }
  strcpy( NewId, &(QualId[Index + 1]) );
  return MakeIdent( NewId, strlen( NewId ) );
}


/***************************************************
 *  Return the id of a new identifier "H_nnn".
 ***************************************************/

tIdent GenIdents_HelpVar( void )
{
  HelpVarCnt++;
  return( GenIdent2( 'H', HelpVarCnt ) );
}

/*----------------------------------------------------------------------------*/

/***************************/
/* Initialize this module. */
/***************************/

void GenIdents_Init( void )
{
  if ( !isInitialised ) {
      isInitialised = TRUE;

    RenameCnt     = 0;
    LabelCnt      = 0;
    ParamCnt      = 0;
    WithCnt       = 0;
    StructCnt     = 10; /* because 1 to 9 are reserved for SYSTEM_C */
    GlobalPtrCnt  = 0;
    LocalPtrCnt   = 0;
    BoundCnt      = 0;
    ReturnCnt     = 0;
    StringCnt     = 0;
    HelpVarCnt    = 0;

    IdentEXIT = MakeIdent( "EXIT", 4 );	  /* Name-Prefix for Labels */
  } /* endif */
}
